// ############################################################################################
//  vdmyenv.cpp
//  Implementation of _CVortexAudioEnv class using dummy (NULL) audio system
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <string.h>
// ***** VortexGE headers
#include <vmdisp.h>
#include <auddmy/vdmysf.h>
#include <auddmy/vdmyenv.h>

// ============================================================================================
// Exported variables declaration
// ============================================================================================
// ***** The one and the only DSP environment instance
_CVortexAudioEnv *_pcsAudioEnv_ = NULL;

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The audio environment class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexAudioEnv::_CVortexAudioEnv()
{
  if(_pcsAudioEnv_ == NULL) _pcsAudioEnv_ = this;
}
_CVortexAudioEnv::~_CVortexAudioEnv()
{
  // ..... Set reference to NULL
  _pcsAudioEnv_ = NULL;
}

// ***** Initialization functions
_vbool _CVortexAudioEnv::QueryDeviceName(_vchar *pcResult)
{
  // ..... Check
  if(_pcsAudioEnv_ != this)
  {
    VSetErrorCode(VERR_MultipleInstance);
    VError(_F_, _L_, "Cannot have multiple instance of _CVortexAudioEnv");
    return(false);
  }

  // ..... Just return "dummy"
  strcpy(pcResult, "dummy");
  return(true);
}
_vbool _CVortexAudioEnv::OpenDevice(_vchar *pcDeviceName)
{
  pcDeviceName = pcDeviceName;
  return(true);
}
_vbool _CVortexAudioEnv::InitializeDevice(const _TVortexAudSettings *ptPrefered, _TVortexAudSettings *ptRealized)
{
  if((ptRealized) && (ptRealized)) memcpy(ptRealized, ptPrefered, sizeof(_TVortexAudSettings));
  VWarning(_F_, _L_, "Using dummy (NULL) audio system !!!");
  return(true);
}
_vbool _CVortexAudioEnv::ActivateDevice(_vvoid)
{
  return(true);
}
_vvoid _CVortexAudioEnv::DeactivateDevice(_vvoid)
{
}

// ============================================================================================
// End of vdmyenv.cpp
// ============================================================================================
