// ############################################################################################
//  vdspenv.h
//  Definition & implementation of _CVortexAudioEnv class using direct access to /dev/dsp
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VDSPENV_H_
#define _VDSPENV_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <fcntl.h>
#include <unistd.h>
// ***** VortexGE headers
#include <vthread.h>
#include <vllsbcf.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** Maximum sound buffers can be played together
#define _MAX_SB_ 16

// ============================================================================================
// Foward references of classes
// ============================================================================================
// ***** From "vdspsb.h"
class _CVortexSoundBuffer;

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The audio environment class
class _CVortexAudioEnv
{
  protected:
    // Variables
    struct _TDSPData
    {
      _vchar              acDevName[64];   // The audio device path
      _vint32             n32DSPFD;        // The audio device handle
      _TVortexAudioFormat tFormat;         // The audio data format
      _vuint32            un32FragSize;    // The fragment size in bytes
      _vuint32            un32FragSize2;   // The fragment size in bytes / 2
      _vuint32            un32FragSize4;   // The fragment size in bytes / 4
      _vuint32            un32PBuffSize;   // The primary sound buffer size in bytes
      _vuchar             *pucPBuff;       // The primary sound buffer 
      _vuchar             *pucPBuffMax;    // The primary sound buffer maximum address
      _vuchar             *pucPBuffRead;   // The primary sound buffer read address
      _vuchar             *pucPBuffWrite;  // The primary sound buffer write address
      _vuint8             un8ToWrite;      // Total number of blocks to be written
    } mt_tD; 
    struct _TPSPData
    {
      _TVortexPlayedSBData tData;          // The played sound buffer data
      _CVortexSoundBuffer  *pcsSB;         // The sound buffer
      
    } mt_atPSB[_MAX_SB_];                  // List of played sound buffer
    _CVortexThread mt_csTWriter;           // The device writer thread
    _vvoid (*mt_pfnDoSoundMixing)(_vvoid); // The pointer to the sound mixing function
    // Sound buffer functions
    _vvoid _ClearBuffer_(_vuchar *pucBuff, _vuint32 un32BuffSize, _vuint8 un8Sign, _vuint8 un8Bits);
    
  public:
    // Constructor & destructor
    _CVortexAudioEnv();
    ~_CVortexAudioEnv();
    // Initialization functions
    _vbool QueryDeviceName(_vchar *pcResult);
    _vbool OpenDevice(_vchar *pcDeviceName);
    _vbool InitializeDevice(const _TVortexAudSettings *ptPrefered = NULL, _TVortexAudSettings *ptRealized = NULL);
    _vbool ActivateDevice(_vvoid);
    _vvoid DeactivateDevice(_vvoid);
    inline _vvoid SuspendDevice(_vvoid);
    inline _vvoid ResumeDevice(_vvoid);
    // Sound functions
    inline _vvoid UpdateBuffer(_vvoid);
  
  // Friend classes
  friend class _CVortexSoundBuffer;
    
  // Friend functions
  friend _vuint64 _DSPWriter_(_TVortexThreadData *ptThreadData);
  friend _vvoid _DSPMixerU8m_(_vvoid);
  friend _vvoid _DSPMixerU8s_(_vvoid);
  friend _vvoid _DSPMixerS8m_(_vvoid);
  friend _vvoid _DSPMixerS8s_(_vvoid);
  friend _vvoid _DSPMixerU16m_(_vvoid);
  friend _vvoid _DSPMixerU16s_(_vvoid);
  friend _vvoid _DSPMixerS16m_(_vvoid);
  friend _vvoid _DSPMixerS16s_(_vvoid);
  friend _vvoid _DSPIMixerU16m_(_vvoid);
  friend _vvoid _DSPIMixerU16s_(_vvoid);
  friend _vvoid _DSPIMixerS16m_(_vvoid);
  friend _vvoid _DSPIMixerS16s_(_vvoid);
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The audio environment class
// --------------------------------------------------------------------------------------------
// ***** Initialization functions
_vvoid _CVortexAudioEnv::SuspendDevice(_vvoid)
{
  mt_csTWriter.Suspend();
}
_vvoid _CVortexAudioEnv::ResumeDevice(_vvoid)
{
  mt_csTWriter.Resume();
}

// ***** Sound functions
_vvoid _CVortexAudioEnv::UpdateBuffer(_vvoid)
{
  // ..... Check
  if(mt_tD.n32DSPFD < 0) return;

  // ..... Do sound mixing
  mt_pfnDoSoundMixing();
}

#endif
// ============================================================================================
// End of vdspenv.h
// ============================================================================================
