// ############################################################################################
//  vdspsf.cpp
//  Implementation of sound functions used by _CVortexSoundBuffer class
//  using direct access to /dev/dsp
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <time.h>
#include <string.h>
// ***** VortexGE headers
#include <auddsp/vdspsb.h>
#include <auddsp/vdspsf.h>

// ============================================================================================
// Reference to external variables
// ============================================================================================
// ***** The one and the only DSP environment instance
extern _CVortexAudioEnv *_pcsAudioEnv_;

// ============================================================================================
// Structures
// ============================================================================================
// ***** Used to swap values
struct _TSwapValue
{
  _vint8 n8Val1;
  _vint8 n8Val2;
};

// ============================================================================================
// Exported functions implementation
// ============================================================================================
// ***** Writer functions
_vuint64 _DSPWriter_(_TVortexThreadData *ptThreadData)
{
  while(!(*ptThreadData->pbExit))
  {
    // Suspended ?
    if((*ptThreadData->pbSuspend) || (_pcsAudioEnv_->mt_tD.un8ToWrite > 3))
    {
      usleep(100);
      continue;
    }
    
    // Write data
    write(_pcsAudioEnv_->mt_tD.n32DSPFD, _pcsAudioEnv_->mt_tD.pucPBuffRead,
          _pcsAudioEnv_->mt_tD.un32FragSize);
    _pcsAudioEnv_->mt_tD.un8ToWrite++;
    
    // Increment read pointer
    _pcsAudioEnv_->mt_tD.pucPBuffRead += _pcsAudioEnv_->mt_tD.un32FragSize;
    if(_pcsAudioEnv_->mt_tD.pucPBuffRead > _pcsAudioEnv_->mt_tD.pucPBuffMax)
      _pcsAudioEnv_->mt_tD.pucPBuffRead = _pcsAudioEnv_->mt_tD.pucPBuff;
  }
  return(0);
}

// ***** Normal sound mixing functions
_vvoid _DSPMixerU8m_(_vvoid) // *** Unchecked ***
{
  // ..... Static variables
  static _vuint8  un8Cnt;
  static _vuint32 un32Cnt;
  static _vuint8  *pun8Buff;
  static _vint8   n8Data;
  static _vint32  n32Sum;

  // ..... Check
  if(_pcsAudioEnv_->mt_tD.un8ToWrite == 0) return;
    
  // ..... Mix sound
  pun8Buff = (_vuint8 *) _pcsAudioEnv_->mt_tD.pucPBuffWrite;
  for(un32Cnt = 0; un32Cnt < _pcsAudioEnv_->mt_tD.un32FragSize4; un32Cnt++)
  {
    n32Sum = 0;
    for(un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
    {
      if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB)
      {
        if(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying)
        {
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB->_GetData_(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData, (_vuchar *) &n8Data);
          n32Sum += n8Data;
        }
        else
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      }
    }
    if(n32Sum < -128)     *pun8Buff++ = 0;
    else if(n32Sum > 127) *pun8Buff++ = 255;
    else                  *pun8Buff++ = _vint8(n32Sum) ^ 0x80;
  }
  _pcsAudioEnv_->mt_tD.un8ToWrite--;
    
  // ..... Increment write pointer
  _pcsAudioEnv_->mt_tD.pucPBuffWrite += _pcsAudioEnv_->mt_tD.un32FragSize;
  if(_pcsAudioEnv_->mt_tD.pucPBuffWrite > _pcsAudioEnv_->mt_tD.pucPBuffMax)
    _pcsAudioEnv_->mt_tD.pucPBuffWrite = _pcsAudioEnv_->mt_tD.pucPBuff;
}
_vvoid _DSPMixerU8s_(_vvoid) // *** Unchecked ***
{
  // ..... Static variables
  static _vuint8  un8Cnt;
  static _vuint32 un32Cnt;
  static _vuint8  *pun8Buff;
  static _vint8   an8Data[2];
  static _vint32  an32Sum[2];

  // ..... Check
  if(_pcsAudioEnv_->mt_tD.un8ToWrite == 0) return;
    
  // ..... Mix sound
  pun8Buff = (_vuint8 *) _pcsAudioEnv_->mt_tD.pucPBuffWrite;
  for(un32Cnt = 0; un32Cnt < _pcsAudioEnv_->mt_tD.un32FragSize4; un32Cnt++)
  {
    an32Sum[1] = an32Sum[0] = 0;
    for(un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
    {
      if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB)
      {
        if(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying)
        {
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB->_GetData_(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData, (_vuchar *) an8Data);
          an32Sum[0] += an8Data[0];
          an32Sum[1] += an8Data[1];
        }
        else
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      }
    }
    if(an32Sum[0] < -128)     *pun8Buff++ = 0;
    else if(an32Sum[0] > 127) *pun8Buff++ = 255;
    else                      *pun8Buff++ = _vint8(an32Sum[0]) ^ 0x80;
    if(an32Sum[1] < -128)     *pun8Buff++ = 0;
    else if(an32Sum[1] > 127) *pun8Buff++ = 255;
    else                      *pun8Buff++ = _vint8(an32Sum[1]) ^ 0x80;
  }
  _pcsAudioEnv_->mt_tD.un8ToWrite--;
    
  // ..... Increment write pointer
  _pcsAudioEnv_->mt_tD.pucPBuffWrite += _pcsAudioEnv_->mt_tD.un32FragSize;
  if(_pcsAudioEnv_->mt_tD.pucPBuffWrite > _pcsAudioEnv_->mt_tD.pucPBuffMax)
    _pcsAudioEnv_->mt_tD.pucPBuffWrite = _pcsAudioEnv_->mt_tD.pucPBuff;
}
_vvoid _DSPMixerS8m_(_vvoid) // *** Unchecked ***
{
  // ..... Static variables
  static _vuint8  un8Cnt;
  static _vuint32 un32Cnt;
  static _vint8   *pn8Buff, n8Data;
  static _vint32  n32Sum;

  // ..... Check
  if(_pcsAudioEnv_->mt_tD.un8ToWrite == 0) return;
    
  // ..... Mix sound
  pn8Buff = (_vint8 *) _pcsAudioEnv_->mt_tD.pucPBuffWrite;
  for(un32Cnt = 0; un32Cnt < _pcsAudioEnv_->mt_tD.un32FragSize4; un32Cnt++)
  {
    n32Sum = 0;
    for(un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
    {
      if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB)
      {
        if(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying)
        {
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB->_GetData_(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData, (_vuchar *) &n8Data);
          n32Sum += n8Data;
        }
        else
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      }
    }
    if(n32Sum < -128)     *pn8Buff++ = -128;
    else if(n32Sum > 127) *pn8Buff++ = 127;
    else                  *pn8Buff++ = _vint8(n32Sum);
  }
  _pcsAudioEnv_->mt_tD.un8ToWrite--;
    
  // ..... Increment write pointer
  _pcsAudioEnv_->mt_tD.pucPBuffWrite += _pcsAudioEnv_->mt_tD.un32FragSize;
  if(_pcsAudioEnv_->mt_tD.pucPBuffWrite > _pcsAudioEnv_->mt_tD.pucPBuffMax)
    _pcsAudioEnv_->mt_tD.pucPBuffWrite = _pcsAudioEnv_->mt_tD.pucPBuff;
}
_vvoid _DSPMixerS8s_(_vvoid) // *** Unchecked ***
{
  // ..... Static variables
  static _vuint8  un8Cnt;
  static _vuint32 un32Cnt;
  static _vint8   *pn8Buff, an8Data[2];
  static _vint32  an32Sum[2];

  // ..... Check
  if(_pcsAudioEnv_->mt_tD.un8ToWrite == 0) return;
    
  // ..... Mix sound
  pn8Buff = (_vint8 *) _pcsAudioEnv_->mt_tD.pucPBuffWrite;
  for(un32Cnt = 0; un32Cnt < _pcsAudioEnv_->mt_tD.un32FragSize4; un32Cnt++)
  {
    an32Sum[1] = an32Sum[0] = 0;
    for(un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
    {
      if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB)
      {
        if(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying)
        {
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB->_GetData_(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData, (_vuchar *) an8Data);
          an32Sum[0] += an8Data[0];
          an32Sum[1] += an8Data[1];
        }
        else
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      }
    }
    if(an32Sum[0] < -128)     *pn8Buff++ = -128;
    else if(an32Sum[0] > 127) *pn8Buff++ = 127;
    else                      *pn8Buff++ = _vint8(an32Sum[0]);
    if(an32Sum[1] < -128)     *pn8Buff++ = -128;
    else if(an32Sum[1] > 127) *pn8Buff++ = 127;
    else                      *pn8Buff++ = _vint8(an32Sum[1]);
  }
  _pcsAudioEnv_->mt_tD.un8ToWrite--;
    
  // ..... Increment write pointer
  _pcsAudioEnv_->mt_tD.pucPBuffWrite += _pcsAudioEnv_->mt_tD.un32FragSize;
  if(_pcsAudioEnv_->mt_tD.pucPBuffWrite > _pcsAudioEnv_->mt_tD.pucPBuffMax)
    _pcsAudioEnv_->mt_tD.pucPBuffWrite = _pcsAudioEnv_->mt_tD.pucPBuff;
}
_vvoid _DSPMixerU16m_(_vvoid) // *** Unchecked ***
{
  // ..... Static variables
  static _vuint8  un8Cnt;
  static _vuint32 un32Cnt;
  static _vuint16 *pun16Buff;
  static _vint16  n16Data;
  static _vint32  n32Sum;

  // ..... Check
  if(_pcsAudioEnv_->mt_tD.un8ToWrite == 0) return;
    
  // ..... Mix sound
  pun16Buff = (_vuint16 *) _pcsAudioEnv_->mt_tD.pucPBuffWrite;
  for(un32Cnt = 0; un32Cnt < _pcsAudioEnv_->mt_tD.un32FragSize4; un32Cnt++)
  {
    n32Sum = 0;
    for(un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
    {
      if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB)
      {
        if(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying)
        {
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB->_GetData_(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData, (_vuchar *) &n16Data);
          n32Sum += n16Data;
        }
        else
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      }
    }
    if(n32Sum < -32768)     *pun16Buff++ = 0;
    else if(n32Sum > 32767) *pun16Buff++ = 65535;
    else                    *pun16Buff++ = _vuint16(n32Sum) ^ 0x8000;
  }
  _pcsAudioEnv_->mt_tD.un8ToWrite--;
    
  // ..... Increment write pointer
  _pcsAudioEnv_->mt_tD.pucPBuffWrite += _pcsAudioEnv_->mt_tD.un32FragSize;
  if(_pcsAudioEnv_->mt_tD.pucPBuffWrite > _pcsAudioEnv_->mt_tD.pucPBuffMax)
    _pcsAudioEnv_->mt_tD.pucPBuffWrite = _pcsAudioEnv_->mt_tD.pucPBuff;
}
_vvoid _DSPMixerU16s_(_vvoid) // *** Unchecked ***
{
  // ..... Static variables
  static _vuint8  un8Cnt;
  static _vuint32 un32Cnt;
  static _vuint16 *pun16Buff;
  static _vint16  an16Data[2];
  static _vint32  an32Sum[2];

  // ..... Check
  if(_pcsAudioEnv_->mt_tD.un8ToWrite == 0) return;
    
  // ..... Mix sound
  pun16Buff = (_vuint16 *) _pcsAudioEnv_->mt_tD.pucPBuffWrite;
  for(un32Cnt = 0; un32Cnt < _pcsAudioEnv_->mt_tD.un32FragSize4; un32Cnt++)
  {
    an32Sum[1] = an32Sum[0] = 0;
    for(un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
    {
      if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB)
      {
        if(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying)
        {
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB->_GetData_(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData, (_vuchar *) an16Data);
          an32Sum[0] += an16Data[0];
          an32Sum[1] += an16Data[1];
        }
        else
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      }
    }
    if(an32Sum[0] < -32768)     *pun16Buff++ = 0;
    else if(an32Sum[0] > 32767) *pun16Buff++ = 65535;
    else                        *pun16Buff++ = _vuint16(an32Sum[0]) ^ 0x8000;
    if(an32Sum[1] < -32768)     *pun16Buff++ = 0;
    else if(an32Sum[1] > 32767) *pun16Buff++ = 65535;
    else                        *pun16Buff++ = _vuint16(an32Sum[1]) ^ 0x8000;
  }
  _pcsAudioEnv_->mt_tD.un8ToWrite--;
    
  // ..... Increment write pointer
  _pcsAudioEnv_->mt_tD.pucPBuffWrite += _pcsAudioEnv_->mt_tD.un32FragSize;
  if(_pcsAudioEnv_->mt_tD.pucPBuffWrite > _pcsAudioEnv_->mt_tD.pucPBuffMax)
    _pcsAudioEnv_->mt_tD.pucPBuffWrite = _pcsAudioEnv_->mt_tD.pucPBuff;
}
_vvoid _DSPMixerS16m_(_vvoid) // *** Unchecked ***
{
  // ..... Static variables
  static _vuint8  un8Cnt;
  static _vuint32 un32Cnt;
  static _vint16  *pn16Buff, n16Data;
  static _vint32  n32Sum;

  // ..... Check
  if(_pcsAudioEnv_->mt_tD.un8ToWrite == 0) return;
    
  // ..... Mix sound
  pn16Buff = (_vint16 *) _pcsAudioEnv_->mt_tD.pucPBuffWrite;
  for(un32Cnt = 0; un32Cnt < _pcsAudioEnv_->mt_tD.un32FragSize4; un32Cnt++)
  {
    n32Sum = 0;
    for(un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
    {
      if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB)
      {
        if(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying)
        {
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB->_GetData_(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData, (_vuchar *) &n16Data);
          n32Sum += n16Data;
        }
        else
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      }
    }
    if(n32Sum < -32768)     *pn16Buff++ = -32768;
    else if(n32Sum > 32767) *pn16Buff++ = 32767;
    else                    *pn16Buff++ = _vint16(n32Sum);
  }
  _pcsAudioEnv_->mt_tD.un8ToWrite--;
    
  // ..... Increment write pointer
  _pcsAudioEnv_->mt_tD.pucPBuffWrite += _pcsAudioEnv_->mt_tD.un32FragSize;
  if(_pcsAudioEnv_->mt_tD.pucPBuffWrite > _pcsAudioEnv_->mt_tD.pucPBuffMax)
    _pcsAudioEnv_->mt_tD.pucPBuffWrite = _pcsAudioEnv_->mt_tD.pucPBuff;
}
_vvoid _DSPMixerS16s_(_vvoid)
{
  // ..... Static variables
  static _vuint8  un8Cnt;
  static _vuint32 un32Cnt;
  static _vint16  *pn16Buff, an16Data[2];
  static _vint32  an32Sum[2];

  // ..... Check
  if(_pcsAudioEnv_->mt_tD.un8ToWrite == 0) return;
    
  // ..... Mix sound
  pn16Buff = (_vint16 *) _pcsAudioEnv_->mt_tD.pucPBuffWrite;
  for(un32Cnt = 0; un32Cnt < _pcsAudioEnv_->mt_tD.un32FragSize4; un32Cnt++)
  {
    an32Sum[1] = an32Sum[0] = 0;
    for(un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
    {
      if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB)
      {
        if(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying)
        {
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB->_GetData_(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData, (_vuchar *) an16Data);
          an32Sum[0] += an16Data[0];
          an32Sum[1] += an16Data[1];
        }
        else
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      }
    }
    if(an32Sum[0] < -32768)     *pn16Buff++ = -32768;
    else if(an32Sum[0] > 32767) *pn16Buff++ = 32767;
    else                        *pn16Buff++ = _vint16(an32Sum[0]);
    if(an32Sum[1] < -32768)     *pn16Buff++ = -32768;
    else if(an32Sum[1] > 32767) *pn16Buff++ = 32767;
    else                        *pn16Buff++ = _vint16(an32Sum[1]);
  }
  _pcsAudioEnv_->mt_tD.un8ToWrite--;
    
  // ..... Increment write pointer
  _pcsAudioEnv_->mt_tD.pucPBuffWrite += _pcsAudioEnv_->mt_tD.un32FragSize;
  if(_pcsAudioEnv_->mt_tD.pucPBuffWrite > _pcsAudioEnv_->mt_tD.pucPBuffMax)
    _pcsAudioEnv_->mt_tD.pucPBuffWrite = _pcsAudioEnv_->mt_tD.pucPBuff;
}

// ***** Endian inverted sound mixing functions
_vvoid _DSPIMixerU16m_(_vvoid) // *** Unchecked ***
{
  // ..... Static variables
  static _vuint8  un8Cnt;
  static _vuint32 un32Cnt;
  static _vuint16 *pun16Buff, un16Tmp;
  static _vint16  n16Data;
  static _vint32  n32Sum;

  // ..... Check
  if(_pcsAudioEnv_->mt_tD.un8ToWrite == 0) return;
    
  // ..... Mix sound
  pun16Buff = (_vuint16 *) _pcsAudioEnv_->mt_tD.pucPBuffWrite;
  for(un32Cnt = 0; un32Cnt < _pcsAudioEnv_->mt_tD.un32FragSize4; un32Cnt++)
  {
    n32Sum = 0;
    for(un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
    {
      if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB)
      {
        if(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying)
        {
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB->_GetData_(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData, (_vuchar *) &n16Data);
          n32Sum += n16Data;
        }
        else
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      }
    }
    if(n32Sum < -32768)     un16Tmp = 0;
    else if(n32Sum > 32767) un16Tmp = 65535;
    else                    un16Tmp = _vuint16(n32Sum) ^ 0x8000;
    ((_TSwapValue *) pun16Buff)->n8Val1 = ((_TSwapValue *) &un16Tmp)->n8Val2;
    ((_TSwapValue *) pun16Buff)->n8Val2 = ((_TSwapValue *) &un16Tmp)->n8Val1;
    pun16Buff++;
  }
  _pcsAudioEnv_->mt_tD.un8ToWrite--;
    
  // ..... Increment write pointer
  _pcsAudioEnv_->mt_tD.pucPBuffWrite += _pcsAudioEnv_->mt_tD.un32FragSize;
  if(_pcsAudioEnv_->mt_tD.pucPBuffWrite > _pcsAudioEnv_->mt_tD.pucPBuffMax)
    _pcsAudioEnv_->mt_tD.pucPBuffWrite = _pcsAudioEnv_->mt_tD.pucPBuff;
}
_vvoid _DSPIMixerU16s_(_vvoid) // *** Unchecked ***
{
  // ..... Static variables
  static _vuint8  un8Cnt;
  static _vuint32 un32Cnt;
  static _vuint16 *pun16Buff, un16Tmp;
  static _vint16  an16Data[2];
  static _vint32  an32Sum[2];

  // ..... Check
  if(_pcsAudioEnv_->mt_tD.un8ToWrite == 0) return;
    
  // ..... Mix sound
  pun16Buff = (_vuint16 *) _pcsAudioEnv_->mt_tD.pucPBuffWrite;
  for(un32Cnt = 0; un32Cnt < _pcsAudioEnv_->mt_tD.un32FragSize4; un32Cnt++)
  {
    an32Sum[1] = an32Sum[0] = 0;
    for(un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
    {
      if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB)
      {
        if(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying)
        {
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB->_GetData_(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData, (_vuchar *) an16Data);
          an32Sum[0] += an16Data[0];
          an32Sum[1] += an16Data[1];
        }
        else
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      }
    }
    if(an32Sum[0] < -32768)     un16Tmp = 0;
    else if(an32Sum[0] > 32767) un16Tmp = 65535;
    else                        un16Tmp = _vuint16(an32Sum[0]) ^ 0x8000;
    ((_TSwapValue *) pun16Buff)->n8Val1 = ((_TSwapValue *) &un16Tmp)->n8Val2;
    ((_TSwapValue *) pun16Buff)->n8Val2 = ((_TSwapValue *) &un16Tmp)->n8Val1;
    pun16Buff++;
    if(an32Sum[1] < -32768)     un16Tmp = 0;
    else if(an32Sum[1] > 32767) un16Tmp = 65535;
    else                        un16Tmp = _vuint16(an32Sum[1]) ^ 0x8000;
    ((_TSwapValue *) pun16Buff)->n8Val1 = ((_TSwapValue *) &un16Tmp)->n8Val2;
    ((_TSwapValue *) pun16Buff)->n8Val2 = ((_TSwapValue *) &un16Tmp)->n8Val1;
    pun16Buff++;
  }
  _pcsAudioEnv_->mt_tD.un8ToWrite--;
    
  // ..... Increment write pointer
  _pcsAudioEnv_->mt_tD.pucPBuffWrite += _pcsAudioEnv_->mt_tD.un32FragSize;
  if(_pcsAudioEnv_->mt_tD.pucPBuffWrite > _pcsAudioEnv_->mt_tD.pucPBuffMax)
    _pcsAudioEnv_->mt_tD.pucPBuffWrite = _pcsAudioEnv_->mt_tD.pucPBuff;
}
_vvoid _DSPIMixerS16m_(_vvoid) // *** Unchecked ***
{
  // ..... Static variables
  static _vuint8  un8Cnt;
  static _vuint32 un32Cnt;
  static _vint16  *pn16Buff, n16Data, n16Tmp;
  static _vint32  n32Sum;

  // ..... Check
  if(_pcsAudioEnv_->mt_tD.un8ToWrite == 0) return;
    
  // ..... Mix sound
  pn16Buff = (_vint16 *) _pcsAudioEnv_->mt_tD.pucPBuffWrite;
  for(un32Cnt = 0; un32Cnt < _pcsAudioEnv_->mt_tD.un32FragSize4; un32Cnt++)
  {
    n32Sum = 0;
    for(un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
    {
      if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB)
      {
        if(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying)
        {
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB->_GetData_(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData, (_vuchar *) &n16Data);
          n32Sum += n16Data;
        }
        else
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      }
    }
    if(n32Sum < -32768)     n16Tmp = -32768;
    else if(n32Sum > 32767) n16Tmp = 32767;
    else                    n16Tmp = _vint16(n32Sum);
    ((_TSwapValue *) pn16Buff)->n8Val1 = ((_TSwapValue *) &n16Tmp)->n8Val2;
    ((_TSwapValue *) pn16Buff)->n8Val2 = ((_TSwapValue *) &n16Tmp)->n8Val1;
    pn16Buff++;
  }
  _pcsAudioEnv_->mt_tD.un8ToWrite--;
    
  // ..... Increment write pointer
  _pcsAudioEnv_->mt_tD.pucPBuffWrite += _pcsAudioEnv_->mt_tD.un32FragSize;
  if(_pcsAudioEnv_->mt_tD.pucPBuffWrite > _pcsAudioEnv_->mt_tD.pucPBuffMax)
    _pcsAudioEnv_->mt_tD.pucPBuffWrite = _pcsAudioEnv_->mt_tD.pucPBuff;
}
_vvoid _DSPIMixerS16s_(_vvoid) // *** Unchecked ***
{
  // ..... Static variables
  static _vuint8  un8Cnt;
  static _vuint32 un32Cnt;
  static _vint16  *pn16Buff, an16Data[2], n16Tmp;
  static _vint32  an32Sum[2];

  // ..... Check
  if(_pcsAudioEnv_->mt_tD.un8ToWrite == 0) return;
    
  // ..... Mix sound
  pn16Buff = (_vint16 *) _pcsAudioEnv_->mt_tD.pucPBuffWrite;
  for(un32Cnt = 0; un32Cnt < _pcsAudioEnv_->mt_tD.un32FragSize4; un32Cnt++)
  {
    an32Sum[1] = an32Sum[0] = 0;
    for(un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
    {
      if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB)
      {
        if(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying)
        {
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB->_GetData_(_pcsAudioEnv_->mt_atPSB[un8Cnt].tData, (_vuchar *) an16Data);
          an32Sum[0] += an16Data[0];
          an32Sum[1] += an16Data[1];
        }
        else
          _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      }
    }
    if(an32Sum[0] < -32768)     n16Tmp = -32768;
    else if(an32Sum[0] > 32767) n16Tmp = 32767;
    else                        n16Tmp = _vint16(an32Sum[0]);
    ((_TSwapValue *) pn16Buff)->n8Val1 = ((_TSwapValue *) &n16Tmp)->n8Val2;
    ((_TSwapValue *) pn16Buff)->n8Val2 = ((_TSwapValue *) &n16Tmp)->n8Val1;
    pn16Buff++;
    if(an32Sum[1] < -32768)     n16Tmp = -32768;
    else if(an32Sum[1] > 32767) n16Tmp = 32767;
    else                        n16Tmp = _vint16(an32Sum[1]);
    ((_TSwapValue *) pn16Buff)->n8Val1 = ((_TSwapValue *) &n16Tmp)->n8Val2;
    ((_TSwapValue *) pn16Buff)->n8Val2 = ((_TSwapValue *) &n16Tmp)->n8Val1;
    pn16Buff++;
  }
  _pcsAudioEnv_->mt_tD.un8ToWrite--;
    
  // ..... Increment write pointer
  _pcsAudioEnv_->mt_tD.pucPBuffWrite += _pcsAudioEnv_->mt_tD.un32FragSize;
  if(_pcsAudioEnv_->mt_tD.pucPBuffWrite > _pcsAudioEnv_->mt_tD.pucPBuffMax)
    _pcsAudioEnv_->mt_tD.pucPBuffWrite = _pcsAudioEnv_->mt_tD.pucPBuff;
}

// ============================================================================================
// End of vdspsf.cpp
// ============================================================================================
