// ############################################################################################
//  vaimgctl.cpp
//  Implementation  of _CVortexTexture, _CVortexAnimatedTexture, _CVortexImage,
//  _CVortexAnimatedImage, and _CVortexImageBuilder class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vmdisp.h>
#include <vaimgctl.h>

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The advanced frame control for animated image class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexAImgControl::_CVortexAImgControl()
{
  mt_pcsImgSrc    = NULL;
  mt_ppun8ImgBuff = NULL;
  mt_ptASList     = NULL;
  mt_un32TotalAS  = 0;
  mt_un32SelAS    = 0;
  mt_un32SelFrame = 0;
}
_CVortexAImgControl::~_CVortexAImgControl()
{
  FreeAll();
}

// ***** Main initialization functions
_vbool _CVortexAImgControl::SetImageSource(_CVortexAnimatedImage *pcsAnimatedImage)
{
  // ..... Check
  if(pcsAnimatedImage->mt_un32TotFrame == 0)
  {
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Cannot use an uninitialized _CVortexAnimatedImage");
    return(false);    
  }
  
  // ..... Select new image source
  mt_pcsImgSrc = pcsAnimatedImage;
  
  // ..... Destroy old reference buffer, if any
  if(mt_ppun8ImgBuff) free(mt_ppun8ImgBuff);
    
  // ..... Allocate memory
  mt_ppun8ImgBuff = (_vuint8 **) malloc(mt_pcsImgSrc->mt_un32TotFrame * sizeof(_vuint8 *));
  if(mt_ppun8ImgBuff == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for a new reference buffer");
    return(false);    
  }
  
  // ..... Set references
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_pcsImgSrc->mt_un32TotFrame; un32Cnt++)
    mt_ppun8ImgBuff[un32Cnt] = mt_pcsImgSrc->mt_pun8BufferArray + mt_pcsImgSrc->mt_un32FrameSize * un32Cnt;

  // ..... Refersh data
  mt_un32SelAS     = 0;
  mt_un32SelFrame  = 0;
  mt_un32StepDelay = 0;
  mt_pcsImgSrc->mt_pun8Buffer = mt_pcsImgSrc->mt_pun8BufferArray;
  
  // ..... OK !
  return(true);
}
_vvoid _CVortexAImgControl::FreeAll(_vvoid)
{
  if(mt_ptASList)
  {
    for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32TotalAS; un32Cnt++)
      if(mt_ptASList[un32Cnt].pun32FrameList) free(mt_ptASList[un32Cnt].pun32FrameList);
    free(mt_ptASList);   
  } 
  if(mt_ppun8ImgBuff) free(mt_ppun8ImgBuff);
  mt_ptASList     = NULL;
  mt_ppun8ImgBuff = NULL;
  mt_un32TotalAS  = 0;
  mt_un32SelAS    = 0;
  mt_un32SelFrame = 0;
}

// ***** Animation set management functions
_vbool _CVortexAImgControl::AddAnimationSet(_vuint32 un32TotalASFrames)
{
  // ..... Temporary variables
  _TAnimSet *ptAS;

  // ..... Allocate memory
  ptAS = (_TAnimSet *) malloc((mt_un32TotalAS + 1) * sizeof(_TAnimSet));
  if(ptAS == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for a new animation set");
    return(false);    
  }
  
  // ..... Copy and free old data
  memcpy(ptAS, mt_ptASList, mt_un32TotalAS * sizeof(_TAnimSet));
  memset(&ptAS[mt_un32TotalAS], 0, sizeof(_TAnimSet));
  free(mt_ptASList);
  mt_ptASList = ptAS;
  
  // ..... Allocate memory for frames
  return(_AllocASFrame_(mt_un32TotalAS++, un32TotalASFrames));
}
_vbool _CVortexAImgControl::AddAnimationSets(_vuint32 un32ASCount, _vuint32 aun32TotalASFrames[])
{
  // ..... Temporary variables
  _TAnimSet *ptAS;
  _vuint32  un32Cnt;

  // ..... Allocate memory
  ptAS = (_TAnimSet *) malloc((mt_un32TotalAS + un32ASCount) * sizeof(_TAnimSet));
  if(ptAS == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for new animation sets");
    return(false);    
  }
  
  // ..... Copy and free old data
  memcpy(ptAS, mt_ptASList, mt_un32TotalAS * sizeof(_TAnimSet));
  memset(&ptAS[mt_un32TotalAS], 0, un32ASCount * sizeof(_TAnimSet));
  free(mt_ptASList);
  mt_ptASList = ptAS;
  
  // ..... Allocate memory for frames
  if(aun32TotalASFrames)
  {
    for(un32Cnt = 0; un32Cnt < un32ASCount; un32Cnt++)
      if(!_AllocASFrame_(mt_un32TotalAS + un32Cnt, aun32TotalASFrames[un32Cnt])) return(false);
  }
  mt_un32TotalAS += un32ASCount;
  return(true);
}
_vuint32 _CVortexAImgControl::GetAnimationSetCount(_vvoid)
{
  return(mt_un32TotalAS);
}

// ***** Frame management functions
_vbool _CVortexAImgControl::AddASFrame(_vuint32 un32AnimationSetIndex, _vuint32 un32FrameReference)
{
  // ..... Temporary variables
  _vuint32 *pun32Tmp, un32Tot;

  // ..... Save old datas
  pun32Tmp = mt_ptASList[un32AnimationSetIndex].pun32FrameList;
  un32Tot  = mt_ptASList[un32AnimationSetIndex].un32TotalFrames;

  // ..... Allocate memory
  if(!_AllocASFrame_(un32AnimationSetIndex, mt_ptASList[un32AnimationSetIndex].un32TotalFrames + 1)) return(false);

  // ..... Copy and free old data
  memcpy(mt_ptASList[un32AnimationSetIndex].pun32FrameList, pun32Tmp,  un32Tot * sizeof(_vuint32));
  free(pun32Tmp);

  // ..... Set new data
  mt_ptASList[un32AnimationSetIndex].pun32FrameList[un32Tot] = un32FrameReference;
  
  // ..... OK !
  return(true);
}
_vbool _CVortexAImgControl::AddASFrames(_vuint32 un32AnimationSetIndex, _vuint32 un32FrameCount, _vuint32 aun32FrameReferences[])
{
  // ..... Temporary variables
  _vuint32 *pun32Tmp, un32Tot, un32Cnt;

  // ..... Save old datas
  pun32Tmp = mt_ptASList[un32AnimationSetIndex].pun32FrameList;
  un32Tot  = mt_ptASList[un32AnimationSetIndex].un32TotalFrames;

  // ..... Allocate memory
  if(!_AllocASFrame_(un32AnimationSetIndex, mt_ptASList[un32AnimationSetIndex].un32TotalFrames + un32FrameCount)) return(false);

  // ..... Copy and free old data
  memcpy(mt_ptASList[un32AnimationSetIndex].pun32FrameList, pun32Tmp,  un32Tot * sizeof(_vuint32));
  free(pun32Tmp);

  // ..... Set new datas
  for(un32Cnt = 0; un32Cnt < un32FrameCount; un32Cnt++)
    mt_ptASList[un32AnimationSetIndex].pun32FrameList[un32Tot + un32Cnt] = aun32FrameReferences[un32Cnt];
  
  // ..... OK !
  return(true);
}
_vuint32 _CVortexAImgControl::GetASFrameCount(_vuint32 un32AnimationSetIndex)
{
  return(mt_ptASList[un32AnimationSetIndex].un32TotalFrames);
}
_vvoid _CVortexAImgControl::SetASFrame(_vuint32 un32AnimationSetIndex, _vuint32 un32ASFrameIndex, _vuint32 un32FrameReference)
{
  mt_ptASList[un32AnimationSetIndex].pun32FrameList[un32ASFrameIndex] = un32FrameReference;
}
_vvoid _CVortexAImgControl::SetASFrames(_vuint32 un32AnimationSetIndex, _vuint32 aun32FrameReferences[])
{
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_ptASList[un32AnimationSetIndex].un32TotalFrames; un32Cnt++)
  {
    mt_ptASList[un32AnimationSetIndex].pun32FrameList[un32Cnt] = aun32FrameReferences[un32Cnt];
  }
}
_vbool _CVortexAImgControl::_AllocASFrame_(_vuint32 un32AnimationSetIndex, _vuint32 un32TotalASFrames)
{
  // ..... Allocate memory for frames
  if(un32TotalASFrames == 0) return(true);
  mt_ptASList[un32AnimationSetIndex].pun32FrameList = (_vuint32 *) malloc(un32TotalASFrames * sizeof(_vuint32));
  if(mt_ptASList[un32AnimationSetIndex].pun32FrameList == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for new frames");
    return(false);    
  }
  mt_ptASList[un32AnimationSetIndex].un32TotalFrames = un32TotalASFrames;
  return(true);
}

// ***** File I/O functions
_vbool _CVortexAImgControl::SaveVAC(const _vchar *pcFileName)
{
  pcFileName = pcFileName;
  VSetErrorCode(VERR_NotImplemented);
  VError(_F_, _L_, "This function is not implemented yet");
  return(false);
}
_vbool _CVortexAImgControl::LoadVAC(const _vchar *pcFileName)
{
  pcFileName = pcFileName;
  VSetErrorCode(VERR_NotImplemented);
  VError(_F_, _L_, "This function is not implemented yet");
  return(false);
}

// ============================================================================================
// End of vaimgctl.cpp
// ============================================================================================
