// ############################################################################################
//  vaimgctl.h
//  Definition & implementation  of _CVortexAImgControl class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VAIMGCTL_H_
#define _VAIMGCTL_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vimage.h>

// ============================================================================================
// Exported structures definition
// ============================================================================================
// ***** Animation set structure
struct _TAnimSet
{
  _vuint32 *pun32FrameList; // Frame references to _CVortexAnimatedImage
  _vuint32 un32TotalFrames; // Total frames for this animation set
};

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The advanced frame control for animated image class
class _CVortexAImgControl
{
  protected:
    // Variables
    _CVortexAnimatedImage *mt_pcsImgSrc;     // Reference to _CVortexAnimatedImage
    _vuint8               **mt_ppun8ImgBuff; // Reference to the image source buffer
    _TAnimSet             *mt_ptASList;      // Animaton set list
    _vuint32              mt_un32TotalAS;    // Total animation set
    _vuint32              mt_un32SelAS;      // Selected animation set
    _vuint32              mt_un32SelFrame;   // Selected frame of the selected animation set
    _vuint32              mt_un32StepDelay;  // The step delay
  
    // Frame management functions
    _vbool _AllocASFrame_(_vuint32 un32AnimationSetIndex, _vuint32 un32TotalASFrames);
    // Animation control functions
    inline _vvoid _UpdateImgSrcFrame_(_vvoid);
  
  public:
    // Constructor & destructor
    _CVortexAImgControl();
    ~_CVortexAImgControl();
    // Main initialization functions
    _vbool SetImageSource(_CVortexAnimatedImage *pcsAnimatedImage);
    _vvoid FreeAll(_vvoid);
    // Animation set management functions
    _vbool   AddAnimationSet(_vuint32 un32TotalASFrames = 0);
    _vbool   AddAnimationSets(_vuint32 un32ASCount, _vuint32 aun32TotalASFrames[] = NULL);
    _vuint32 GetAnimationSetCount(_vvoid);
    // Frame management functions
    _vbool   AddASFrame(_vuint32 un32AnimationSetIndex, _vuint32 un32FrameReference);
    _vbool   AddASFrames(_vuint32 un32AnimationSetIndex, _vuint32 un32FrameCount, _vuint32 aun32FrameReferences[]);
    _vuint32 GetASFrameCount(_vuint32 un32AnimationSetIndex);
    _vvoid   SetASFrame(_vuint32 un32AnimationSetIndex, _vuint32 un32ASFrameIndex, _vuint32 un32FrameReference);
    _vvoid   SetASFrames(_vuint32 un32AnimationSetIndex, _vuint32 aun32FrameReferences[]);
    // File I/O functions
    _vbool SaveVAC(const _vchar *pcFileName);
    _vbool LoadVAC(const _vchar *pcFileName);
    // Animation control functions
    inline _vvoid SelectAnimationSet(_vuint32 un32AnimationSetIndex);
    inline _vvoid SelectNextASFrame(_vbool bWarp = false);
    inline _vvoid SelectPrevASFrame(_vbool bWarp = false);
    inline _vvoid SelectNextASFrameDly(_vuint32 un32Delay, _vbool bWarp = false);
    inline _vvoid SelectPrevASFrameDly(_vuint32 un32Delay, _vbool bWarp = false);
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The advanced frame control for animated image class
// --------------------------------------------------------------------------------------------
// ***** Animation control functions
_vvoid _CVortexAImgControl::SelectAnimationSet(_vuint32 un32AnimationSetIndex)
{
  mt_un32SelAS     = un32AnimationSetIndex;
  mt_un32SelFrame  = 0;
  mt_un32StepDelay = 0;
  _UpdateImgSrcFrame_();
}
_vvoid _CVortexAImgControl::SelectNextASFrame(_vbool bWarp)
{
  if(mt_un32SelFrame < (mt_ptASList[mt_un32SelAS].un32TotalFrames - 1))
  {
    mt_un32SelFrame++;
    _UpdateImgSrcFrame_();
  }
  else if(bWarp)
  {
    mt_un32SelFrame = 0;
    _UpdateImgSrcFrame_();
  }
}
_vvoid _CVortexAImgControl::SelectPrevASFrame(_vbool bWarp)
{
  if(mt_un32SelFrame > 0)
  {
    mt_un32SelFrame--;
    _UpdateImgSrcFrame_();
  }
  else if(bWarp)
  {
    mt_un32SelFrame = mt_ptASList[mt_un32SelAS].un32TotalFrames - 1;
    _UpdateImgSrcFrame_();
  }
}
_vvoid _CVortexAImgControl::SelectNextASFrameDly(_vuint32 un32Delay, _vbool bWarp)
{
  // ..... Do delay
  mt_un32StepDelay++;
  if(mt_un32StepDelay < un32Delay) return;
  mt_un32StepDelay = 0;

  // ..... Do step
  SelectNextASFrame(bWarp);
}
_vvoid _CVortexAImgControl::SelectPrevASFrameDly(_vuint32 un32Delay, _vbool bWarp)
{
  // ..... Do delay
  mt_un32StepDelay++;
  if(mt_un32StepDelay < un32Delay) return;
  mt_un32StepDelay = 0;

  // ..... Do step
  SelectNextASFrame(bWarp);
}
_vvoid _CVortexAImgControl::_UpdateImgSrcFrame_(_vvoid)
{
  mt_pcsImgSrc->mt_pun8Buffer = mt_ppun8ImgBuff[mt_ptASList[mt_un32SelAS].pun32FrameList[mt_un32SelFrame]];
}

#endif
// ============================================================================================
// End of vaimgctl.h
// ============================================================================================
