// ############################################################################################
//  vcvtsize.cpp
//  This file will do compile time checking for the size of some data types, this will not
//  be included (archived) with the library.
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vbmp.h>
#include <vcwave.h>
#include <vvidbase.h>

// ============================================================================================
// Special defines
// ============================================================================================
// ***** Used to check the size of simple data types
#define _CHECK_SIZE_(Name, Size) typedef char _dummy##Name[int(sizeof(Name) == Size) - 1]

// ============================================================================================
// Check the size of some data types, just in case not compiled with GCC
// ============================================================================================
// ***** Create a dummy enumeration
typedef enum {_dummy_enum_value_} _vdenum_;

// ***** Check if all simple data types have the correct size
_CHECK_SIZE_(_vchar,  1); _CHECK_SIZE_(_vuchar,  1);
_CHECK_SIZE_(_vint8,  1); _CHECK_SIZE_(_vuint8,  1);
_CHECK_SIZE_(_vint16, 2); _CHECK_SIZE_(_vuint16, 2);
_CHECK_SIZE_(_vint32, 4); _CHECK_SIZE_(_vuint32, 4);
_CHECK_SIZE_(_vint64, 8); _CHECK_SIZE_(_vuint64, 8);
_CHECK_SIZE_(_vfloat, 4); _CHECK_SIZE_(_vdouble, 8);
_CHECK_SIZE_(_vbool,  1); _CHECK_SIZE_(_vdenum_, sizeof(_vint32));

// ***** Check if some packed structures have the correct size
//       (if so, the others should be also correct)
_CHECK_SIZE_(_TRGBTRIPLE,          3);
_CHECK_SIZE_(_TRGBQUAD ,           4);
_CHECK_SIZE_(_TBITMAPFILEHEADER,  14);
_CHECK_SIZE_(_TBITMAPCOREHEADER,  12);
_CHECK_SIZE_(_TBITMAPINFOHEADER,  40);
_CHECK_SIZE_(_TCWAVECHUNK,        12);
_CHECK_SIZE_(_TFORMATSUBCHUNKPCM, 24);
_CHECK_SIZE_(_TDATASUBCHUNK,       8);
_CHECK_SIZE_(_TCWFPCMHEADER,      44);
_CHECK_SIZE_(_TVortexUN8RGB,       3);

// ============================================================================================
// End of vcvtsize.cpp
// ============================================================================================
