// ############################################################################################
//  vbmp.h
//  Definition of structures used by Canonical WAVE file
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  Thanks to Craig Stuart Sapp (craig@ccrma.stanford.edu), for the Canonical WAVE file format.
//
//  NOTE : The WAVE file format is a subset of Microsoft's RIFF specification for the storage
//         of multimedia files.
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VCWAVE_H_
#define _VCWAVE_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vbtypes.h>

// ============================================================================================
// Exported structures definition
// ============================================================================================
// ***** Used by Canonical WAVE file
struct _TCWAVECHUNK
{
  _vuchar  aucChunkID[4];     // Always contains string "RIFF"
  _vuint32 un32ChunkSize;     // 4 + (8 + un32SubChunk1Size) + (8 + un32SubChunk2Size)
  _vuchar  aucFormat[4];      // Always contains string "WAVE"
} __attribute__((packed));
struct _TFORMATSUBCHUNKPCM
{
  _vuchar  aucSubChunk1ID[4]; // Always contains string "fmt "
  _vuint32 un32SubChunk1Size; // For PCM always 16
  _vuint16 un16AudioFormat;   // For PCM always 1
  _vuint16 un16Channels;      // 1 = mono, 2 = stereo
  _vuint32 un32SampleRate;    // The sampling frequency
  _vuint32 un32ByteRate;      // un32SampleRate * un16Channels * (un16BitsPerSample / 8)
  _vuint16 un16BlockAlign;    // un16Channels * (un16BitsPerSample / 8)
  _vuint16 un16BitsPerSample; // 8 = 8 bits, 16 = 16 bits
} __attribute__((packed));
struct _TDATASUBCHUNK
{
  _vuchar  aucSubChunk2ID[4]; // Always contains string "data"
  _vuint32 un32SubChunk2Size; // <number of samples> * un16Channels * (un16BitsPerSample / 8)
} __attribute__((packed));
struct _TCWFPCMHEADER
{
  _TCWAVECHUNK        tChunk;
  _TFORMATSUBCHUNKPCM tFormat;
  _TDATASUBCHUNK      tData;
} __attribute__((packed));

#endif
// ============================================================================================
// End of vcwave.h
// ============================================================================================
