// ############################################################################################
//  vfcline.h
//  Impelementation of fogged _CVortexFogUFCeilLine, _CVortexFogUGCeilLine, 
//  _CVortexFogTFCeilLine, and _CVortexFogTGCeilLine classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VFCLINE_H_
#define _VFCLINE_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <v3denv.h>
#include <vpoint2d.h>

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The fogged-untextured-flat-shaded ceil-line class
class _CVortexFogUFCeilLine
{
  protected:
    // Variables
    _vint32 mt_n32X1, mt_n32Y1, mt_n32X2, mt_n32Y2;
    _vint32 mt_n32X, mt_n32Y, mt_n32Z, mt_n32XStep, mt_n32ZStep;
    _vint32 mt_n32F, mt_n32FStep;
    _vint32 mt_n32EdgeHeight;
  public:
    // Constructor
    inline _CVortexFogUFCeilLine();
    // Ceil line functions
    inline _vvoid Init(_TVortexPoint2D tP1, _TVortexPoint2D tP2);
    inline _vvoid ClipTop(const _vint32 n32Top);
    inline _vvoid Step(_vvoid);
    inline _vvoid Step(_vint32 n32Amount);
    inline _vint32 GetX(_vvoid);
    inline _vint32 GetY(_vvoid);
    inline _vint32 GetZ(_vvoid);
    inline _vint32 GetF(_vvoid);
    inline _vint32 GetHeight(_vvoid);
};

// ***** The fogged-untextured-gouraud-shaded ceil-line class
class _CVortexFogUGCeilLine : public _CVortexFogUFCeilLine 
{
  protected:
    // Variables
    _vint32 mt_n32R, mt_n32RStep;
    _vint32 mt_n32G, mt_n32GStep;
    _vint32 mt_n32B, mt_n32BStep;
  public:
    // Constructor
    inline _CVortexFogUGCeilLine();
    // Ceil line functions
    inline _vvoid Init(_TVortexPoint2D tP1, _TVortexPoint2D tP2);
    inline _vvoid ClipTop(const _vint32 n32Top);
    inline _vvoid Step(_vvoid);
    inline _vvoid Step(_vint32 n32Amount);
    inline _vint32 GetR(_vvoid);
    inline _vint32 GetG(_vvoid);
    inline _vint32 GetB(_vvoid);
};

// ***** The fogged-textured-flat-shaded ceil-line class
class _CVortexFogTFCeilLine : public _CVortexFogUFCeilLine
{
  protected:
    // Variables
    _vint32 mt_n32U, mt_n32V, mt_n32UStep, mt_n32VStep;
  public:
    // Constructor
    inline _CVortexFogTFCeilLine();
    // Ceil line functions
    inline _vvoid Init(_TVortexPoint2D tP1, _TVortexPoint2D tP2);
    inline _vvoid ClipTop(const _vint32 n32Top);
    inline _vvoid Step(_vvoid);
    inline _vvoid Step(_vint32 n32Amount);
    inline _vint32 GetU(_vvoid);
    inline _vint32 GetV(_vvoid);
};

// ***** The fogged-textured-gouroud-shaded ceil-line class
class _CVortexFogTGCeilLine : public _CVortexFogUGCeilLine
{
  protected:
    // Variables
    _vint32 mt_n32U, mt_n32V, mt_n32UStep, mt_n32VStep;
  public:
    // Constructor
    inline _CVortexFogTGCeilLine();
    // Ceil line functions
    inline _vvoid Init(_TVortexPoint2D tP1, _TVortexPoint2D tP2);
    inline _vvoid ClipTop(const _vint32 n32Top);
    inline _vvoid Step(_vvoid);
    inline _vvoid Step(_vint32 n32Amount);
    inline _vint32 GetU(_vvoid);
    inline _vint32 GetV(_vvoid);
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The fogged-untextured-flat-shaded ceil-line class
// --------------------------------------------------------------------------------------------
// ***** Constructor
_CVortexFogUFCeilLine::_CVortexFogUFCeilLine()
{
  mt_n32EdgeHeight = 0;
}

// ***** Ceil-line functions
_vvoid _CVortexFogUFCeilLine::Init(_TVortexPoint2D tP1, _TVortexPoint2D tP2)
{
  // ..... Temporary variables
  _vint32 n32Z1, n32Z2;
  
  // ..... Calculate the initial values for polygon edge
  mt_n32X1 = tP1.n32X; mt_n32Y1 = tP1.n32Y; n32Z1 = tP1.n32Z;
  mt_n32X2 = tP2.n32X; mt_n32Y2 = tP2.n32Y; n32Z2 = tP2.n32Z;
  
  mt_n32EdgeHeight = mt_n32Y2 - mt_n32Y1;
  
  mt_n32X = (mt_n32X1 << _XSTEP_PREC_)  + _CEIL_FRACT_;
  mt_n32Y = mt_n32Y1;
  mt_n32Z = n32Z1 + _pcsAct3DEnv_->mt_tRB->n32ZTrans;
  mt_n32F = tP1.n32F;

  if(mt_n32EdgeHeight)
  {
    mt_n32XStep = ((mt_n32X2 - mt_n32X1) << _XSTEP_PREC_) / mt_n32EdgeHeight;
    mt_n32ZStep = (n32Z2 - n32Z1) / mt_n32EdgeHeight;
    mt_n32FStep = (tP2.n32F - tP1.n32F) / mt_n32EdgeHeight;
  }
  else
  {
    mt_n32XStep = 0;
    mt_n32ZStep = 0;
    mt_n32FStep = 0;
  }
}
_vvoid _CVortexFogUFCeilLine::ClipTop(const _vint32 n32Top)
{
  // ..... Temporary variables
  _vint32 n32Step, n32Height;
  _vfloat fXSlope;
  
  // ..... Clip the polygon edge againts the top of the viewport
  // ..... Must be called directly after calling FSInit()
  n32Height = mt_n32Y2 - mt_n32Y1;
  if((mt_n32Y < n32Top) && (n32Height))
  {
    n32Step = n32Top - mt_n32Y;
    fXSlope = _vfloat(mt_n32X2 - mt_n32X1) / _vfloat(n32Height);
    
    mt_n32X = (mt_n32X1 << _XSTEP_PREC_) +
                 (_vint32(fXSlope * _vfloat(n32Step)) << _XSTEP_PREC_) +
                 _CEIL_FRACT_;
    mt_n32Y = n32Top;
    mt_n32Z += (mt_n32ZStep * n32Step);
    mt_n32F += (mt_n32FStep * n32Step);
    mt_n32EdgeHeight -= n32Step;
  }
}
_vvoid _CVortexFogUFCeilLine::Step(_vvoid)
{
  mt_n32X += mt_n32XStep;
  mt_n32Y++;
  mt_n32Z += mt_n32ZStep;
  mt_n32F += mt_n32FStep;
  mt_n32EdgeHeight--;
}

_vvoid _CVortexFogUFCeilLine::Step(_vint32 n32Amount)
{
  mt_n32X += (mt_n32XStep * n32Amount);
  mt_n32Y += n32Amount;
  mt_n32Z += (mt_n32ZStep * n32Amount);
  mt_n32F += (mt_n32FStep * n32Amount);
  mt_n32EdgeHeight -= n32Amount;
}
_vint32 _CVortexFogUFCeilLine::GetX(_vvoid)
{
  return(mt_n32X >> _XSTEP_PREC_);
}
_vint32 _CVortexFogUFCeilLine::GetY(_vvoid)
{
  return(mt_n32Y);
}
_vint32 _CVortexFogUFCeilLine::GetZ(_vvoid)
{
  return(mt_n32Z);
}
_vint32 _CVortexFogUFCeilLine::GetF(_vvoid)
{
  return(mt_n32F);
}
_vint32 _CVortexFogUFCeilLine::GetHeight(_vvoid)
{
  return(mt_n32EdgeHeight);
}

// --------------------------------------------------------------------------------------------
// The fogged-untextured-gouraud-shaded ceil-line class
// --------------------------------------------------------------------------------------------
// ***** Constructor
_CVortexFogUGCeilLine::_CVortexFogUGCeilLine()
{
}

// ***** Ceil-line functions
_vvoid _CVortexFogUGCeilLine::Init(_TVortexPoint2D tP1, _TVortexPoint2D tP2)
{
  // ..... Temporary variables
  _vint32 n32Z1, n32Z2;
  
  // ..... Calculate the initial values for polygon edge
  mt_n32X1 = tP1.n32X; mt_n32Y1 = tP1.n32Y; n32Z1 = tP1.n32Z;
  mt_n32X2 = tP2.n32X; mt_n32Y2 = tP2.n32Y; n32Z2 = tP2.n32Z;
  
  mt_n32EdgeHeight = mt_n32Y2 - mt_n32Y1;
  
  mt_n32X = (mt_n32X1 << _XSTEP_PREC_)  + _CEIL_FRACT_;
  mt_n32Y = mt_n32Y1;
  mt_n32Z = n32Z1 + _pcsAct3DEnv_->mt_tRB->n32ZTrans;
  mt_n32F = tP1.n32F;
  mt_n32R = tP1.n32R;
  mt_n32G = tP1.n32G;
  mt_n32B = tP1.n32B;
  
  if(mt_n32EdgeHeight)
  {
    mt_n32XStep = ((mt_n32X2 - mt_n32X1) << _XSTEP_PREC_) / mt_n32EdgeHeight;
    mt_n32ZStep = (n32Z2 - n32Z1) / mt_n32EdgeHeight;
    mt_n32FStep = (tP2.n32F - tP1.n32F) / mt_n32EdgeHeight;
    mt_n32RStep = (tP2.n32R - tP1.n32R) / mt_n32EdgeHeight;
    mt_n32GStep = (tP2.n32G - tP1.n32G) / mt_n32EdgeHeight;
    mt_n32BStep = (tP2.n32B - tP1.n32B) / mt_n32EdgeHeight;
  }
  else
  {
    mt_n32XStep = 0;
    mt_n32ZStep = 0;
    mt_n32FStep = 0;
    mt_n32RStep = 0;
    mt_n32GStep = 0;
    mt_n32BStep = 0;
  }
}
_vvoid _CVortexFogUGCeilLine::ClipTop(const _vint32 n32Top)
{
  // ..... Temporary variables
  _vint32 n32Step, n32Height;
  _vfloat fXSlope;
  
  // ..... Clip the polygon edge againts the top of the viewport
  // ..... Must be called directly after calling GSInit()
  n32Height = mt_n32Y2 - mt_n32Y1;
  if((mt_n32Y < n32Top) && (n32Height))
  {
    n32Step = n32Top - mt_n32Y;
    fXSlope = _vfloat(mt_n32X2 - mt_n32X1) / _vfloat(n32Height);
    
    mt_n32X = (mt_n32X1 << _XSTEP_PREC_) +
                 (_vint32(fXSlope * _vfloat(n32Step)) << _XSTEP_PREC_) +
                 _CEIL_FRACT_;
    mt_n32Y = n32Top;
    mt_n32Z += (mt_n32ZStep * n32Step);
    mt_n32F += (mt_n32FStep * n32Step);
    mt_n32R += (mt_n32RStep * n32Step);
    mt_n32G += (mt_n32GStep * n32Step);
    mt_n32B += (mt_n32BStep * n32Step);
    mt_n32EdgeHeight -= n32Step;
  }
}
_vvoid _CVortexFogUGCeilLine::Step(_vvoid)
{
  mt_n32X += mt_n32XStep;
  mt_n32Y++;
  mt_n32Z += mt_n32ZStep;
  mt_n32F += mt_n32FStep;
  mt_n32R += mt_n32RStep;
  mt_n32G += mt_n32GStep;
  mt_n32B += mt_n32BStep;
  mt_n32EdgeHeight--;
}_vvoid _CVortexFogUGCeilLine::Step(_vint32 n32Amount)
{
  mt_n32X += (mt_n32XStep * n32Amount);
  mt_n32Y += n32Amount;
  mt_n32Z += (mt_n32ZStep * n32Amount);
  mt_n32F += (mt_n32FStep * n32Amount);
  mt_n32R += (mt_n32RStep * n32Amount);
  mt_n32G += (mt_n32GStep * n32Amount);
  mt_n32B += (mt_n32BStep * n32Amount);
  mt_n32EdgeHeight -= n32Amount;
}
_vint32 _CVortexFogUGCeilLine::GetR(_vvoid)
{
  return(mt_n32R);
}
_vint32 _CVortexFogUGCeilLine::GetG(_vvoid)
{
  return(mt_n32G);
}
_vint32 _CVortexFogUGCeilLine::GetB(_vvoid)
{
  return(mt_n32B);
}

// --------------------------------------------------------------------------------------------
// The fogged-textured-flat-shaded ceil-line class
// --------------------------------------------------------------------------------------------
// ***** Constructor
_CVortexFogTFCeilLine::_CVortexFogTFCeilLine()
{
}

// ***** Ceil-line functions
_vvoid _CVortexFogTFCeilLine::Init(_TVortexPoint2D tP1, _TVortexPoint2D tP2)
{
  // ..... Temporary variables
  _vint32 n32Z1, n32Z2;
  
  // ..... Calculate the initial values for polygon edge
  mt_n32X1 = tP1.n32X; mt_n32Y1 = tP1.n32Y; n32Z1 = tP1.n32Z;
  mt_n32X2 = tP2.n32X; mt_n32Y2 = tP2.n32Y; n32Z2 = tP2.n32Z;
  
  mt_n32EdgeHeight = mt_n32Y2 - mt_n32Y1;
  
  mt_n32X = (mt_n32X1 << _XSTEP_PREC_)  + _CEIL_FRACT_;
  mt_n32Y = mt_n32Y1;
  mt_n32Z = n32Z1 + _pcsAct3DEnv_->mt_tRB->n32ZTrans;
  mt_n32F = tP1.n32F;
  mt_n32U = tP1.n32U;
  mt_n32V = tP1.n32V;
  
  if(mt_n32EdgeHeight)
  {
    mt_n32XStep = ((mt_n32X2 - mt_n32X1) << _XSTEP_PREC_) / mt_n32EdgeHeight;
    mt_n32ZStep = (n32Z2 - n32Z1) / mt_n32EdgeHeight;
    mt_n32FStep = (tP2.n32F - tP1.n32F) / mt_n32EdgeHeight;
    mt_n32UStep = (tP2.n32U - tP1.n32U) / mt_n32EdgeHeight;
    mt_n32VStep = (tP2.n32V - tP1.n32V) / mt_n32EdgeHeight;
  }
  else
  {
    mt_n32XStep = 0;
    mt_n32ZStep = 0;
    mt_n32FStep = 0;
    mt_n32UStep = 0;
    mt_n32VStep = 0;
  }
}
_vvoid _CVortexFogTFCeilLine::ClipTop(const _vint32 n32Top)
{
  // ..... Temporary variables
  _vint32 n32Step, n32Height;
  _vfloat fXSlope;
  
  // ..... Clip the polygon edge againts the top of the viewport
  // ..... Must be called directly after calling FSInit()
  n32Height = mt_n32Y2 - mt_n32Y1;
  if((mt_n32Y < n32Top) && (n32Height))
  {
    n32Step = n32Top - mt_n32Y;
    fXSlope = _vfloat(mt_n32X2 - mt_n32X1) / _vfloat(n32Height);
    
    mt_n32X = (mt_n32X1 << _XSTEP_PREC_) +
                 (_vint32(fXSlope * _vfloat(n32Step)) << _XSTEP_PREC_) +
                 _CEIL_FRACT_;
    mt_n32Y = n32Top;
    mt_n32Z += (mt_n32ZStep * n32Step);
    mt_n32F += (mt_n32FStep * n32Step);
    mt_n32U += (mt_n32UStep * n32Step);
    mt_n32V += (mt_n32VStep * n32Step);
    mt_n32EdgeHeight -= n32Step;
  }
}
_vvoid _CVortexFogTFCeilLine::Step(_vvoid)
{
  mt_n32X += mt_n32XStep;
  mt_n32Y++;
  mt_n32Z += mt_n32ZStep;
  mt_n32F += mt_n32FStep;
  mt_n32U += mt_n32UStep;
  mt_n32V += mt_n32VStep;
  mt_n32EdgeHeight--;
}
_vvoid _CVortexFogTFCeilLine::Step(_vint32 n32Amount)
{
  mt_n32X += (mt_n32XStep * n32Amount);
  mt_n32Y += n32Amount;
  mt_n32Z += (mt_n32ZStep * n32Amount);
  mt_n32F += (mt_n32FStep * n32Amount);
  mt_n32U += (mt_n32UStep * n32Amount);
  mt_n32V += (mt_n32VStep * n32Amount);
  mt_n32EdgeHeight -= n32Amount;
}
_vint32 _CVortexFogTFCeilLine::GetU(_vvoid)
{
  return(mt_n32U);
}
_vint32 _CVortexFogTFCeilLine::GetV(_vvoid)
{
  return(mt_n32V);
}

// --------------------------------------------------------------------------------------------
// The fogged-textured-gouroud-shaded ceil-line class
// --------------------------------------------------------------------------------------------
// ***** Constructor
_CVortexFogTGCeilLine::_CVortexFogTGCeilLine()
{
}

// ***** Ceil-line functions
_vvoid _CVortexFogTGCeilLine::Init(_TVortexPoint2D tP1, _TVortexPoint2D tP2)
{
  // ..... Temporary variables
  _vint32 n32Z1, n32Z2;
  
  // ..... Calculate the initial values for polygon edge
  mt_n32X1 = tP1.n32X; mt_n32Y1 = tP1.n32Y; n32Z1 = tP1.n32Z;
  mt_n32X2 = tP2.n32X; mt_n32Y2 = tP2.n32Y; n32Z2 = tP2.n32Z;
  
  mt_n32EdgeHeight = mt_n32Y2 - mt_n32Y1;
  
  mt_n32X = (mt_n32X1 << _XSTEP_PREC_)  + _CEIL_FRACT_;
  mt_n32Y = mt_n32Y1;
  mt_n32Z = n32Z1 + _pcsAct3DEnv_->mt_tRB->n32ZTrans;
  mt_n32F = tP1.n32F;
  mt_n32U = tP1.n32U;
  mt_n32V = tP1.n32V;
  mt_n32R = tP1.n32R;
  mt_n32G = tP1.n32G;
  mt_n32B = tP1.n32B;
  
  if(mt_n32EdgeHeight)
  {
    mt_n32XStep = ((mt_n32X2 - mt_n32X1) << _XSTEP_PREC_) / mt_n32EdgeHeight;
    mt_n32ZStep = (n32Z2 - n32Z1) / mt_n32EdgeHeight;
    mt_n32FStep = (tP2.n32F - tP1.n32F) / mt_n32EdgeHeight;
    mt_n32UStep = (tP2.n32U - tP1.n32U) / mt_n32EdgeHeight;
    mt_n32VStep = (tP2.n32V - tP1.n32V) / mt_n32EdgeHeight;
    mt_n32RStep = (tP2.n32R - tP1.n32R) / mt_n32EdgeHeight;
    mt_n32GStep = (tP2.n32G - tP1.n32G) / mt_n32EdgeHeight;
    mt_n32BStep = (tP2.n32B - tP1.n32B) / mt_n32EdgeHeight;
  }
  else
  {
    mt_n32XStep = 0;
    mt_n32ZStep = 0;
    mt_n32FStep = 0;
    mt_n32UStep = 0;
    mt_n32VStep = 0;
    mt_n32RStep = 0;
    mt_n32GStep = 0;
    mt_n32BStep = 0;
  }
}
_vvoid _CVortexFogTGCeilLine::ClipTop(const _vint32 n32Top)
{
  // ..... Temporary variables
  _vint32 n32Step, n32Height;
  _vfloat fXSlope;
  
  // ..... Clip the polygon edge againts the top of the viewport
  // ..... Must be called directly after calling FSInit()
  n32Height = mt_n32Y2 - mt_n32Y1;
  if((mt_n32Y < n32Top) && (n32Height))
  {
    n32Step = n32Top - mt_n32Y;
    fXSlope = _vfloat(mt_n32X2 - mt_n32X1) / _vfloat(n32Height);
    
    mt_n32X = (mt_n32X1 << _XSTEP_PREC_) +
                 (_vint32(fXSlope * _vfloat(n32Step)) << _XSTEP_PREC_) +
                 _CEIL_FRACT_;
    mt_n32Y = n32Top;
    mt_n32Z += (mt_n32ZStep * n32Step);
    mt_n32F += (mt_n32FStep * n32Step);
    mt_n32U += (mt_n32UStep * n32Step);
    mt_n32V += (mt_n32VStep * n32Step);
    mt_n32R += (mt_n32RStep * n32Step);
    mt_n32G += (mt_n32GStep * n32Step);
    mt_n32B += (mt_n32BStep * n32Step);
    mt_n32EdgeHeight -= n32Step;
  }
}
_vvoid _CVortexFogTGCeilLine::Step(_vvoid)
{
  mt_n32X += mt_n32XStep;
  mt_n32Y++;
  mt_n32Z += mt_n32ZStep;
  mt_n32F += mt_n32FStep;
  mt_n32U += mt_n32UStep;
  mt_n32V += mt_n32VStep;
  mt_n32R += mt_n32RStep;
  mt_n32G += mt_n32GStep;
  mt_n32B += mt_n32BStep;
  mt_n32EdgeHeight--;
}

_vvoid _CVortexFogTGCeilLine::Step(_vint32 n32Amount)
{
  mt_n32X += (mt_n32XStep * n32Amount);
  mt_n32Y += n32Amount;
  mt_n32Z += (mt_n32ZStep * n32Amount);
  mt_n32F += (mt_n32FStep * n32Amount);
  mt_n32U += (mt_n32UStep * n32Amount);
  mt_n32V += (mt_n32VStep * n32Amount);
  mt_n32R += (mt_n32RStep * n32Amount);
  mt_n32G += (mt_n32GStep * n32Amount);
  mt_n32B += (mt_n32BStep * n32Amount);
  mt_n32EdgeHeight -= n32Amount;
}

_vint32 _CVortexFogTGCeilLine::GetU(_vvoid)
{
  return(mt_n32U);
}
_vint32 _CVortexFogTGCeilLine::GetV(_vvoid)
{
  return(mt_n32V);
}

#endif
// ============================================================================================
// End of vfcline.h
// ============================================================================================
