// ############################################################################################
//  vfprojs.cpp
//  Implementation of fogged projection functions of _CVortexPanel3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vpanel3d.h>

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 3d panel class
// --------------------------------------------------------------------------------------------
// ***** Fogged, untextured/uncolored, flat shaded
_vvoid _CVortexPanel3D::FogUFProject(_vvoid)
{
  // ..... Temporary variables
  _vuint8         un8Cnt, un8Start, un8End, un8OutCnt = 0;
  _vfloat         fOneOverZ, fDeltaZ, fT;
  _CVortexPoint3D csZClip[5];
  _vint32         n32Fog;
  
  // ..... Initialization
  mt_tP.un8SPCount = 4;
  un8Start        = 3;
  
  // ..... Loop trough all edges of panel using S&H algorithm
  for(un8End = 0; un8End < mt_tP.un8SPCount; un8End++)
  {
    if(mt_tP.csVPoint[un8Start]->m_tWorld.fZ >= _pcsAct3DEnv_->mt_un32MinZ)
    {
      if(mt_tP.csVPoint[un8End]->m_tWorld.fZ >= _pcsAct3DEnv_->mt_un32MinZ)
      {
        // Entirely inside front view volume
        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8End]->m_tWorld.fX;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8End]->m_tWorld.fY;
        csZClip[un8OutCnt].m_tWorld.fZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ;
        un8OutCnt++;
      }
      else
      {
        // Leaving view volume
        fDeltaZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ;
        fT = _vfloat(_pcsAct3DEnv_->mt_un32MinZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ) / fDeltaZ;

        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8Start]->m_tWorld.fX +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fX - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fX) * fT;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8Start]->m_tWorld.fY +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fY - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fY) * fT;
        csZClip[un8OutCnt].m_tWorld.fZ = _pcsAct3DEnv_->mt_un32MinZ; 
        un8OutCnt++;
      }
    }
    else
    {
      if(mt_tP.csVPoint[un8End]->m_tWorld.fZ >= _pcsAct3DEnv_->mt_un32MinZ)
      {
        // Entering view volume
        fDeltaZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ;
        fT = _vfloat(_pcsAct3DEnv_->mt_un32MinZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ) / fDeltaZ;

        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8Start]->m_tWorld.fX +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fX - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fX) * fT;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8Start]->m_tWorld.fY +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fY - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fY) * fT;
        csZClip[un8OutCnt].m_tWorld.fZ = _pcsAct3DEnv_->mt_un32MinZ; 
        un8OutCnt++;
        
        // Add an extra edge to list
        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8End]->m_tWorld.fX;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8End]->m_tWorld.fY;
        csZClip[un8OutCnt].m_tWorld.fZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ;
        un8OutCnt++;
      }
      else
      {
        // Entirely outside clipping boundary
      }
    }
    // Advance to next vertex
    un8Start = un8End;
  }
  
  // ..... Store the number of vertices from un8OutCnt
  mt_tP.un8SPCount = un8OutCnt;
  
  // ..... Project panel's points
  for(un8Cnt = 0; un8Cnt < un8OutCnt; un8Cnt++)
  {
    fOneOverZ = _vfloat(1.0) / csZClip[un8Cnt].m_tWorld.fZ;
    mt_tP.tSPoint[un8Cnt].n32X = _vuint32(csZClip[un8Cnt].m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale * fOneOverZ +
                                          _pcsAct3DEnv_->mt_fXShift);
    mt_tP.tSPoint[un8Cnt].n32Y = _vuint32(csZClip[un8Cnt].m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale * fOneOverZ +
                                          _pcsAct3DEnv_->mt_fYShift);
    mt_tP.tSPoint[un8Cnt].n32Z = _vuint32(fOneOverZ * _vfloat(1 << _ZSTEP_PREC_));

    n32Fog = _vint32(csZClip[un8Cnt].m_tWorld.fZ - _pcsAct3DEnv_-> mt_un32MinZ) >> _pcsAct3DEnv_->mt_un32FogShift;
    if(n32Fog < 0)
      mt_tP.tSPoint[un8Cnt].n32F = 0;
    else if(n32Fog > MAX_COLOR_VALUE)
      mt_tP.tSPoint[un8Cnt].n32F = MAX_COLOR_VALUE << _ISTEP_PREC_;
    else
      mt_tP.tSPoint[un8Cnt].n32F = n32Fog << _ISTEP_PREC_;
  }
}

// ***** Fogged, untextured/uncolored, gouroud shaded
_vvoid _CVortexPanel3D::FogUGProject(_vvoid)
{
  // ..... Temporary variables
  _vuint8         un8Cnt, un8Start, un8End, un8OutCnt = 0;
  _vfloat         fOneOverZ, fDeltaZ, fT;
  _CVortexPoint3D csZClip[5];
  _TVortexUN8RGB  atClip[5];
  _vint32         n32Fog;
  
  // ..... Initialization
  mt_tP.un8SPCount = 4;
  un8Start         = 3;
  
  // ..... Loop trough all edges of panel using S&H algorithm
  for(un8End = 0; un8End < mt_tP.un8SPCount; un8End++)
  {
    if(mt_tP.csVPoint[un8Start]->m_tWorld.fZ >= _pcsAct3DEnv_->mt_un32MinZ)
    {
      if(mt_tP.csVPoint[un8End]->m_tWorld.fZ >= _pcsAct3DEnv_->mt_un32MinZ)
      {
        // Entirely inside front view volume
        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8End]->m_tWorld.fX;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8End]->m_tWorld.fY;
        csZClip[un8OutCnt].m_tWorld.fZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ;
        atClip[un8OutCnt].un8R = mt_tP.atTotal[un8End].un8R;
        atClip[un8OutCnt].un8G = mt_tP.atTotal[un8End].un8G;
        atClip[un8OutCnt].un8B = mt_tP.atTotal[un8End].un8B;
        un8OutCnt++;
      }
      else
      {
        // Leaving view volume
        fDeltaZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ;
        fT = _vfloat(_pcsAct3DEnv_->mt_un32MinZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ) / fDeltaZ;

        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8Start]->m_tWorld.fX +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fX - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fX) * fT;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8Start]->m_tWorld.fY +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fY - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fY) * fT;
        csZClip[un8OutCnt].m_tWorld.fZ = _pcsAct3DEnv_->mt_un32MinZ; 
        atClip[un8OutCnt].un8R = mt_tP.atTotal[un8Start].un8R +
                                 _vuint8(_vfloat(mt_tP.atTotal[un8End].un8R - mt_tP.atTotal[un8Start].un8R) * fT);
        atClip[un8OutCnt].un8G = mt_tP.atTotal[un8Start].un8G +
                                 _vuint8(_vfloat(mt_tP.atTotal[un8End].un8G - mt_tP.atTotal[un8Start].un8G) * fT);
        atClip[un8OutCnt].un8B = mt_tP.atTotal[un8Start].un8B +
                                 _vuint8(_vfloat(mt_tP.atTotal[un8End].un8B - mt_tP.atTotal[un8Start].un8B) * fT);
        un8OutCnt++;
      }
    }
    else
    {
      if(mt_tP.csVPoint[un8End]->m_tWorld.fZ >= _pcsAct3DEnv_->mt_un32MinZ)
      {
        // Entering view volume
        fDeltaZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ;
        fT = _vfloat(_pcsAct3DEnv_->mt_un32MinZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ) / fDeltaZ;

        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8Start]->m_tWorld.fX +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fX - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fX) * fT;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8Start]->m_tWorld.fY +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fY - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fY) * fT;
        csZClip[un8OutCnt].m_tWorld.fZ = _pcsAct3DEnv_->mt_un32MinZ; 
        atClip[un8OutCnt].un8R = mt_tP.atTotal[un8Start].un8R +
                                 _vuint8(_vfloat(mt_tP.atTotal[un8End].un8R - mt_tP.atTotal[un8Start].un8R) * fT);
        atClip[un8OutCnt].un8G = mt_tP.atTotal[un8Start].un8G +
                                 _vuint8(_vfloat(mt_tP.atTotal[un8End].un8G - mt_tP.atTotal[un8Start].un8G) * fT);
        atClip[un8OutCnt].un8B = mt_tP.atTotal[un8Start].un8B +
                                 _vuint8(_vfloat(mt_tP.atTotal[un8End].un8B - mt_tP.atTotal[un8Start].un8B) * fT);
        un8OutCnt++;
        
        // Add an extra edge to list
        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8End]->m_tWorld.fX;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8End]->m_tWorld.fY;
        csZClip[un8OutCnt].m_tWorld.fZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ;
        atClip[un8OutCnt].un8R = mt_tP.atTotal[un8End].un8R;
        atClip[un8OutCnt].un8G = mt_tP.atTotal[un8End].un8G;
        atClip[un8OutCnt].un8B = mt_tP.atTotal[un8End].un8B;
        un8OutCnt++;
      }
      else
      {
        // Entirely outside clipping boundary
      }
    }
    // Advance to next vertex
    un8Start = un8End;
  }
  
  // ..... Store the number of vertices from un8OutCnt
  mt_tP.un8SPCount = un8OutCnt;
  
  // ..... Project panel's points
  for(un8Cnt = 0; un8Cnt < un8OutCnt; un8Cnt++)
  {
    fOneOverZ = _vfloat(1.0) / csZClip[un8Cnt].m_tWorld.fZ;
    mt_tP.tSPoint[un8Cnt].n32X = _vuint32(csZClip[un8Cnt].m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale * fOneOverZ +
                                          _pcsAct3DEnv_->mt_fXShift);
    mt_tP.tSPoint[un8Cnt].n32Y = _vuint32(csZClip[un8Cnt].m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale * fOneOverZ +
                                          _pcsAct3DEnv_->mt_fYShift);
    mt_tP.tSPoint[un8Cnt].n32Z = _vuint32(fOneOverZ * _vfloat(1 << _ZSTEP_PREC_));
    mt_tP.tSPoint[un8Cnt].n32R = _vint32(atClip[un8Cnt].un8R) << _ISTEP_PREC_;
    mt_tP.tSPoint[un8Cnt].n32G = _vint32(atClip[un8Cnt].un8G) << _ISTEP_PREC_;
    mt_tP.tSPoint[un8Cnt].n32B = _vint32(atClip[un8Cnt].un8B) << _ISTEP_PREC_;

    n32Fog = _vint32(csZClip[un8Cnt].m_tWorld.fZ - _pcsAct3DEnv_-> mt_un32MinZ) >> _pcsAct3DEnv_->mt_un32FogShift;
    if(n32Fog < 0)
      mt_tP.tSPoint[un8Cnt].n32F = 0;
    else if(n32Fog > MAX_COLOR_VALUE)
      mt_tP.tSPoint[un8Cnt].n32F = MAX_COLOR_VALUE << _ISTEP_PREC_;
    else
      mt_tP.tSPoint[un8Cnt].n32F = n32Fog << _ISTEP_PREC_;
  }
}

// ***** Fogged, textured, not/flat shaded
_vvoid _CVortexPanel3D::FogTFProject(_vvoid)
{
  // ..... Temporary variables
  _vuint8         un8Cnt, un8Start, un8End, un8OutCnt = 0;
  _vfloat         fOneOverZ, fDeltaZ, fT;
  _CVortexPoint3D csZClip[5];
  _TVortexTexUV   tNewUV[5];
  _vint32         n32Fog;
  
  // ..... Initialization
  mt_tP.un8SPCount = 4;
  un8Start        = 3;
  
  // ..... Loop trough all edges of panel using S&H algorithm
  for(un8End = 0; un8End < mt_tP.un8SPCount; un8End++)
  {
    if(mt_tP.csVPoint[un8Start]->m_tWorld.fZ >= _pcsAct3DEnv_->mt_un32MinZ)
    {
      if(mt_tP.csVPoint[un8End]->m_tWorld.fZ >= _pcsAct3DEnv_->mt_un32MinZ)
      {
        // Entirely inside front view volume
        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8End]->m_tWorld.fX;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8End]->m_tWorld.fY;
        csZClip[un8OutCnt].m_tWorld.fZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ;
        tNewUV[un8OutCnt].m_n32U = mt_tP.tUV[un8End].m_n32U;
        tNewUV[un8OutCnt].m_n32V = mt_tP.tUV[un8End].m_n32V;
        un8OutCnt++;
      }
      else
      {
        // Leaving view volume
        fDeltaZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ;
        fT = _vfloat(_pcsAct3DEnv_->mt_un32MinZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ) / fDeltaZ;

        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8Start]->m_tWorld.fX +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fX - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fX) * fT;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8Start]->m_tWorld.fY +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fY - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fY) * fT;
        csZClip[un8OutCnt].m_tWorld.fZ = _pcsAct3DEnv_->mt_un32MinZ;
        tNewUV[un8OutCnt].m_n32U = mt_tP.tUV[un8Start].m_n32U + 
                                   _vint32(_vfloat(mt_tP.tUV[un8End].m_n32U -
                                                   mt_tP.tUV[un8Start].m_n32U) * fT);
        tNewUV[un8OutCnt].m_n32V = mt_tP.tUV[un8Start].m_n32V + 
                                   _vint32(_vfloat(mt_tP.tUV[un8End].m_n32V -
                                                   mt_tP.tUV[un8Start].m_n32V) * fT);
        un8OutCnt++;
      }
    }
    else
    {
      if(mt_tP.csVPoint[un8End]->m_tWorld.fZ >= _pcsAct3DEnv_->mt_un32MinZ)
      {
        // Entering view volume
        fDeltaZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ;
        fT = _vfloat(_pcsAct3DEnv_->mt_un32MinZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ) / fDeltaZ;

        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8Start]->m_tWorld.fX +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fX - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fX) * fT;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8Start]->m_tWorld.fY +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fY - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fY) * fT;
        csZClip[un8OutCnt].m_tWorld.fZ = _pcsAct3DEnv_->mt_un32MinZ; 
        tNewUV[un8OutCnt].m_n32U = mt_tP.tUV[un8Start].m_n32U + 
                                   _vint32(_vfloat(mt_tP.tUV[un8End].m_n32U -
                                                   mt_tP.tUV[un8Start].m_n32U) * fT);
        tNewUV[un8OutCnt].m_n32V = mt_tP.tUV[un8Start].m_n32V + 
                                   _vint32(_vfloat(mt_tP.tUV[un8End].m_n32V -
                                                   mt_tP.tUV[un8Start].m_n32V) * fT);
        un8OutCnt++;
        
        // Add an extra edge to list
        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8End]->m_tWorld.fX;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8End]->m_tWorld.fY;
        csZClip[un8OutCnt].m_tWorld.fZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ;
        tNewUV[un8OutCnt].m_n32U = mt_tP.tUV[un8End].m_n32U;
        tNewUV[un8OutCnt].m_n32V = mt_tP.tUV[un8End].m_n32V;
        un8OutCnt++;
      }
      else
      {
        // Entirely outside clipping boundary
      }
    }
    // Advance to next vertex
    un8Start = un8End;
  }
  
  // ..... Store the number of vertices from un8OutCnt
  mt_tP.un8SPCount = un8OutCnt;
  
  // ..... Project panel's points
  for(un8Cnt = 0; un8Cnt < un8OutCnt; un8Cnt++)
  {
    fOneOverZ = _vfloat(1.0) / csZClip[un8Cnt].m_tWorld.fZ;
    mt_tP.tSPoint[un8Cnt].n32X = _vuint32(csZClip[un8Cnt].m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale * fOneOverZ +
                                          _pcsAct3DEnv_->mt_fXShift);
    mt_tP.tSPoint[un8Cnt].n32Y = _vuint32(csZClip[un8Cnt].m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale * fOneOverZ +
                                          _pcsAct3DEnv_->mt_fYShift);
    mt_tP.tSPoint[un8Cnt].n32Z = _vuint32(fOneOverZ * _vfloat(1 << _ZSTEP_PREC_));
    mt_tP.tSPoint[un8Cnt].n32U = _vuint32(_vfloat(tNewUV[un8Cnt].m_n32U) * fOneOverZ *
                                          _vfloat(1 << _TSTEP_PREC_));
    mt_tP.tSPoint[un8Cnt].n32V = _vuint32(_vfloat(tNewUV[un8Cnt].m_n32V) * fOneOverZ *
                                          _vfloat(1 << _TSTEP_PREC_));

    n32Fog = _vint32(csZClip[un8Cnt].m_tWorld.fZ - _pcsAct3DEnv_-> mt_un32MinZ) >> _pcsAct3DEnv_->mt_un32FogShift;
    if(n32Fog < 0)
      mt_tP.tSPoint[un8Cnt].n32F = 0;
    else if(n32Fog > MAX_COLOR_VALUE)
      mt_tP.tSPoint[un8Cnt].n32F = MAX_COLOR_VALUE << _ISTEP_PREC_;
    else
      mt_tP.tSPoint[un8Cnt].n32F = n32Fog << _ISTEP_PREC_;
  }
}

// ***** Fogged, textured, gouroud shaded
_vvoid _CVortexPanel3D::FogTGProject(_vvoid)
{
  // ..... Temporary variables
  _vuint8         un8Cnt, un8Start, un8End, un8OutCnt = 0;
  _vfloat         fOneOverZ, fDeltaZ, fT;
  _CVortexPoint3D csZClip[5];
  _TVortexTexUV   tNewUV[5];
  _TVortexUN8RGB  atClip[5];  
  _vint32         n32Fog;
  
  // ..... Initialization
  mt_tP.un8SPCount = 4;
  un8Start        = 3;
  
  // ..... Loop trough all edges of panel using S&H algorithm
  for(un8End = 0; un8End < mt_tP.un8SPCount; un8End++)
  {
    if(mt_tP.csVPoint[un8Start]->m_tWorld.fZ >= _pcsAct3DEnv_->mt_un32MinZ)
    {
      if(mt_tP.csVPoint[un8End]->m_tWorld.fZ >= _pcsAct3DEnv_->mt_un32MinZ)
      {
        // Entirely inside front view volume
        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8End]->m_tWorld.fX;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8End]->m_tWorld.fY;
        csZClip[un8OutCnt].m_tWorld.fZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ;
        tNewUV[un8OutCnt].m_n32U = mt_tP.tUV[un8End].m_n32U;
        tNewUV[un8OutCnt].m_n32V = mt_tP.tUV[un8End].m_n32V;
        atClip[un8OutCnt].un8R = mt_tP.atTotal[un8End].un8R;
        atClip[un8OutCnt].un8G = mt_tP.atTotal[un8End].un8G;
        atClip[un8OutCnt].un8B = mt_tP.atTotal[un8End].un8B;
        un8OutCnt++;
      }
      else
      {
        // Leaving view volume
        fDeltaZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ;
        fT = _vfloat(_pcsAct3DEnv_->mt_un32MinZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ) / fDeltaZ;

        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8Start]->m_tWorld.fX +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fX - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fX) * fT;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8Start]->m_tWorld.fY +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fY - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fY) * fT;
        csZClip[un8OutCnt].m_tWorld.fZ = _pcsAct3DEnv_->mt_un32MinZ;
        tNewUV[un8OutCnt].m_n32U = mt_tP.tUV[un8Start].m_n32U + 
                                   _vint32(_vfloat(mt_tP.tUV[un8End].m_n32U -
                                                   mt_tP.tUV[un8Start].m_n32U) * fT);
        tNewUV[un8OutCnt].m_n32V = mt_tP.tUV[un8Start].m_n32V + 
                                   _vint32(_vfloat(mt_tP.tUV[un8End].m_n32V -
                                                   mt_tP.tUV[un8Start].m_n32V) * fT);
        atClip[un8OutCnt].un8R = mt_tP.atTotal[un8Start].un8R +
                                 _vuint8(_vfloat(mt_tP.atTotal[un8End].un8R - mt_tP.atTotal[un8Start].un8R) * fT);
        atClip[un8OutCnt].un8G = mt_tP.atTotal[un8Start].un8G +
                                 _vuint8(_vfloat(mt_tP.atTotal[un8End].un8G - mt_tP.atTotal[un8Start].un8G) * fT);
        atClip[un8OutCnt].un8B = mt_tP.atTotal[un8Start].un8B +
                                 _vuint8(_vfloat(mt_tP.atTotal[un8End].un8B - mt_tP.atTotal[un8Start].un8B) * fT);
        un8OutCnt++;
      }
    }
    else
    {
      if(mt_tP.csVPoint[un8End]->m_tWorld.fZ >= _pcsAct3DEnv_->mt_un32MinZ)
      {
        // Entering view volume
        fDeltaZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ;
        fT = _vfloat(_pcsAct3DEnv_->mt_un32MinZ - mt_tP.csVPoint[un8Start]->m_tWorld.fZ) / fDeltaZ;

        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8Start]->m_tWorld.fX +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fX - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fX) * fT;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8Start]->m_tWorld.fY +
                                         (mt_tP.csVPoint[un8End]->m_tWorld.fY - 
                                          mt_tP.csVPoint[un8Start]->m_tWorld.fY) * fT;
        csZClip[un8OutCnt].m_tWorld.fZ = _pcsAct3DEnv_->mt_un32MinZ; 
        tNewUV[un8OutCnt].m_n32U = mt_tP.tUV[un8Start].m_n32U + 
                                   _vint32(_vfloat(mt_tP.tUV[un8End].m_n32U -
                                                   mt_tP.tUV[un8Start].m_n32U) * fT);
        tNewUV[un8OutCnt].m_n32V = mt_tP.tUV[un8Start].m_n32V + 
                                   _vint32(_vfloat(mt_tP.tUV[un8End].m_n32V -
                                                   mt_tP.tUV[un8Start].m_n32V) * fT);
        atClip[un8OutCnt].un8R = mt_tP.atTotal[un8Start].un8R +
                                 _vuint8(_vfloat(mt_tP.atTotal[un8End].un8R - mt_tP.atTotal[un8Start].un8R) * fT);
        atClip[un8OutCnt].un8G = mt_tP.atTotal[un8Start].un8G +
                                 _vuint8(_vfloat(mt_tP.atTotal[un8End].un8G - mt_tP.atTotal[un8Start].un8G) * fT);
        atClip[un8OutCnt].un8B = mt_tP.atTotal[un8Start].un8B +
                                 _vuint8(_vfloat(mt_tP.atTotal[un8End].un8B - mt_tP.atTotal[un8Start].un8B) * fT);
        un8OutCnt++;
        
        // Add an extra edge to list
        csZClip[un8OutCnt].m_tWorld.fX = mt_tP.csVPoint[un8End]->m_tWorld.fX;
        csZClip[un8OutCnt].m_tWorld.fY = mt_tP.csVPoint[un8End]->m_tWorld.fY;
        csZClip[un8OutCnt].m_tWorld.fZ = mt_tP.csVPoint[un8End]->m_tWorld.fZ;
        tNewUV[un8OutCnt].m_n32U = mt_tP.tUV[un8End].m_n32U;
        tNewUV[un8OutCnt].m_n32V = mt_tP.tUV[un8End].m_n32V;
        atClip[un8OutCnt].un8R = mt_tP.atTotal[un8End].un8R;
        atClip[un8OutCnt].un8G = mt_tP.atTotal[un8End].un8G;
        atClip[un8OutCnt].un8B = mt_tP.atTotal[un8End].un8B;
        un8OutCnt++;
      }
      else
      {
        // Entirely outside clipping boundary
      }
    }
    // Advance to next vertex
    un8Start = un8End;
  }
  
  // ..... Store the number of vertices from un8OutCnt
  mt_tP.un8SPCount = un8OutCnt;
  
  // ..... Project panel's points
  for(un8Cnt = 0; un8Cnt < un8OutCnt; un8Cnt++)
  {
    fOneOverZ = _vfloat(1.0) / csZClip[un8Cnt].m_tWorld.fZ;
    mt_tP.tSPoint[un8Cnt].n32X = _vuint32(csZClip[un8Cnt].m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale * fOneOverZ +
                                          _pcsAct3DEnv_->mt_fXShift);
    mt_tP.tSPoint[un8Cnt].n32Y = _vuint32(csZClip[un8Cnt].m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale * fOneOverZ +
                                          _pcsAct3DEnv_->mt_fYShift);
    mt_tP.tSPoint[un8Cnt].n32Z = _vuint32(fOneOverZ * _vfloat(1 << _ZSTEP_PREC_));
    mt_tP.tSPoint[un8Cnt].n32U = _vuint32(_vfloat(tNewUV[un8Cnt].m_n32U) * fOneOverZ *
                                          _vfloat(1 << _TSTEP_PREC_));
    mt_tP.tSPoint[un8Cnt].n32V = _vuint32(_vfloat(tNewUV[un8Cnt].m_n32V) * fOneOverZ *
                                          _vfloat(1 << _TSTEP_PREC_));
    mt_tP.tSPoint[un8Cnt].n32R = _vint32(atClip[un8Cnt].un8R) << _ISTEP_PREC_;
    mt_tP.tSPoint[un8Cnt].n32G = _vint32(atClip[un8Cnt].un8G) << _ISTEP_PREC_;
    mt_tP.tSPoint[un8Cnt].n32B = _vint32(atClip[un8Cnt].un8B) << _ISTEP_PREC_;  
    
    n32Fog = _vint32(csZClip[un8Cnt].m_tWorld.fZ - _pcsAct3DEnv_-> mt_un32MinZ) >> _pcsAct3DEnv_->mt_un32FogShift;
    if(n32Fog < 0)
      mt_tP.tSPoint[un8Cnt].n32F = 0;
    else if(n32Fog > MAX_COLOR_VALUE)
      mt_tP.tSPoint[un8Cnt].n32F = MAX_COLOR_VALUE << _ISTEP_PREC_;
    else
      mt_tP.tSPoint[un8Cnt].n32F = n32Fog << _ISTEP_PREC_;
  }
}

// ============================================================================================
// End of vfprojs.cpp
// ============================================================================================
