// ############################################################################################
//  vx11.h
//  Definition & implementation of _CVortexVideoEnv class using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VX11_H_
#define _VX11_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <string.h>
#include <sys/select.h>
// ***** VortexGE headers
#include <vkbsym.h>
#include <vfastmem.h>
#include <vvidbase.h>
#include <vidx11/vx11dlg.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** Default fonts
//       If these failed, try to change "iso8859" into "*" or change "times" to to "*".
//       Another font also can be choosed using xfontsel from terminal
#define _DEFAULT_FONT_SMALL_ "*-times-*-r-*-*-12-*-*-*-*-*-iso8859-*" 
#define _DEFAULT_FONT_BIG_   "*-times-*-r-*-*-24-*-*-*-*-*-iso8859-*" 

// ***** Maximum values
#define _MAX_X11_VKS_ 256
#define _MAX_X11_VBS_ 64
#define _MAX_EVENTS_  32

// ============================================================================================
// Reference to External variables
// ============================================================================================
// ***** The VortexGE symbols
extern _EVortexKeySymbol _aeX11VKS_[_MAX_X11_VKS_];
extern _EVortexBtnSymbol _aeX11VBS_[_MAX_X11_VBS_];

// ============================================================================================
// Enumerations
// ============================================================================================
// ***** Message box size constants
typedef enum _EVMS_ { ESizeTiny, ESizeSmall, ESizeMedium, 
                      ESizeLarge, ESizeHuge } _EVortexMsgSize;

                      // ***** Message box icon constants
typedef enum _EVMI_ { EIconInformation, EIconQuestion, 
                      EIconWarning, EIconError } _EVortexMsgIcon;
                      
// ***** Message box button constants
typedef enum _EVMB_ { EBtnOK, EBtnOKCancel, EBtnYesNo, EBtnYesNoCancel,
                      EBtnAbortRetry, EBtnAbortRetryIgnore }_EVortexMsgButton;

                      // ***** Message box return value constants
typedef enum _EVMRV_ { EOK, ECancel, EYes, ENo,
                       EAbort, ERetry, EIgnore  } _EVortexMsgRetVal;

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The X11 display class
class _CVortexVideoEnv
{
  protected:
    // Variables
    struct _TX11Base
    {
      Display     *ptDisplay;         // Connection to the X server
      _vint32     n32Screen;          // Default screen number
      _vuint32    un32Depth;          // Display depth of screen
      Visual      *ptVisual;          // Default visual type of screen
      Window      tRootWindow;        // The root window
      Colormap    tColormap;          // Colormap of the main window
      struct _TColormap
      {
        _vuint8  un8Count;            // Total number of user colors
        XColor   *ptColors;           // The user color table
      } tColormapData;
      _vuint32    un32BlackPixel ;    // Black pixel value of screen
      _vuint32    un32WhitePixel ;    // White pixel value of scren
      GC          tMainGC;            // The main GC
      XFontStruct *ptDefFontSmall;    // The default small font
      XFontStruct *ptDefFontBig;      // The default big font
      _vuint32    un32Width;          // Width of the main window
      _vuint32    un32Height;         // Height of the main window
      Window      tMainWindow;        // The main window
      Pixmap      tMainPixmap;        // The main window's pixmap
      Cursor      tDefCurArrow;       // The default "arrow" cursor
      Cursor      tDefCurWait;        // The default "wait" cursor
      _vbool      bAutoRepeat;        // The keyboard auto-repeat flag
      _vbool      bInitialized;       // Indicate that the main window is initialized
    } mt_tX11; 
    struct _TX11CurentlyActiveGCSettings
    {
      XFontStruct *ptFont;            // Current active font
      _vuint32    un32Foreground;     // Current active foreground color
      _vint32     n32FillStyle;       // Current active fill style
    } mt_tX11CS;
    _TVortexStyle mt_tX11Style;       // Currently used style
    struct _TRenderImage
    {
      _vuint32  un32Width;            // Width of the image
      _vuint32  un32Height;           // Height of the image
      _vuint32  un32Dim;              // Dimension of the image data buffer
      _vuint32  un32Size;             // Size of the image data buffer
      _vuchar  *pucImage;             // The image data
      XImage   *ptImage;              // The image
      _vuint16 *pun16Pal;             // Palette table of the screen
      _vuchar  *pucLUT;               // Color translation LUT
      _vbool   bInitialized;          // Indicate that the render image is initialized
    } mt_tRI;
    _TVortexRenderBuffer _sse mt_tRB;      // The render buffer  
    _vbool mt_bAllInitialized;        // Indicate that this class in completely initialized
    _vbool mt_bActive;                // Indicate that the main window has focus
    _vbool mt_bVisible;               // Indicate that the main window is visible
    struct _TX11EventBuffer
    {
      XEvent  tEmptyEvent;            // An empty event
      XEvent  atEvents[_MAX_EVENTS_]; // Events container
      XEvent  *ptEvent;               // Pointer to an event
      _vuint8 un8DPos;                // Dispatch position
      _vuint8 un8CPos;                // Check position
    } mt_tEB;
    
    // Initialization functions
    _vvoid _CreateColorMap_(_vbool bUseExclusiveColormap);
    _vvoid _InitColorLUT_(_vvoid);
    _vvoid _CreateDefaultCursor_(_vvoid);
    // Rendering functions
    inline _vvoid _ConvertRB2RI_(_vvoid); 
    // Simple dialog functions
    _vvoid _OBox_(const _vchar *pcCaption, const _vchar *pcText, _EVortexMsgIcon eIcon, _EVortexMsgSize eSize);
    _EVortexMsgRetVal _OCYNARBox_(const _vchar *pcCaption, const _vchar *pcText, _EVortexMsgIcon eIcon, _EVortexMsgSize eSize, _vuint8 un8Type);
    _EVortexMsgRetVal _YNCARIBox_(const _vchar *pcCaption, const _vchar *pcText, _EVortexMsgIcon eIcon, _EVortexMsgSize eSize, _vuint8 un8Type);
    // Event pumping functions
    inline _vbool _XPending_(_vvoid);
    inline _vbool _HandleInternalEvents_(XEvent &tEvent);
    inline KeySym _TranslateKeySymbol_(XEvent *ptEvent);

  public:
    // Constructor & destructor
    _CVortexVideoEnv();
    ~_CVortexVideoEnv();
    // Initialization functions
    _vbool CreateMainWindow(_vuint32 un32Width, _vuint32 un32Height, const _vchar *pcWindowTitle,
                            _vbool bUseExclusiveColormap = false, const _vchar *pcDisplayName = NULL);
    _vbool CreateRenderImage(_vuint32 un32Width = 0, _vuint32 un32Height = 0);
    _vbool CreateRenderBuffer(_vuint32 un32Width = 0, _vuint32 un32Height = 0);
    // Controling functions
    inline _vvoid MoveWindow(_vuint32 un32XPos,_vuint32 un32YPos);
    inline _vvoid AutoRepeatOn(_vvoid);
    inline _vvoid AutoRepeatOff(_vvoid);
    // Event pumping functions
    inline _vvoid ClearEvent(_vvoid);
    inline _vvoid DispatchEvent(_vvoid);
    inline _vvoid BlockedDispatchEvent(_vvoid);
    inline _vbool CheckEvent(_vvoid);
    // Keyboard event functions
    inline _vbool EventKeyPress(_vvoid);
    inline _vbool EventKeyRelease(_vvoid);
    inline _vint32 GetRawKeyCode(_vvoid);
    inline _EVortexKeySymbol GetKeySym(_vvoid);
    // Mouse event functions
    inline _vbool EventButtonPress(_vvoid);
    inline _vbool EventButtonRelease(_vvoid);
    inline _vint32 GetRawButtonCode(_vvoid);
    inline _EVortexBtnSymbol GetBtnSym(_vvoid);
    inline _vbool EventMouseMove(_vvoid);
    inline _vint32 GetMouseX(_vvoid);
    inline _vint32 GetMouseY(_vvoid);
    inline _vbool GetButton1State(_vvoid);
    inline _vbool GetButton2State(_vvoid);
    inline _vbool GetButton3State(_vvoid);
    inline _vbool GetButton4State(_vvoid);
    inline _vbool GetButton5State(_vvoid);
    // Other event functions
    inline _vbool EventExpose(_vvoid);
    // Window state functions
    inline _vbool IsActive(_vvoid);
    inline _vbool IsInactive(_vvoid);
    inline _vbool IsVisible(_vvoid);
    inline _vbool IsInvisible(_vvoid);
    // Rendering functions
    inline _vuint8 *RBuffData(_vvoid);
    inline _vvoid  DisplayRenderImage(_vvoid);
    inline _vvoid  DisplayRenderImage(_vint32 n32X, _vint32 n32Y);
    // Drawing functions
    inline _vvoid UseDefaultArrowCursor(_vvoid);
    inline _vvoid UseDefaultWaitCursor(_vvoid);
    inline _vvoid UseDefaultSmallFont(_vvoid);
    inline _vvoid UseDefaultBigFont(_vvoid);
    inline _vvoid SetFont(_TVortexFont *ptFont);
    inline _TVortexColor CreateColor(_vuint8 un8R, _vuint8 un8G, _vuint8 un8B);
    inline _vvoid SetForegroundColor(_TVortexColor un32Color);
    inline _vvoid SetFillStyle(_vint32 n32FillStyle);
    inline _vvoid DrawText(_vint32 n32X, _vint32 n32Y, const _vchar *pcText, _vuint32 un32Len); 
    inline _vvoid DrawTextXCenter(_vint32 n32Y, const _vchar *pcText, _vuint32 un32Len); 
    inline _vvoid DrawTextML(_vint32 n32X, _vint32 n32Y, const _vchar *pcText); 
    inline _vvoid DrawButton(_CVortexButton &csVButton, _vbool bPressed = false);
    inline _vvoid DrawButton(_CVortexButton &csVvButton, _vint32 n32MouseX, _vint32 n32MouseY, _vbool bPressed = false);
    inline _vvoid DrawPanelBar(_CVortexPanelBar &csVPanelBar);
    inline _vvoid DrawIcon2424SCol(_vint32 n32X, _vint32 n32Y, _vuint8 aun8Data[72]);
    inline _vvoid DrawMsgIcon(_vint32 n32X, _vint32 n32Y, _EVortexMsgIcon eVMsgIcon);
    inline _vvoid UpdateMainWindow(_vvoid);
    // Creation functions
    _vbool CreateButton(_CVortexButton &csVButton, 
                        _vint32 n32X1, _vint32 n32Y1, _vint32 n32X2, _vint32 n32Y2,
                        const _vchar *pcText, _vbool bInsideOutline = false,
                        _vbool bThickOutline = false, _vbool bDuplicateString = true);
    _vbool CreatePanelBar(_CVortexPanelBar &csVPanelBar, 
                          _vint32 n32X1, _vint32 n32Y1, _vint32 n32X2, _vint32 n32Y2,
                          const _vchar *pcCaption, _vbool bDuplicateString = true);
    // Simple dialog functions
    _EVortexMsgRetVal MessageBox(const _vchar *pcCaption, const _vchar *pcText,
                                 _EVortexMsgIcon eIcon = EIconInformation,
                                 _EVortexMsgButton eButton = EBtnOK,
                                 _EVortexMsgSize eSize = ESizeSmall);
    // Special overloaded operators (to support aligned memory allocation for SSE)
    inline _vvoid *operator new      (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
    inline _vvoid *operator new[]    (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
    inline _vvoid  operator delete   (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }
    inline _vvoid  operator delete[] (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }

  // Friend classes
  friend class _CVortex3DEnv;
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The X11 display class
// --------------------------------------------------------------------------------------------
// ***** Controling functions
_vvoid _CVortexVideoEnv::MoveWindow(_vuint32 un32XPos,_vuint32 un32YPos)
{
  XMoveWindow(mt_tX11.ptDisplay, mt_tX11.tMainWindow, un32XPos, un32YPos);
  XSync(mt_tX11.ptDisplay, true);
}
_vvoid _CVortexVideoEnv::AutoRepeatOn(_vvoid)
{
  XAutoRepeatOn(mt_tX11.ptDisplay);
  mt_tX11.bAutoRepeat = true;
}
_vvoid _CVortexVideoEnv::AutoRepeatOff(_vvoid)
{
  XAutoRepeatOff(mt_tX11.ptDisplay);
  mt_tX11.bAutoRepeat = false;
}

// ***** Event pumping functions
_vvoid _CVortexVideoEnv::ClearEvent(_vvoid)
{
  memset(&mt_tEB, 0, sizeof(mt_tEB));
  mt_tEB.ptEvent = &mt_tEB.tEmptyEvent;
  XSync(mt_tX11.ptDisplay, true);
}
_vvoid _CVortexVideoEnv::DispatchEvent(_vvoid)
{
  if(mt_bVisible)
  {
    while(_XPending_())
    {
      XNextEvent(mt_tX11.ptDisplay, &mt_tEB.atEvents[mt_tEB.un8DPos]);
      if(!_HandleInternalEvents_(mt_tEB.atEvents[mt_tEB.un8DPos]))
      {
        mt_tEB.un8DPos++;
        if(mt_tEB.un8DPos >= _MAX_EVENTS_) mt_tEB.un8DPos = 0;
      }
    }
    if(mt_bActive) XSync(mt_tX11.ptDisplay, true);
  }
  else
    BlockedDispatchEvent();
}
inline _vvoid _CVortexVideoEnv::BlockedDispatchEvent(_vvoid)
{
  XNextEvent(mt_tX11.ptDisplay, &mt_tEB.atEvents[mt_tEB.un8DPos]);
  if(!_HandleInternalEvents_(mt_tEB.atEvents[mt_tEB.un8DPos]))
  {
    mt_tEB.un8DPos++;
    if(mt_tEB.un8DPos >= _MAX_EVENTS_) mt_tEB.un8DPos = 0;
  }
}
_vbool _CVortexVideoEnv::CheckEvent(_vvoid)
{
  if(mt_tEB.un8CPos == mt_tEB.un8DPos)
  {
    mt_tEB.ptEvent = &mt_tEB.tEmptyEvent;
    return(false);
  }
  mt_tEB.ptEvent = & mt_tEB.atEvents[mt_tEB.un8CPos];
  mt_tEB.un8CPos++;
  if(mt_tEB.un8CPos >= _MAX_EVENTS_) mt_tEB.un8CPos = 0;
  return(true);
}
_vbool _CVortexVideoEnv::_XPending_(_vvoid)
{
  // ..... Static variables
  static struct  timeval tTimeOut = {0, 0};  
  static fd_set  tFDS;
  static _vint32 n32X11FD;
  
  // ..... Check for queued events
  XFlush(mt_tX11.ptDisplay);
  if(XEventsQueued(mt_tX11.ptDisplay, QueuedAlready)) return(true);
  
  // ..... If not, check whether the file descriptor has changed status
  n32X11FD = ConnectionNumber(mt_tX11.ptDisplay);  
  FD_ZERO(&tFDS);
  FD_SET(n32X11FD, &tFDS);
  if(select(n32X11FD + 1, &tFDS, NULL, NULL, &tTimeOut) == 1)
    if(XPending(mt_tX11.ptDisplay)) return(true);
  
  // ..... No queued events found
  return(false);
}
_vbool _CVortexVideoEnv::_HandleInternalEvents_(XEvent &tEvent)
{
  // ..... Static variables
  static _vbool bFocus   = true;
  static _vbool bMap     = true;
  static _vbool bVisible = true;

  // ..... Temporary variables
  _vbool bInt = false;

  // ..... Handle events
  if(tEvent.type == FocusIn)
  {
    bFocus = true;
    bInt   = true;
  }
  else if(tEvent.type == FocusOut)
  {
    bFocus = false;
    bInt   = true;
  }
  else if(tEvent.type == MapNotify)
  {
    bMap = true;
    bInt = true;
  }
  else if(tEvent.type == UnmapNotify)
  {
    bMap = false;
    bInt = true;
  }
  else if(tEvent.type == VisibilityNotify)
  {
    bVisible = (tEvent.xvisibility.state != VisibilityFullyObscured);
    bInt     = true;
  }

  // ..... Adjust the application states
  if(bInt)
  {
    mt_bVisible = bMap & bVisible;
    mt_bActive  = bFocus & mt_bVisible;
    if(mt_bActive)
    {
      if(mt_tX11.bAutoRepeat)
        XAutoRepeatOn(mt_tX11.ptDisplay);
      else
        XAutoRepeatOff(mt_tX11.ptDisplay);
    }
    else
      XAutoRepeatOn(mt_tX11.ptDisplay);
  }

  // .... Tell whether the internal event is handled
  return(bInt);
}
KeySym _CVortexVideoEnv::_TranslateKeySymbol_(XEvent *ptEvent)
{
  return(XKeycodeToKeysym(mt_tX11.ptDisplay, ptEvent->xkey.keycode, 0));
}

// ***** Keyboard event functions
_vbool _CVortexVideoEnv::EventKeyPress(_vvoid)
{
  return(mt_tEB.ptEvent->type == KeyPress);
}
_vbool _CVortexVideoEnv::EventKeyRelease(_vvoid)
{
  return(mt_tEB.ptEvent->type == KeyRelease);
}
_vint32 _CVortexVideoEnv::GetRawKeyCode(_vvoid)
{
  return(mt_tEB.ptEvent->xkey.keycode);
}
_EVortexKeySymbol _CVortexVideoEnv::GetKeySym(_vvoid)
{
  // ..... Temporary variables
  KeySym   tKS;
  _vuint32 un32KS;
  
  // ..... Get the X11 key symbol
  tKS = _TranslateKeySymbol_(mt_tEB.ptEvent);

  // ..... Translate the symbol  
  if(tKS)
  {
    switch(tKS & 0xFFFFFF00)
    {
      case 0x00000000: // Latin 1
      case 0x00000100: // Latin 2
      case 0x00000200: // Latin 3 
      case 0x00000300: // Latin 4
      case 0x00000400: // Katakana
      case 0x00000500: // Arabic
      case 0x00000600: // Cyrillic
      case 0x00000700: // Greek
      case 0x00000C00: // Hebrew
      case 0x00000D00: // Thai
      case 0x00000E00: // Korean
      case 0x00001400: // Armenian
      case 0x00001500: // Georgian
      case 0x00001600: // Caucasius
      case 0x00001E00: // Vietnamese
        un32KS = tKS & 0x000000FF;
        if((un32KS >= 'A') && (un32KS <= 'Z')) un32KS += ('a' - 'A');
        return(_EVortexKeySymbol(un32KS));
      case 0x0000FF00:
        return(_aeX11VKS_[tKS & 0x000000FF]);
      default:      
        return(VKS_UNKNOWN);
    }
  }
  return(VKS_UNKNOWN);
}

// ***** Mouse event functions
_vbool _CVortexVideoEnv::EventButtonPress(_vvoid)
{
  return(mt_tEB.ptEvent->type == ButtonPress);
}
_vbool _CVortexVideoEnv::EventButtonRelease(_vvoid)
{
  return(mt_tEB.ptEvent->type == ButtonRelease);
}
_vint32 _CVortexVideoEnv::GetRawButtonCode(_vvoid)
{
  return(mt_tEB.ptEvent->xbutton.button);
}
_EVortexBtnSymbol _CVortexVideoEnv::GetBtnSym(_vvoid)
{
  return(_aeX11VBS_[mt_tEB.ptEvent->xbutton.button]);
}
_vbool _CVortexVideoEnv::EventMouseMove(_vvoid)
{
  return(mt_tEB.ptEvent->type == MotionNotify);
}
_vint32 _CVortexVideoEnv::GetMouseX(_vvoid)
{
  return(mt_tEB.ptEvent->xmotion.x);
}
_vint32 _CVortexVideoEnv::GetMouseY(_vvoid)
{
  return(mt_tEB.ptEvent->xmotion.y);
}
_vbool _CVortexVideoEnv::GetButton1State(_vvoid)
{
  return(mt_tEB.ptEvent->xmotion.state & Button1Mask);
}
_vbool _CVortexVideoEnv::GetButton2State(_vvoid)
{
  return(mt_tEB.ptEvent->xmotion.state & Button2Mask);
}
_vbool _CVortexVideoEnv::GetButton3State(_vvoid)
{
  return(mt_tEB.ptEvent->xmotion.state & Button3Mask);
}
_vbool _CVortexVideoEnv::GetButton4State(_vvoid)
{
  return(mt_tEB.ptEvent->xmotion.state & Button4Mask);
}
_vbool _CVortexVideoEnv::GetButton5State(_vvoid)
{
  return(mt_tEB.ptEvent->xmotion.state & Button5Mask);
}

// ***** Other event functions
_vbool _CVortexVideoEnv::EventExpose(_vvoid)
{
 return((mt_tEB.ptEvent->type == Expose) & (mt_tEB.ptEvent->xexpose.count == 0));
}

// ***** Window state functions
_vbool _CVortexVideoEnv::IsActive(_vvoid)
{
  return(mt_bActive);
}
_vbool _CVortexVideoEnv::IsInactive(_vvoid)
{
  return(!mt_bActive);
}
_vbool _CVortexVideoEnv::IsVisible(_vvoid)
{
  return(mt_bVisible);
}
_vbool _CVortexVideoEnv::IsInvisible(_vvoid)
{
  return(!mt_bVisible);
}

// ***** Rendering functions
_vuint8 *_CVortexVideoEnv::RBuffData(_vvoid)
{
  return(mt_tRB.pun8Buffer);
}
_vvoid _CVortexVideoEnv::DisplayRenderImage(_vvoid)
{
  _ConvertRB2RI_();
  XPutImage(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, mt_tRI.ptImage,
            0, 0, 0, 0, mt_tRI.un32Width, mt_tRI.un32Height);
}
_vvoid _CVortexVideoEnv::DisplayRenderImage(_vint32 n32X, _vint32 n32Y)
{
  _ConvertRB2RI_();
  XPutImage(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, mt_tRI.ptImage,
            0, 0, n32X, n32Y, mt_tRI.un32Width, mt_tRI.un32Height);
}
_vvoid _CVortexVideoEnv::_ConvertRB2RI_(_vvoid)
{
  
  // Same size, 24 bpp
  if(mt_tRB.un32UT == 0)
    CopyMem4X8((_vuint32 *) mt_tRI.pucImage, (_vuint32 *) mt_tRI.pucImage, mt_tRB.un32Size48);
  // Same size, 16 bpp
  else if(mt_tRB.un32UT == 1)
  {
    _vuint16 *pun16Tmp = (_vuint16 *) mt_tRI.pucImage;
    _vuint8  *pun8Tmp  = mt_tRB.pun8Buffer;
    for(_vuint32 un32Cnt = 0; un32Cnt < mt_tRI.un32Dim; un32Cnt++)
    {
      *pun16Tmp++ = (*pun8Tmp >> 3) | (*++pun8Tmp << 3) | (*++pun8Tmp << 8);
      pun8Tmp++;
      pun8Tmp++;
    }
  }
  // Same size, 8 bpp
  else if(mt_tRB.un32UT == 2)
  {
    _vuchar *pucTmp  = mt_tRI.pucImage;
    _vuint8 *pun8Tmp = mt_tRB.pun8Buffer;
    for(_vuint32 un32Cnt = 0; un32Cnt < mt_tRI.un32Dim; un32Cnt++)
    {
      *pucTmp++ = mt_tRI.pucLUT[(*pun8Tmp >> 3) | (*++pun8Tmp << 2) | (*++pun8Tmp << 7)];
      pun8Tmp++;
      pun8Tmp++;
    }
  }
  // Different size, 24 bpp
  else if(mt_tRB.un32UT == 3)
  {
    _vuint32 *pun32Tmp1 = (_vuint32 *) mt_tRI.pucImage;
    _vuint32 *pun32Tmp2 = mt_tRB.pun32LUT;
    for(_vuint32 un32Cnt = 0; un32Cnt < mt_tRI.un32Dim; un32Cnt++)
      *pun32Tmp1++ = ((_vuint32 *) mt_tRB.pun8Buffer)[*pun32Tmp2++];
  }
  // Different size, 16 bpp
  else if(mt_tRB.un32UT == 4)
  {
    _vuint16 *pun16Tmp = (_vuint16 *) mt_tRI.pucImage;
    _vuint32 *pun32Tmp = mt_tRB.pun32LUT;
    for(_vuint32 un32Cnt = 0; un32Cnt < mt_tRI.un32Dim; un32Cnt++)
    {
      *pun16Tmp++ = (mt_tRB.pun8Buffer[*pun32Tmp] >> 3)   | 
                    (mt_tRB.pun8Buffer[*++pun32Tmp] << 3) | 
                    (mt_tRB.pun8Buffer[*++pun32Tmp] << 8);
      pun32Tmp++;
    }
  }
  // Different size, 8 bpp
  else
  {
    _vuchar  *pucTmp   = mt_tRI.pucImage;
    _vuint32 *pun32Tmp = mt_tRB.pun32LUT;
    for(_vuint32 un32Cnt = 0; un32Cnt < mt_tRI.un32Dim; un32Cnt++)
    {
      *pucTmp++ = mt_tRI.pucLUT[(mt_tRB.pun8Buffer[*pun32Tmp] >> 3)   |
                                (mt_tRB.pun8Buffer[*++pun32Tmp] << 2) |
                                (mt_tRB.pun8Buffer[*++pun32Tmp] << 7)];
      pun32Tmp++;
    }
  }
}

// Drawing functions
_vvoid _CVortexVideoEnv::UseDefaultArrowCursor(_vvoid)
{
  XDefineCursor(mt_tX11.ptDisplay, mt_tX11.tMainWindow, mt_tX11.tDefCurArrow);
}
_vvoid _CVortexVideoEnv::UseDefaultWaitCursor(_vvoid)
{
  XDefineCursor(mt_tX11.ptDisplay, mt_tX11.tMainWindow, mt_tX11.tDefCurWait);
}
_vvoid  _CVortexVideoEnv::UseDefaultSmallFont(_vvoid)
{
  SetFont(mt_tX11.ptDefFontSmall);
}
_vvoid  _CVortexVideoEnv::UseDefaultBigFont(_vvoid)
{
  SetFont(mt_tX11.ptDefFontBig);
}
_vvoid _CVortexVideoEnv::SetFont(_TVortexFont *ptFont)
{
  if(mt_tX11CS.ptFont == ptFont) return;
  mt_tX11CS.ptFont = ptFont;
  XSetFont(mt_tX11.ptDisplay, mt_tX11.tMainGC, mt_tX11CS.ptFont->fid);
}
_TVortexColor _CVortexVideoEnv::CreateColor(_vuint8 un8R, _vuint8 un8G, _vuint8 un8B)
{
  if(mt_tX11.un32Depth == 24) return(X11_RGB24(un8R, un8G, un8B));
  if(mt_tX11.un32Depth == 16) return(X11_RGB16(un8R, un8G, un8B));
  return(mt_tRI.pucLUT[(_vuint32(un8R & 0xF8) << 7) | (_vuint32(un8G & 0xF8) << 2) | _vuint32(un8B >> 3)]);
}
_vvoid _CVortexVideoEnv::SetForegroundColor(_TVortexColor un32Color)
{ 
  if(mt_tX11CS.un32Foreground == un32Color) return;
  mt_tX11CS.un32Foreground = un32Color;
  XSetForeground(mt_tX11.ptDisplay, mt_tX11.tMainGC, mt_tX11CS.un32Foreground);
}
_vvoid _CVortexVideoEnv::SetFillStyle(_vint32 n32FillStyle)
{
  if(mt_tX11CS.n32FillStyle == n32FillStyle) return;
  mt_tX11CS.n32FillStyle = n32FillStyle;
  XSetFillStyle(mt_tX11.ptDisplay, mt_tX11.tMainGC, mt_tX11CS.n32FillStyle);
}
_vvoid _CVortexVideoEnv::DrawText(_vint32 n32X, _vint32 n32Y, const _vchar *pcText, _vuint32 un32Len)
{
  XDrawString(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32X, n32Y, pcText, un32Len);
}
_vvoid _CVortexVideoEnv::DrawTextXCenter(_vint32 n32Y, const _vchar *pcText, _vuint32 un32Len)
{
  XDrawString(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, 
              (mt_tX11.un32Width - XTextWidth(mt_tX11CS.ptFont, pcText, un32Len)) >> 1, n32Y,
              pcText, un32Len);
}
_vvoid _CVortexVideoEnv::DrawTextML(_vint32 n32X, _vint32 n32Y, const _vchar *pcText)
{
  // ..... Temporary variables
  _vchar *pcTmp;
  _vchar *pcSrc;

  // ..... Draw text
  SetForegroundColor(mt_tX11Style.un32Text);
  pcSrc = strdup(pcText);
  pcTmp = strtok(pcSrc, "\n");
  while(pcTmp)
  {
    XDrawString(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, 
                n32X, n32Y, pcTmp, strlen(pcTmp));
    n32Y += (mt_tX11Style.ptFont->ascent + mt_tX11Style.ptFont->descent);
    pcTmp = strtok(NULL, "\n");
  }
  free(pcSrc);
}
_vvoid _CVortexVideoEnv::DrawButton(_CVortexButton &csVButton, _vbool bPressed)
{
  // ..... Draw button's face
  SetForegroundColor(mt_tX11Style.un32Face);
  SetFillStyle(mt_tX11Style.n32FillStyle);
  XFillRectangle(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC,
                 csVButton.mt_tFace.x, csVButton.mt_tFace.y,
                 csVButton.mt_tFace.width, csVButton.mt_tFace.height);

  // ..... Draw button's text
  SetForegroundColor(mt_tX11Style.un32Text);
  SetFont(mt_tX11Style.ptFont);
  XDrawString(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, 
              csVButton.mt_n32Xt, csVButton.mt_n32Yt,
              csVButton.mt_pcText, csVButton.mt_un32TextLen);

  // ..... Draw button's border
  SetForegroundColor(mt_tX11Style.un32WholeBorder);
  XDrawRectangle(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC,
                 csVButton.mt_tOut.x, csVButton.mt_tOut.y,
                 csVButton.mt_tOut.width, csVButton.mt_tOut.height);
  if(csVButton.mt_bThickOutline)
  {
    XDrawRectangle(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC,
                   csVButton.mt_tOutThick.x, csVButton.mt_tOutThick.y,
                   csVButton.mt_tOutThick.width, csVButton.mt_tOutThick.height);
  }
  if(csVButton.mt_bInsideOutline)
  {
    XDrawRectangle(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC,
                   csVButton.mt_tIn.x, csVButton.mt_tIn.y,
                   csVButton.mt_tIn.width, csVButton.mt_tIn.height);
  }
      
  if(bPressed)
  {
    SetForegroundColor(mt_tX11Style.un32DarkBorder);
    XDrawSegments(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, csVButton.mt_atSegTL, 4);
    SetForegroundColor(mt_tX11Style.un32LightBorder);
    XDrawSegments(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, csVButton.mt_atSegBR, 4);
  }
  else
  {
    SetForegroundColor(mt_tX11Style.un32LightBorder);
    XDrawSegments(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, csVButton.mt_atSegTL, 4);
    SetForegroundColor(mt_tX11Style.un32DarkBorder);
    XDrawSegments(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, csVButton.mt_atSegBR, 4);
  }
}
_vvoid _CVortexVideoEnv::DrawButton(_CVortexButton &csVButton, _vint32 n32MouseX, _vint32 n32MouseY, _vbool bPressed)
{
  if(csVButton.MouseOver(n32MouseX, n32MouseY))
    DrawButton(csVButton, bPressed);
  else
    DrawButton(csVButton);
}
_vvoid _CVortexVideoEnv::DrawPanelBar(_CVortexPanelBar &csVPanelBar)
{
  // ..... Draw the panel-bar's face
  SetForegroundColor(mt_tX11Style.un32Face);
  SetFillStyle(mt_tX11Style.n32FillStyle);
  XFillRectangles(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, csVPanelBar.mt_atFace, 2);

  // ..... Draw the panel-bar's border
  SetForegroundColor(mt_tX11Style.un32WholeBorder);
  XDrawRectangles(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, csVPanelBar.mt_atOut, 2);
  SetForegroundColor(mt_tX11Style.un32LightBorder);
  XDrawSegments(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, csVPanelBar.mt_atSegTL, 8);
  XDrawSegments(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, csVPanelBar.mt_atSegLL, 3);
  SetForegroundColor(mt_tX11Style.un32DarkBorder);
  XDrawSegments(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, csVPanelBar.mt_atSegBR, 8);
  XDrawSegments(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, &csVPanelBar.mt_atSegLL[3], 3);
  SetForegroundColor(mt_tX11Style.un32Face);
  
  // ..... Draw the panel-bar's caption
  XFillRectangle(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC,
                 csVPanelBar.mt_n32XC, csVPanelBar.mt_atFace[0].y,
                 csVPanelBar.mt_n32XCW, csVPanelBar.mt_atFace[0].height);
  SetFont(mt_tX11Style.ptFont);
  SetForegroundColor(mt_tX11Style.un32DarkBorder);
  XDrawString(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, 
              csVPanelBar.mt_n32XC + 5, csVPanelBar.mt_n32YC + 1,
              csVPanelBar.mt_pcCaption, csVPanelBar.mt_un32CaptionLen);
  SetForegroundColor(mt_tX11Style.un32Text);
  XDrawString(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, 
              csVPanelBar.mt_n32XC + 6, csVPanelBar.mt_n32YC,
              csVPanelBar.mt_pcCaption, csVPanelBar.mt_un32CaptionLen);
}
_vvoid _CVortexVideoEnv::DrawIcon2424SCol(_vint32 n32X, _vint32 n32Y, _vuint8 aun8Data[72])
{
  // ..... Temporary variables
  _vint32 n32XPos, n32YCnt, n32Cnt;
  
  // ..... Draw icon
  n32YCnt = 0;
  for(n32Cnt = 0; n32Cnt < 24; n32Cnt++)
  {
    n32XPos = n32X;
    if(aun8Data[n32YCnt] & 0x80) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x40) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x20) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x10) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x08) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x04) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x02) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x01) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    n32YCnt++;
    if(aun8Data[n32YCnt] & 0x80) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x40) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x20) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x10) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x08) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x04) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x02) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x01) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    n32YCnt++;
    if(aun8Data[n32YCnt] & 0x80) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x40) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x20) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x10) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x08) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x04) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x02) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    if(aun8Data[n32YCnt] & 0x01) XDrawPoint(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC, n32XPos, n32Y); n32XPos++;
    n32YCnt++;
    n32Y++;
  }      
}
_vvoid _CVortexVideoEnv::DrawMsgIcon(_vint32 n32X, _vint32 n32Y, _EVortexMsgIcon eVMsgIcon)
{
  // ..... Draw icon border
  SetForegroundColor(mt_tX11Style.un32IconBorder);
  XDrawRectangle(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC,
                 n32X++, n32Y++, 33, 33);
  SetForegroundColor(mt_tX11Style.un32IconBackground);
  SetFillStyle(mt_tX11Style.n32FillStyle);
  XFillRectangle(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainGC,
                 n32X, n32Y, 32, 32);

  // ..... Draw the icon
  n32X += 4; n32Y += 4;
  if(eVMsgIcon == EIconInformation)
  {
    _vuint8 aun8I[72] = { 0x00,0x38,0x00, 0x00,0x38,0x00, 0x00,0x00,0x00,
                          0x00,0x00,0x00, 0x00,0x38,0x00, 0x00,0x7C,0x00,
                          0x00,0xFC,0x00, 0x00,0xDC,0x00, 0x00,0x1C,0x00,
                          0x00,0x1C,0x00, 0x00,0x1C,0x00, 0x00,0x1C,0x00,
                          0x00,0x1C,0x00, 0x00,0x1C,0x00, 0x00,0x1C,0x00, 
                          0x00,0x1C,0x00, 0x00,0x1C,0x00, 0x00,0x1C,0x00, 
                          0x00,0x1C,0x00, 0x00,0x1C,0x00, 0x00,0x1d,0X80,
                          0x00,0x1F,0X80, 0x00,0x1F,0X00, 0x00,0x0E,0X00 };
    SetForegroundColor(mt_tX11Style.un32IconInformation);
    DrawIcon2424SCol(n32X, n32Y, aun8I);
  }
  else if(eVMsgIcon == EIconQuestion)
  {
    _vuint8 aun8Q[72] = { 0x00,0x3C,0X00, 0x00,0x7E,0x00, 0x00,0xFF,0x00,
                          0x01,0xE7,0x80, 0x03,0xC3,0xC0, 0x07,0x81,0xE0,
                          0x0F,0x00,0xF0, 0x0E,0x00,0x70, 0x00,0x00,0x70, 
                          0x00,0x00,0xF0, 0x00,0x01,0xE0, 0x00,0x03,0xC0, 
                          0x00,0x07,0x80, 0x00,0x0F,0x00, 0x00,0x1E,0x00, 
                          0x00,0x3C,0x00, 0x00,0x38,0x00, 0x00,0x38,0x00, 
                          0x00,0x38,0x00, 0x00,0x38,0x00, 0x00,0x00,0x00, 
                          0x00,0x00,0x00, 0x00,0x38,0x00, 0x00,0x38,0x00 };
    
    SetForegroundColor(mt_tX11Style.un32IconQuestion);
    DrawIcon2424SCol(n32X, n32Y, aun8Q);
  }
  else if(eVMsgIcon == EIconWarning)
  {
    _vuint8 aun8W[72] = { 0x00,0xFE,0x00, 0x00,0xFE,0x00, 0x00,0xFE,0x00,
                          0x00,0xFE,0x00, 0x00,0xFE,0x00, 0x00,0xFE,0x00,
                          0x00,0x7C,0x00, 0x00,0x7C,0x00, 0x00,0x7C,0x00, 
                          0x00,0x7C,0x00, 0x00,0x7C,0x00, 0x00,0x7C,0x00, 
                          0x00,0x7C,0x00, 0x00,0x7C,0x00, 0x00,0x7C,0x00, 
                          0x00,0x38,0x00, 0x00,0x38,0x00, 0x00,0x38,0x00,
                          0x00,0x38,0x00, 0x00,0x38,0x00, 0x00,0x00,0x00,
                          0x00,0x00,0x00, 0x00,0x38,0x00, 0x00,0x38,0x00 };
    SetForegroundColor(mt_tX11Style.un32IconWarning);
    DrawIcon2424SCol(n32X, n32Y, aun8W);
  }
  else if(eVMsgIcon == EIconError)
  {
    _vuint8 aun8E[72] = { 0x20,0x00,0x04, 0x70,0x00,0x0E, 0xF8,0x00,0x1F,
                          0x7C,0x00,0x3E, 0x3E,0x00,0x7C, 0x1F,0x00,0xF8,
                          0x0F,0x81,0xF0, 0x07,0xC3,0xE0, 0x03,0xE7,0xC0,
                          0x01,0xFF,0x80, 0x00,0xFF,0x00, 0x00,0x7E,0x00,
                          0x00,0x7E,0x00, 0x00,0xFF,0x00, 0x01,0xFF,0x80,
                          0x03,0xE7,0xC0, 0x07,0xC3,0xE0, 0x0F,0x81,0xF0,
                          0x1F,0x00,0xF8, 0x3E,0x00,0x7C, 0x7C,0x00,0x3E,
                          0xF8,0x00,0x1F, 0x70,0x00,0x0E, 0x20,0x00,0x04 };
    SetForegroundColor(mt_tX11Style.un32IconError);
    DrawIcon2424SCol(n32X, n32Y, aun8E);
  }
}
_vvoid _CVortexVideoEnv::UpdateMainWindow(_vvoid)
{
  XCopyArea(mt_tX11.ptDisplay, mt_tX11.tMainPixmap, mt_tX11.tMainWindow, mt_tX11.tMainGC,
            0, 0, mt_tX11.un32Width, mt_tX11.un32Height, 0, 0);
}
    
#endif
// ============================================================================================
// End of vx11.h
// ============================================================================================
