// ############################################################################################
//  vx11cur.h
//  Definition of default cursors used by _CVortexVideoEnv class using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VX11CUR_H_
#define _VX11CUR_H_

// ============================================================================================
// Defines
// ============================================================================================
// ***** Default "arrow" cursor
//   FDEC BA98 7654 3210              FDEC BA98 7654 3210 
// 0 .... .... .... ....  0000      0 .... .... .... ..00  0003
// 1 .... .... .... ..O.  0002      1 .... .... .... 0000  000F
// 2 .... .... .... OO..  000C      2 .... .... ..00 000.  003E
// 3 .... .... ..OO OO..  003C      3 .... .... 0000 000.  00FE
// 4 .... .... OOO. O...  00E8      4 .... ..00 0000 00..  03FC
// 5 .... ..OO O..O O...  0398      5 .... 0000 0000 00..  0FFC
// 6 .... OOO. ...O ....  0E10      6 ..00 0000 0000 0...  3FF8
// 7 ..OO O... ..OO ....  3830      7 .000 0000 0000 0...  7FF8
// 8 .OOO OOOO ..O. ....  7F20      8 0000 0000 0000 ....  FFF0
// 9 .... ...O .OO. ....  0160      9 0000 0000 0000 ....  FFF0
// A .... ...O .O.. ....  0140      A .... ..00 000. ....  03E0
// B .... ...O OO.. ....  01C0      B .... ..00 000. ....  03E0
// C .... ...O O... ....  0180      C .... ..00 00.. ....  03C0
// D .... ...O O... ....  0180      D .... ..00 00.. ....  03C0
// E ... ....O .... ....  0100      E .... ..00 0... ....  0380
// F ... ..... ... .....  0000      F .... ..00 .... ....  0300
#define _VX11CUR_ARROW_      { 0x0000, 0x0002, 0x000C, 0x003C, 0x00E8, 0x0398, 0x0E10, 0x3830, \
                               0x7F20, 0x0160, 0x0140, 0x01C0, 0x0180, 0x0180, 0x0100, 0x0000 }
#define _VX11CUR_ARROW_MASK_ { 0x0003, 0x000F, 0x003E, 0x00FE, 0x03FC, 0x0FFC, 0x3FF8, 0x7FF8, \
                               0xFFF0, 0xFFF0, 0x03E0, 0x03E0, 0x03C0, 0x03C0, 0x0380, 0x0300 }
#define _VX11CUR_ARROW_X_HP_   0
#define _VX11CUR_ARROW_Y_HP_   0
#define _VX11CUR_ARROW_X_SIZE_ 16
#define _VX11CUR_ARROW_Y_SIZE_ 16

// ***** Default "wait" cursor
//   FDEC BA98 7654 3210              FDEC BA98 7654 3210 
// 0 .... .... .... ....  0000      0 ..oo oooo oooo oo..  3FFC
// 1 ...o oooo oooo o...  1FF8      1 ..oo oooo oooo oo..  3FFC
// 2 ...o .... .... o...  1008      2 ..oo oooo oooo oo..  3FFC
// 3 .... o... ...o ....  0810      3 ...o oo.. ..oo o...  1C38
// 4 .... o... ...o ....  0810      4 ...o oo.. ..oo o...  1C38
// 5 .... .o.. ..o. ....  0420      5 .... ooo. .ooo ....  0E70
// 6 .... .o.. ..o. ....  0420      6 .... ooo. .ooo ....  0E70
// 7 .... ..o. .o.. ....  0240      7 .... .ooo ooo. ....  07E0
// 8 .... ..o. .o.. ....  0240      8 .... .ooo ooo. ....  07E0
// 9 .... .o.. ..o. ....  0420      9 .... ooo. .ooo ....  0E70
// A .... .o.. ..o. ....  0420      A .... ooo. .ooo ....  0E70
// B .... o... ...o ....  0810      B ...o oo.. ..oo o...  1C38
// C .... o... ...o ....  0810      C ...o oo.. ..oo o...  1C38
// D ...o .... .... o...  1008      D ..oo oooo oooo oo..  3FFC
// E ...o oooo oooo o...  1FF8      E ..oo oooo oooo oo..  3FFC
// F .... .... .... ....  0000      F ..oo oooo oooo oo..  3FFC
#define _VX11CUR_WAIT_      { 0x0000, 0x1FF8, 0x1008, 0x0810, 0x0810, 0x0420, 0x0420, 0x0240, \
                              0x0240, 0x0420, 0x0420, 0x0810, 0x0810, 0x1008, 0x1FF8, 0x0000 }
#define _VX11CUR_WAIT_MASK_ { 0x3FFC, 0x3FFC, 0x3FFC, 0x1C38, 0x1C38, 0x0E70, 0x0E70, 0x07E0, \
                              0x07E0, 0x0E70, 0x0E70, 0x1C38, 0x1C38, 0x3FFC, 0x3FFC, 0x3FFC }
#define _VX11CUR_WAIT_X_HP_   7
#define _VX11CUR_WAIT_Y_HP_   7
#define _VX11CUR_WAIT_X_SIZE_ 16
#define _VX11CUR_WAIT_Y_SIZE_ 16

#endif
// ============================================================================================
// End of vx11cur.h
// ============================================================================================
