// ############################################################################################
//  vllsbcf.cpp
//  Implementation of low-level sound buffer functions used by _CVortexAudioEnv and
//  _CVortexSoundBuffer
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGEs headers
#include <vllsbcf.h>
    
// ============================================================================================
// Local variables declaration
// ============================================================================================
// ***** Used by converter functions
static _vuint32 un32Pos0, un32Pos1;

// ===========================================================================================
// Exported functions implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// Converter functions without frequency scalling
// --------------------------------------------------------------------------------------------
_vvoid _8Mn8Mn_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff) // *** Unchecked ***
{
  *pucDstBuff = *tData.pucLeftCur++;
  if(tData.pucLeftCur > tData.pucLeftMax)
  {
    tData.pucLeftCur  = tData.pucLeft;
    tData.bPlaying    = tData.bLooping;
  }
}
_vvoid _8Mn8St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff) // *** Unchecked ***
{
  pucDstBuff[1] = pucDstBuff[0] = *tData.pucLeftCur++;
  if(tData.pucLeftCur > tData.pucLeftMax)
  {
    tData.pucLeftCur  = tData.pucLeft;
    tData.bPlaying    = tData.bLooping;
  }
}
_vvoid _8Mn16Mn_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff) // *** Unchecked ***
{
  pucDstBuff[0] = 0;
  pucDstBuff[1] = *tData.pucLeftCur++;
  if(tData.pucLeftCur > tData.pucLeftMax)
  {
    tData.pucLeftCur = tData.pucLeft;
    tData.bPlaying   = tData.bLooping;
  }
}
_vvoid _8Mn16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff)
{
  pucDstBuff[2] = pucDstBuff[0] = 0;
  pucDstBuff[3] = pucDstBuff[1] = *tData.pucLeftCur++;
  if(tData.pucLeftCur > tData.pucLeftMax)
  {
    tData.pucLeftCur = tData.pucLeft;
    tData.bPlaying   = tData.bLooping;
  }
}
_vvoid _8St8St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff) // *** Unchecked ***
{
  pucDstBuff[0] = *tData.pucLeftCur++;
  pucDstBuff[1] = *tData.pucRightCur++;
  if(tData.pucLeftCur > tData.pucLeftMax)
  {
    tData.pucLeftCur  = tData.pucLeft;
    tData.pucRightCur = tData.pucRight;
    tData.bPlaying    = tData.bLooping;
  }
}
_vvoid _8St16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff)
{
  pucDstBuff[2] = pucDstBuff[0] = 0;
  pucDstBuff[1] = *tData.pucLeftCur++;
  pucDstBuff[3] = *tData.pucRightCur++;
  if(tData.pucLeftCur > tData.pucLeftMax)
  {
    tData.pucLeftCur  = tData.pucLeft;
    tData.pucRightCur = tData.pucRight;
    tData.bPlaying    = tData.bLooping;
  }
}
_vvoid _16Mn16Mn_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff) // *** Unchecked ***
{
  pucDstBuff[0] = *tData.pucLeftCur++;
  pucDstBuff[1] = *tData.pucLeftCur++;
  if(tData.pucLeftCur > tData.pucLeftMax)
  {
    tData.pucLeftCur = tData.pucLeft;
    tData.bPlaying   = tData.bLooping;
  }
}
_vvoid _16Mn16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff)
{
  pucDstBuff[2] = pucDstBuff[0] = *tData.pucLeftCur++;
  pucDstBuff[3] = pucDstBuff[1] = *tData.pucLeftCur++;
  if(tData.pucLeftCur > tData.pucLeftMax)
  {
    tData.pucLeftCur = tData.pucLeft;
    tData.bPlaying   = tData.bLooping;
  }
}
_vvoid _16St16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff)
{
  pucDstBuff[0] = *tData.pucLeftCur++;
  pucDstBuff[1] = *tData.pucLeftCur++;
  pucDstBuff[2] = *tData.pucRightCur++;
  pucDstBuff[3] = *tData.pucRightCur++;
  if(tData.pucLeftCur > tData.pucLeftMax)
  {
    tData.pucLeftCur  = tData.pucLeft;
    tData.pucRightCur = tData.pucRight;
    tData.bPlaying    = tData.bLooping;
  }
}

// --------------------------------------------------------------------------------------------
// Converter functions with frequency scalling
// --------------------------------------------------------------------------------------------
_vvoid _X8Mn8Mn_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff) // *** Unchecked ***
{
  un32Pos0 = tData.un32PosCur >> _WSTEP_PREC_;
  *pucDstBuff = tData.pucLeft[un32Pos0];
  tData.un32PosCur += tData.un32PosStep;
  if(tData.un32PosCur > tData.un32PosMax)
  {
    tData.un32PosCur = 0;
    tData.bPlaying   = tData.bLooping;
  }
}
_vvoid _X8Mn8St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff) // *** Unchecked ***
{
  un32Pos0 = tData.un32PosCur >> _WSTEP_PREC_;
  pucDstBuff[1] = pucDstBuff[0] = tData.pucLeft[un32Pos0];
  tData.un32PosCur += tData.un32PosStep;
  if(tData.un32PosCur > tData.un32PosMax)
  {
    tData.un32PosCur = 0;
    tData.bPlaying   = tData.bLooping;
  }
}
_vvoid _X8Mn16Mn_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff) // *** Unchecked ***
{
  pucDstBuff[0] = 0;
  pucDstBuff[1] = tData.pucLeft[tData.un32PosCur >> _WSTEP_PREC_];
  tData.un32PosCur += tData.un32PosStep;
  if(tData.un32PosCur > tData.un32PosMax)
  {
    tData.un32PosCur = 0;
    tData.bPlaying   = tData.bLooping;
  }
}
_vvoid _X8Mn16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff)
{
  pucDstBuff[2] = pucDstBuff[0] = 0;
  pucDstBuff[3] = pucDstBuff[1] = tData.pucLeft[tData.un32PosCur >> _WSTEP_PREC_];
  tData.un32PosCur += tData.un32PosStep;
  if(tData.un32PosCur > tData.un32PosMax)
  {
    tData.un32PosCur = 0;
    tData.bPlaying   = tData.bLooping;
  }
}
_vvoid _X8St8St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff) // *** Unchecked ***
{
  un32Pos0 = tData.un32PosCur >> _WSTEP_PREC_;
  pucDstBuff[0] = tData.pucLeft[un32Pos0];
  pucDstBuff[1] = tData.pucRight[un32Pos0];
  tData.un32PosCur += tData.un32PosStep;
  if(tData.un32PosCur > tData.un32PosMax)
  {
    tData.un32PosCur = 0;
    tData.bPlaying   = tData.bLooping;
  }
}
_vvoid _X8St16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff)
{
  pucDstBuff[2] = pucDstBuff[0] = 0;
  un32Pos0 = tData.un32PosCur >> _WSTEP_PREC_;
  pucDstBuff[1] = tData.pucLeft[un32Pos0];
  pucDstBuff[3] = tData.pucRight[un32Pos0];
  tData.un32PosCur += tData.un32PosStep;
  if(tData.un32PosCur > tData.un32PosMax)
  {
    tData.un32PosCur = 0;
    tData.bPlaying   = tData.bLooping;
  }
}
_vvoid _X16Mn16Mn_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff) // *** Unchecked ***
{
  un32Pos0 = (tData.un32PosCur >> _WSTEP_PREC_) & 0xFFFFFFFE;
  un32Pos1 = un32Pos0 | 0x00000001;
  pucDstBuff[0] = tData.pucLeft[un32Pos0];
  pucDstBuff[1] = tData.pucLeft[un32Pos1];
  tData.un32PosCur += tData.un32PosStep;
  if(tData.un32PosCur > tData.un32PosMax)
  {
    tData.un32PosCur = 0;
    tData.bPlaying   = tData.bLooping;
  }
}
_vvoid _X16Mn16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff)
{
  un32Pos0 = (tData.un32PosCur >> _WSTEP_PREC_) & 0xFFFFFFFE;
  un32Pos1 = un32Pos0 | 0x00000001;
  pucDstBuff[2] = pucDstBuff[0] = tData.pucLeft[un32Pos0];
  pucDstBuff[3] = pucDstBuff[1] = tData.pucLeft[un32Pos1];
  tData.un32PosCur += tData.un32PosStep;
  if(tData.un32PosCur > tData.un32PosMax)
  {
    tData.un32PosCur = 0;
    tData.bPlaying   = tData.bLooping;
  }
}
_vvoid _X16St16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff)
{
  un32Pos0 = (tData.un32PosCur >> _WSTEP_PREC_) & 0xFFFFFFFE;
  un32Pos1 = un32Pos0 | 0x00000001;
  pucDstBuff[0] = tData.pucLeft[un32Pos0];
  pucDstBuff[1] = tData.pucLeft[un32Pos1];
  pucDstBuff[2] = tData.pucRight[un32Pos0];
  pucDstBuff[3] = tData.pucRight[un32Pos1];
  tData.un32PosCur += tData.un32PosStep;
  if(tData.un32PosCur > tData.un32PosMax)
  {
    tData.un32PosCur = 0;
    tData.bPlaying   = tData.bLooping;
  }
}

// ============================================================================================
// End of vllsbcf.cpp
// ============================================================================================
