// ############################################################################################
//  vllsbcf.h
//  Prototype of low-level sound buffer functions used by _CVortexAudioEnv and
//  _CVortexSoundBuffer
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VLLSBCF_H_
#define _VLLSBCF_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vbtypes.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** Fixed point constants
#define _WSTEP_PREC_ 8
#define _MAX_WBUFF_SIZE_ _vuint32(1 << (31 - _WSTEP_PREC_))

// ============================================================================================
// Enumerations
// ============================================================================================
// ***** Raw data formats
typedef enum _EVRF_ { ERF_U8Mono, ERF_U8Stereo, ERF_S8Mono, ERF_S8Stereo, 
                      ERF_U16LEMono, ERF_U16LEStereo, ERF_S16LEMono, ERF_S16LEStereo,
                      ERF_U16BEMono, ERF_U16BEStereo, ERF_S16BEMono, ERF_S16BEStereo 
                    } _EVortexRawFormat;

// ============================================================================================
// Exported structures definition
// ============================================================================================
// ***** Audio settings
struct _TVortexAudSettings
{
  _vint32 n32Format;    // The audio data format
  _vint32 n32Frequency; // The sampling rate
  _vint32 n32Channels;  // The number of channels
};

// ***** The sound format 
struct _TVortexAudioFormat
{
  _vuint8  un8Bits;       // Bit depth (8 or 16)
  _vuint8  un8Sign;       // 0 = unsigned, 1 = signed
  _vuint8  un8Endian;     // 0 = LE, 1 = BE
  _vuint8  un8Channels;   // 1 = mono, 2 = stereo
  _vuint32 un32Frequency; // Sampling frequency
};

// ***** The sound buffer data
struct _TVortexSBData
{
  _vuchar  *pucLeft;     // The left/mono channel buffer
  _vuchar  *pucRight;    // The right channel buffer
  _vuchar  *pucLeftMax;  // The left/mono channel buffer maximum address
  _vuchar  *pucRightMax; // The right channel buffer maximum address
  _vuint32 un32PosMax;   // The maximum position (for frequency scalling)
  _vuint32 un32PosStep;  // The position step (for frequency scalling)
};

// ***** The played sound buffer data
struct _TVortexPlayedSBData
{
  _vuchar  *pucLeft;     // The left/mono channel buffer
  _vuchar  *pucRight;    // The right channel buffer
  _vuchar  *pucLeftMax;  // The left/mono channel buffer maximum address
  _vuchar  *pucRightMax; // The right channel buffer maximum address
  _vbool   bPlaying;     // The "playing" flag
  _vbool   bLooping;     // The "looping" flag
  _vuchar  *pucLeftCur;  // The left/mono channel buffer current address
  _vuchar  *pucRightCur; // The right channel buffer current address
  _vuint32 un32PosCur;   // The current position (for frequency scalling)
  _vuint32 un32PosMax;   // The maximum position (for frequency scalling)
  _vuint32 un32PosStep;  // The position step (for frequency scalling)
};

// ============================================================================================
// Exported functions prototype
// ============================================================================================
// ***** Converter functions without frequency scalling
//       All functions for 16-bit sound buffer will only work for little-endian computers
extern _vvoid _8Mn8Mn_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _8Mn8St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _8Mn16Mn_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _8Mn16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _8St8St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _8St16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _16Mn16Mn_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _16Mn16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _16St16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);

// ***** Converter functions with frequency scalling
//       All functions for 16-bit sound buffer will only work for little-endian computers
extern _vvoid _X8Mn8Mn_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _X8Mn8St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _X8Mn16Mn_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _X8Mn16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _X8St8St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _X8St16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _X16Mn16Mn_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _X16Mn16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
extern _vvoid _X16St16St_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);

#endif
// ============================================================================================
// End of vllsbcf.h
// ============================================================================================
