// ############################################################################################
//  vmat3d.cpp
//  Implementation of _CVortexMatrix3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <math.h>
// ***** VortexGE headers
#include <vmat3d.h>
#include <vmdisp.h>
#include <vluts.h>

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 3D matrix class
// --------------------------------------------------------------------------------------------
// ***** Display functions
_vvoid _CVortexMatrix3D::DumpMatrix(_vvoid)
{
  VInfo(_F_, _L_, "|   %+10.3f   %+10.3f   %+10.3f   %+10.3f   |\n|   %+10.3f   %+10.3f   %+10.3f   %+10.3f   |\n|   %+10.3f   %+10.3f   %+10.3f   %+10.3f   |\n|   %+10.3f   %+10.3f   %+10.3f   %+10.3f   |\n",
        mt_aafMatrix[0][0], mt_aafMatrix[0][1], mt_aafMatrix[0][2], mt_aafMatrix[0][3], mt_aafMatrix[1][0], mt_aafMatrix[1][1], mt_aafMatrix[1][2], mt_aafMatrix[1][3], 
        mt_aafMatrix[2][0], mt_aafMatrix[2][1], mt_aafMatrix[2][2], mt_aafMatrix[2][3], mt_aafMatrix[3][0], mt_aafMatrix[3][1], mt_aafMatrix[3][2], mt_aafMatrix[3][3]);
}

// ***** Set operations
_vvoid _CVortexMatrix3D::SetToLookAtMatrix(const _vfloat afCamPosXYZ[3], const _vfloat afCamTargetXYZ[3])
{
  // ..... Static variables
  static _vfloat afDefUpVect[3] = {0, 1, 0};
  
  // ..... Create matrix
  SetToLookAtMatrix(afCamPosXYZ, afCamTargetXYZ, afDefUpVect);
}
_vvoid _CVortexMatrix3D::SetToLookAtMatrix(const _vfloat afCamPosXYZ[3], const _vfloat afCamTargetXYZ[3], const _vfloat afUpVectXYZ[3])
{
  // ..... Temporary variables
  _vfloat afX[3], afY[3], afZ[3], fMagX, fMagY, fMagZ, fDotX, fDotY, fDotZ;
  
  // ..... Calculate the z-axis
  afZ[0] = afCamTargetXYZ[0] - afCamPosXYZ[0];
  afZ[1] = afCamTargetXYZ[1] - afCamPosXYZ[1];
  afZ[2] = afCamTargetXYZ[2] - afCamPosXYZ[2];
  fMagZ  = sqrt(afZ[0] * afZ[0] + afZ[1] * afZ[1] + afZ[2] * afZ[2]);
  afZ[0] /= fMagZ; afZ[1] /= fMagZ; afZ[2] /= fMagZ;
  
  // ..... Calculate the x-axis
  afX[0] = afUpVectXYZ[1] * afZ[2] - afUpVectXYZ[2] * afZ[1];
  afX[1] = afUpVectXYZ[2] * afZ[0] - afUpVectXYZ[0] * afZ[2]; 
  afX[2] = afUpVectXYZ[0] * afZ[1] - afUpVectXYZ[1] * afZ[0];
  fMagX  = sqrt(afX[0] * afX[0] + afX[1] * afX[1] + afX[2] * afX[2]);
  afX[0] /= fMagX; afX[1] /= fMagX; afX[2] /= fMagX;
  
  // ..... Calculate the y-axis
  afY[0] = afZ[1] * afX[2] - afZ[2] * afX[1];
  afY[1] = afZ[2] * afX[0] - afZ[0] * afX[2];
  afY[2] = afZ[0] * afX[1] - afZ[1] * afX[0];
  fMagY  = sqrt(afY[0] * afY[0] + afY[1] * afY[1] + afY[2] * afY[2]);
  
  // ..... Calculate the dot product
  fDotX = afX[0] * afCamPosXYZ[0] + afX[1] * afCamPosXYZ[1] + afX[2] * afCamPosXYZ[2];
  fDotY = afY[0] * afCamPosXYZ[0] + afY[1] * afCamPosXYZ[1] + afY[2] * afCamPosXYZ[2];
  fDotZ = afZ[0] * afCamPosXYZ[0] + afZ[1] * afCamPosXYZ[1] + afZ[2] * afCamPosXYZ[2];

  // ..... Set matrix 
  mt_aafMatrix[0][0] =  afX[0]; mt_aafMatrix[0][1] =  afY[0]; mt_aafMatrix[0][2] =  afZ[0]; mt_aafMatrix[0][3] = 0;
  mt_aafMatrix[1][0] =  afX[1]; mt_aafMatrix[1][1] =  afY[1]; mt_aafMatrix[1][2] =  afZ[1]; mt_aafMatrix[1][3] = 0;
  mt_aafMatrix[2][0] =  afX[2]; mt_aafMatrix[2][1] =  afY[2]; mt_aafMatrix[2][2] =  afZ[2]; mt_aafMatrix[2][3] = 0;
  mt_aafMatrix[3][0] = -fDotX;  mt_aafMatrix[3][1] = -fDotY;  mt_aafMatrix[3][2] = -fDotZ;  mt_aafMatrix[3][3] = 0;
}
_vvoid _CVortexMatrix3D::SetToShearingShear(_vfloat fX, _vfloat fY)
{
  // ..... Create shearing matrix  
  mt_aafMatrix[0][0] = 1; mt_aafMatrix[0][1] = 0; mt_aafMatrix[0][2] = fX; mt_aafMatrix[0][3] = 0;
  mt_aafMatrix[1][0] = 0; mt_aafMatrix[1][1] = 1; mt_aafMatrix[1][2] = fY; mt_aafMatrix[1][3] = 0;
  mt_aafMatrix[2][0] = 0; mt_aafMatrix[2][1] = 0; mt_aafMatrix[2][2] = 1;  mt_aafMatrix[2][3] = 0;
  mt_aafMatrix[3][0] = 0; mt_aafMatrix[3][1] = 0; mt_aafMatrix[3][2] = 0;  mt_aafMatrix[3][3] = 1;
}
_vvoid _CVortexMatrix3D::SetToScallingMatrix(_vfloat fX, _vfloat fY, _vfloat fZ)
{
  // ..... Create scalling matrix  
  mt_aafMatrix[0][0] = fX; mt_aafMatrix[0][1] = 0;  mt_aafMatrix[0][2] = 0;  mt_aafMatrix[0][3] = 0;
  mt_aafMatrix[1][0] = 0;  mt_aafMatrix[1][1] = fY; mt_aafMatrix[1][2] = 0;  mt_aafMatrix[1][3] = 0;
  mt_aafMatrix[2][0] = 0;  mt_aafMatrix[2][1] = 0;  mt_aafMatrix[2][2] = fZ; mt_aafMatrix[2][3] = 0;
  mt_aafMatrix[3][0] = 0;  mt_aafMatrix[3][1] = 0;  mt_aafMatrix[3][2] = 0;  mt_aafMatrix[3][3] = 1;
}
_vvoid _CVortexMatrix3D::SetToRotatationMatrix(_vint32 n32X, _vint32 n32Y, _vint32 n32Z)
{
  // ..... Temporary variables
  _vfloat aafMat[4][4];

  // ..... Create identity matrix  
  SetIdentity(mt_aafMatrix);
  
  // ..... Create Z rotation matrix
  aafMat[0][0] =  COS(n32Z); aafMat[0][1] = SIN(n32Z); aafMat[0][2] = 0; aafMat[0][3] = 0;
  aafMat[1][0] = -SIN(n32Z); aafMat[1][1] = COS(n32Z); aafMat[1][2] = 0; aafMat[1][3] = 0;
  aafMat[2][0] = 0;          aafMat[2][1] = 0;         aafMat[2][2] = 1; aafMat[2][3] = 0;
  aafMat[3][0] = 0;          aafMat[3][1] = 0;         aafMat[3][2] = 0; aafMat[3][3] = 1;
  _MergeMatrices_(mt_aafMatrix, aafMat);
  
  // ..... Create X rotation matrix
  aafMat[0][0] = 1; aafMat[0][1] = 0;          aafMat[0][2] = 0;         aafMat[0][3] = 0;
  aafMat[1][0] = 0; aafMat[1][1] =  COS(n32X); aafMat[1][2] = SIN(n32X); aafMat[1][3] = 0;
  aafMat[2][0] = 0; aafMat[2][1] = -SIN(n32X); aafMat[2][2] = COS(n32X); aafMat[2][3] = 0;
  aafMat[3][0] = 0; aafMat[3][1] = 0;          aafMat[3][2] = 0;         aafMat[3][3] = 1;
  _MergeMatrices_(mt_aafMatrix, aafMat);

  // ..... Create Y rotation matrix
  aafMat[0][0] = COS(n32Y); aafMat[0][1] = 0; aafMat[0][2] = -SIN(n32Y); aafMat[0][3] = 0;
  aafMat[1][0] = 0;         aafMat[1][1] = 1; aafMat[1][2] = 0;          aafMat[1][3] = 0;
  aafMat[2][0] = SIN(n32Y); aafMat[2][1] = 0; aafMat[2][2] = COS(n32Y);  aafMat[2][3] = 0;
  aafMat[3][0] = 0;         aafMat[3][1] = 0; aafMat[3][2] = 0;          aafMat[3][3] = 1;
  _MergeMatrices_(mt_aafMatrix, aafMat);
}
_vvoid _CVortexMatrix3D::SetToTranslationMatrix(_vfloat fX, _vfloat fY, _vfloat fZ)
{
  // ..... Create translation matrix  
  mt_aafMatrix[0][0] = 1;  mt_aafMatrix[0][1] = 0;  mt_aafMatrix[0][2] = 0;  mt_aafMatrix[0][3] = 0;
  mt_aafMatrix[1][0] = 0;  mt_aafMatrix[1][1] = 1;  mt_aafMatrix[1][2] = 0;  mt_aafMatrix[1][3] = 0;
  mt_aafMatrix[2][0] = 0;  mt_aafMatrix[2][1] = 0;  mt_aafMatrix[2][2] = 1;  mt_aafMatrix[2][3] = 0;
  mt_aafMatrix[3][0] = fX; mt_aafMatrix[3][1] = fY; mt_aafMatrix[3][2] = fZ; mt_aafMatrix[3][3] = 1;
}

// ***** Merged transformation operations  
_vvoid _CVortexMatrix3D::Shear(_vfloat fX, _vfloat fY)
{
  // ..... Temporary variables
  _vfloat aafMat[4][4];

  // ..... Create shearing matrix  
  aafMat[0][0] = 1; aafMat[0][1] = 0; aafMat[0][2] = fX; aafMat[0][3] = 0;
  aafMat[1][0] = 0; aafMat[1][1] = 1; aafMat[1][2] = fY; aafMat[1][3] = 0;
  aafMat[2][0] = 0; aafMat[2][1] = 0; aafMat[2][2] = 1;  aafMat[2][3] = 0;
  aafMat[3][0] = 0; aafMat[3][1] = 0; aafMat[3][2] = 0;  aafMat[3][3] = 1;
  
  // ..... Merge with the main matrix
  _MergeMatrix_(aafMat);
}
_vvoid _CVortexMatrix3D::Scale(_vfloat fX, _vfloat fY, _vfloat fZ)
{
  // ..... Temporary variables
  _vfloat aafMat[4][4];
  
  // ..... Create scalling matrix  
  aafMat[0][0] = fX; aafMat[0][1] = 0;  aafMat[0][2] = 0;  aafMat[0][3] = 0;
  aafMat[1][0] = 0;  aafMat[1][1] = fY; aafMat[1][2] = 0;  aafMat[1][3] = 0;
  aafMat[2][0] = 0;  aafMat[2][1] = 0;  aafMat[2][2] = fZ; aafMat[2][3] = 0;
  aafMat[3][0] = 0;  aafMat[3][1] = 0;  aafMat[3][2] = 0;  aafMat[3][3] = 1;

  // ..... Merge with the main matrix
  _MergeMatrix_(aafMat);
}
_vvoid _CVortexMatrix3D::Rotate(_vint32 n32X, _vint32 n32Y, _vint32 n32Z)
{
  // ..... Temporary variables
  _vfloat aafMat[4][4], aafRMatrix[4][4];

  // ..... Create identity matrix  
  SetIdentity(aafRMatrix);
  
  // ..... Create Z rotation matrix
  aafMat[0][0] =  COS(n32Z); aafMat[0][1] = SIN(n32Z); aafMat[0][2] = 0; aafMat[0][3] = 0;
  aafMat[1][0] = -SIN(n32Z); aafMat[1][1] = COS(n32Z); aafMat[1][2] = 0; aafMat[1][3] = 0;
  aafMat[2][0] = 0;          aafMat[2][1] = 0;         aafMat[2][2] = 1; aafMat[2][3] = 0;
  aafMat[3][0] = 0;          aafMat[3][1] = 0;         aafMat[3][2] = 0; aafMat[3][3] = 1;
  _MergeMatrices_(aafRMatrix, aafMat);
  
  // ..... Create X rotation matrix
  aafMat[0][0] = 1; aafMat[0][1] = 0;          aafMat[0][2] = 0;         aafMat[0][3] = 0;
  aafMat[1][0] = 0; aafMat[1][1] =  COS(n32X); aafMat[1][2] = SIN(n32X); aafMat[1][3] = 0;
  aafMat[2][0] = 0; aafMat[2][1] = -SIN(n32X); aafMat[2][2] = COS(n32X); aafMat[2][3] = 0;
  aafMat[3][0] = 0; aafMat[3][1] = 0;          aafMat[3][2] = 0;         aafMat[3][3] = 1;
  _MergeMatrices_(aafRMatrix, aafMat);

  // ..... Create Y rotation matrix
  aafMat[0][0] = COS(n32Y); aafMat[0][1] = 0; aafMat[0][2] = -SIN(n32Y); aafMat[0][3] = 0;
  aafMat[1][0] = 0;         aafMat[1][1] = 1; aafMat[1][2] = 0;          aafMat[1][3] = 0;
  aafMat[2][0] = SIN(n32Y); aafMat[2][1] = 0; aafMat[2][2] = COS(n32Y);  aafMat[2][3] = 0;
  aafMat[3][0] = 0;         aafMat[3][1] = 0; aafMat[3][2] = 0;          aafMat[3][3] = 1;
  _MergeMatrices_(aafRMatrix, aafMat);
  
  // ..... Merge with the main matrix
  _MergeMatrix_(aafRMatrix);
}
_vvoid _CVortexMatrix3D::RotateX(_vint32 n32X)
{
  // ..... Temporary variables
  _vfloat aafMat[4][4];

  // ..... Create and merge X rotation matrix
  aafMat[0][0] = 1; aafMat[0][1] = 0;          aafMat[0][2] = 0;         aafMat[0][3] = 0;
  aafMat[1][0] = 0; aafMat[1][1] =  COS(n32X); aafMat[1][2] = SIN(n32X); aafMat[1][3] = 0;
  aafMat[2][0] = 0; aafMat[2][1] = -SIN(n32X); aafMat[2][2] = COS(n32X); aafMat[2][3] = 0;
  aafMat[3][0] = 0; aafMat[3][1] = 0;          aafMat[3][2] = 0;         aafMat[3][3] = 1;
  _MergeMatrix_(aafMat);
}
_vvoid _CVortexMatrix3D::RotateY(_vint32 n32Y)
{
  // ..... Temporary variables
  _vfloat aafMat[4][4];

  // ..... Create and merge Y rotation matrix
  aafMat[0][0] = COS(n32Y); aafMat[0][1] = 0; aafMat[0][2] = -SIN(n32Y); aafMat[0][3] = 0;
  aafMat[1][0] = 0;         aafMat[1][1] = 1; aafMat[1][2] = 0;          aafMat[1][3] = 0;
  aafMat[2][0] = SIN(n32Y); aafMat[2][1] = 0; aafMat[2][2] = COS(n32Y);  aafMat[2][3] = 0;
  aafMat[3][0] = 0;         aafMat[3][1] = 0; aafMat[3][2] = 0;          aafMat[3][3] = 1;
  _MergeMatrix_(aafMat);
}
_vvoid _CVortexMatrix3D::RotateZ(_vint32 n32Z)
{
  // ..... Temporary variables
  _vfloat aafMat[4][4];

  // ..... Create and merge Z rotation matrix
  aafMat[0][0] =  COS(n32Z); aafMat[0][1] = SIN(n32Z); aafMat[0][2] = 0; aafMat[0][3] = 0;
  aafMat[1][0] = -SIN(n32Z); aafMat[1][1] = COS(n32Z); aafMat[1][2] = 0; aafMat[1][3] = 0;
  aafMat[2][0] = 0;          aafMat[2][1] = 0;         aafMat[2][2] = 1; aafMat[2][3] = 0;
  aafMat[3][0] = 0;          aafMat[3][1] = 0;         aafMat[3][2] = 0; aafMat[3][3] = 1;
  _MergeMatrix_(aafMat);
}
_vvoid _CVortexMatrix3D::Translate(_vfloat fX, _vfloat fY, _vfloat fZ)
{
  // ..... Temporary variables
  _vfloat aafMat[4][4];
  
  // ..... Create translation matrix  
  aafMat[0][0] = 1;  aafMat[0][1] = 0;  aafMat[0][2] = 0;  aafMat[0][3] = 0;
  aafMat[1][0] = 0;  aafMat[1][1] = 1;  aafMat[1][2] = 0;  aafMat[1][3] = 0;
  aafMat[2][0] = 0;  aafMat[2][1] = 0;  aafMat[2][2] = 1;  aafMat[2][3] = 0;
  aafMat[3][0] = fX; aafMat[3][1] = fY; aafMat[3][2] = fZ; aafMat[3][3] = 1;

  // ..... Merge with the main matrix
  _MergeMatrix_(aafMat);
}

// ============================================================================================
// End of vmat3d.cpp
// ============================================================================================
