// ############################################################################################
//  vmat3d.h
//  Definition & implementation of _CVortexMatrix3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VMATRIX3D_H_
#define _VMATRIX3D_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vvect3d.h>
#include <vpoint3d.h>

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The 3D matrix class
class _CVortexMatrix3D
{
  protected:
    // Variables
    _vfloat _sse mt_aafMatrix[4][4];  // The combined matrix data
    // Merged transformation operations
    inline _vvoid _MergeMatrix_(const _vfloat aafNewMatrix[4][4]);
    inline _vvoid _MergeMatrices_(_vfloat aafDstMatrix[4][4], const _vfloat aafSrcMatrix[4][4]);

  public:
    // Constructor
    inline _CVortexMatrix3D();
    // Display functions
    _vvoid DumpMatrix(_vvoid);
    // Transformation functions
    inline _TVortexVector3D &Transform(_TVortexVector3D &tVVector3D);
    inline _CVortexPoint3D &Transform(_CVortexPoint3D &csVPoint3D);
    inline _CVortexPoint3D &PreDynTransform(_CVortexPoint3D &csVPoint3D);
    inline _CVortexPoint3D &PstDynTransform(_CVortexPoint3D &csVPoint3D);
    // Set operations
    _vvoid SetToLookAtMatrix(const _vfloat afCamPosXYZ[3], const _vfloat afCamTargetXYZ[3]);
    _vvoid SetToLookAtMatrix(const _vfloat afCamPosXYZ[3], const _vfloat afCamTargetXYZ[3], const _vfloat afUpVectXYZ[3]);
    _vvoid SetToShearingShear(_vfloat fX, _vfloat fY);
    _vvoid SetToScallingMatrix(_vfloat fX, _vfloat fY, _vfloat fZ);
    _vvoid SetToRotatationMatrix(_vint32 n32X, _vint32 n32Y, _vint32 n32Z);
    _vvoid SetToTranslationMatrix(_vfloat fX, _vfloat fY, _vfloat fZ);
    inline _vvoid Identity(_vvoid);
    inline _vvoid SetIdentity(_vfloat aafDstMatrix[4][4]);
    inline _vvoid SetIdentity(_CVortexMatrix3D &csVDstMatrix);
    inline _vvoid SetMatrix(_vfloat const aafNewMatrix[4][4]);
    inline _vvoid SetMatrix(_CVortexMatrix3D &csVNewMatrix);
    // Merged transformation operations  
    _vvoid Shear(_vfloat fX, _vfloat fY);
    _vvoid Scale(_vfloat fX, _vfloat fY, _vfloat fZ);
    _vvoid Rotate(_vint32 n32X, _vint32 n32Y, _vint32 n32Z);
    _vvoid RotateX(_vint32 n32X);
    _vvoid RotateY(_vint32 n32Y);
    _vvoid RotateZ(_vint32 n32Z);
    _vvoid Translate(_vfloat fX, _vfloat fY, _vfloat fZ);
    // Transpose operations
    inline _vvoid TransposeMatrices(_vfloat aafDstMatrix[4][4], const _vfloat aafSrcMatrix[4][4]);
    inline _vvoid TransposeMatrices(_CVortexMatrix3D &csVDstMatrix, _CVortexMatrix3D &csVSrcMatrix);
    // Copy operations
    inline _vfloat *GetMatrixArray(_vvoid);
    inline _vvoid CopyMatrices(_vfloat aafDstMatrix[4][4], const _vfloat aafSrcMatrix[4][4]);
    inline _vvoid CopyMatrices(_CVortexMatrix3D &csVDstMatrix, _CVortexMatrix3D &csVSrcMatrix);
    // Multiplication operations
    inline _vvoid MultMatrixCCN(const _vfloat aafNewMatrix[4][4]);
    inline _vvoid MultMatrixCCN(_CVortexMatrix3D &csVNewMatrix);
    inline _vvoid MultMatrixCNC(const _vfloat aafNewMatrix[4][4]);
    inline _vvoid MultMatrixCNC(_CVortexMatrix3D &csVNewMatrix);
    inline _vvoid MultMatricesDDS(_vfloat aafDstMatrix[4][4], const _vfloat aafSrcMatrix[4][4]);
    inline _vvoid MultMatricesDDS(_CVortexMatrix3D &csVDstMatrix, _CVortexMatrix3D &csVSrcMatrix);
    inline _vvoid MultMatricesDSD(_vfloat aafDstMatrix[4][4], const _vfloat aafSrcMatrix[4][4]);
    inline _vvoid MultMatricesDSD(_CVortexMatrix3D &csVDstMatrix, _CVortexMatrix3D &csVSrcMatrix);
    // Special overloaded operators (to support aligned memory allocation for SSE)
    inline _vvoid *operator new      (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
    inline _vvoid *operator new[]    (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
    inline _vvoid  operator delete   (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }
    inline _vvoid  operator delete[] (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 3D matrix class
// ***** Constructor
_CVortexMatrix3D::_CVortexMatrix3D()
{
}

// ***** Transformation functions
_TVortexVector3D &_CVortexMatrix3D::Transform(_TVortexVector3D &tVVector3D)
{
  // ..... Temporary variables
  _vfloat fX = tVVector3D.fXt;
  _vfloat fY = tVVector3D.fYt;
  _vfloat fZ = tVVector3D.fZt;

  // ..... Transfor vector
  tVVector3D.fXt = (fX * mt_aafMatrix[0][0]) +
                   (fY * mt_aafMatrix[1][0]) +
                   (fZ * mt_aafMatrix[2][0]) +
                   mt_aafMatrix[3][0];
  tVVector3D.fYt = (fX * mt_aafMatrix[0][1]) +
                   (fY * mt_aafMatrix[1][1]) +
                   (fZ * mt_aafMatrix[2][1]) +
                   mt_aafMatrix[3][1];
  tVVector3D.fZt = (fX * mt_aafMatrix[0][2]) +
                   (fY * mt_aafMatrix[1][2]) +
                   (fZ * mt_aafMatrix[2][2]) +
                   mt_aafMatrix[3][2];
  return(tVVector3D);
}
_CVortexPoint3D &_CVortexMatrix3D::Transform(_CVortexPoint3D &csVPoint3D)
{
  csVPoint3D.m_tWorld.fX = (csVPoint3D.m_tLocal.fX * mt_aafMatrix[0][0]) +
                           (csVPoint3D.m_tLocal.fY * mt_aafMatrix[1][0]) +
                           (csVPoint3D.m_tLocal.fZ * mt_aafMatrix[2][0]) +
                           mt_aafMatrix[3][0];
  csVPoint3D.m_tWorld.fY = (csVPoint3D.m_tLocal.fX * mt_aafMatrix[0][1]) +
                           (csVPoint3D.m_tLocal.fY * mt_aafMatrix[1][1]) +
                           (csVPoint3D.m_tLocal.fZ * mt_aafMatrix[2][1]) +
                           mt_aafMatrix[3][1];
  csVPoint3D.m_tWorld.fZ = (csVPoint3D.m_tLocal.fX * mt_aafMatrix[0][2]) +
                           (csVPoint3D.m_tLocal.fY * mt_aafMatrix[1][2]) +
                           (csVPoint3D.m_tLocal.fZ * mt_aafMatrix[2][2]) +
                           mt_aafMatrix[3][2];
  return(csVPoint3D);
}
_CVortexPoint3D &_CVortexMatrix3D::PreDynTransform(_CVortexPoint3D &csVPoint3D)
{
  return(Transform(csVPoint3D));
}
_CVortexPoint3D &_CVortexMatrix3D::PstDynTransform(_CVortexPoint3D &csVPoint3D)
{
  // ..... Temporary variables
  _vfloat fX = csVPoint3D.m_tWorld.fX;
  _vfloat fY = csVPoint3D.m_tWorld.fY;
  _vfloat fZ = csVPoint3D.m_tWorld.fZ;

  // ..... Transfor vector
  csVPoint3D.m_tWorld.fX = (fX * mt_aafMatrix[0][0]) +
                           (fY * mt_aafMatrix[1][0]) +
                           (fZ * mt_aafMatrix[2][0]) +
                           mt_aafMatrix[3][0];
  csVPoint3D.m_tWorld.fY = (fX * mt_aafMatrix[0][1]) +
                           (fY * mt_aafMatrix[1][1]) +
                           (fZ * mt_aafMatrix[2][1]) +
                           mt_aafMatrix[3][1];
  csVPoint3D.m_tWorld.fZ = (fX * mt_aafMatrix[0][2]) +
                           (fY * mt_aafMatrix[1][2]) +
                           (fZ * mt_aafMatrix[2][2]) +
                           mt_aafMatrix[3][2];
  return(csVPoint3D);
}

// ***** Set operations
_vvoid _CVortexMatrix3D::Identity(_vvoid)
{
  SetIdentity(mt_aafMatrix);
}
_vvoid _CVortexMatrix3D::SetIdentity(_vfloat aafDstMatrix[4][4])
{
#ifdef _ASM_SSE_
  static _vfloat _sse fOne = _vfloat(1.0);
  _vuint32 edi;
  __asm__ __volatile__
  (
    //                                   XPOS - 0 1 2 3
    "xorps  %%xmm0, %%xmm0        \n" // xmm0 = 0 0 0 0
    "movss (%%esi), %%xmm0        \n" // xmm0 = 1 0 0 0
    "movaps %%xmm0, %%xmm1        \n" // xmm1 = 1 0 0 0
    "movaps %%xmm0, %%xmm2        \n" // xmm2 = 1 0 0 0
    "movaps %%xmm0, %%xmm3        \n" // xmm3 = 1 0 0 0
    "shufps $0xF1, %%xmm1, %%xmm1 \n" // xmm1 = 0 1 0 0
    "shufps $0x4F, %%xmm2, %%xmm2 \n" // xmm2 = 0 0 1 0
    "shufps $0x1F, %%xmm3, %%xmm3 \n" // xmm3 = 0 0 0 1
    "movaps %%xmm0,   (%%edi)     \n" // ROW0 = 1 0 0 0
    "movaps %%xmm1, 16(%%edi)     \n" // ROW1 = 0 1 0 0
    "movaps %%xmm2, 32(%%edi)     \n" // ROW2 = 0 0 1 0
    "movaps %%xmm3, 48(%%edi)     \n" // ROW3 = 0 0 0 1
    : "=D" (edi)                       \
    :  "0" (aafDstMatrix), "S" (&fOne) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "memory"
  );
#elif defined(_ASM_MMX_)
  static _vfloat afOZZO[4] = {1, 0, 0, 1};
  _vuint32 edi;
  __asm__ __volatile__
  (
    "movq (%%esi), %%mm0   \n" //  mm0 = 1 0
    "pxor %%mm1, %%mm1     \n" //  mm1 = 0 0
    "movq 8(%%esi), %%mm2  \n" //  mm2 = 0 1
    "movq %%mm0,   (%%edi) \n" // ROW0 = 1 0 X X
    "movq %%mm1,  8(%%edi) \n" // ROW0 = 1 0 0 0
    "movq %%mm2, 16(%%edi) \n" // ROW1 = 0 1 X X
    "movq %%mm1, 24(%%edi) \n" // ROW1 = 0 1 0 0
    "movq %%mm1, 32(%%edi) \n" // ROW2 = 0 0 X X
    "movq %%mm0, 40(%%edi) \n" // ROW2 = 0 0 1 0
    "movq %%mm1, 48(%%edi) \n" // ROW3 = 0 0 X X
    "movq %%mm2, 56(%%edi) \n" // ROW3 = 0 0 0 1
    "emms                  \n"
    : "=D" (edi)                        \
    :  "0" (aafDstMatrix), "S" (afOZZO) \
    : "mm0", "mm1", "mm2", "memory"
  );
#else
  aafDstMatrix[0][0] = 1; aafDstMatrix[0][1] = 0; aafDstMatrix[0][2] = 0; aafDstMatrix[0][3] = 0;
  aafDstMatrix[1][0] = 0; aafDstMatrix[1][1] = 1; aafDstMatrix[1][2] = 0; aafDstMatrix[1][3] = 0;
  aafDstMatrix[2][0] = 0; aafDstMatrix[2][1] = 0; aafDstMatrix[2][2] = 1; aafDstMatrix[2][3] = 0;
  aafDstMatrix[3][0] = 0; aafDstMatrix[3][1] = 0; aafDstMatrix[3][2] = 0; aafDstMatrix[3][3] = 1;
#endif
}
_vvoid _CVortexMatrix3D::SetIdentity(_CVortexMatrix3D &csVDstMatrix)
{
  SetIdentity(csVDstMatrix.mt_aafMatrix);
}
_vvoid _CVortexMatrix3D::SetMatrix(_vfloat const aafNewMatrix[4][4])
{
  CopyMatrices(mt_aafMatrix, aafNewMatrix);
}
_vvoid _CVortexMatrix3D::SetMatrix(_CVortexMatrix3D &csVNewMatrix)
{
  SetMatrix(csVNewMatrix.mt_aafMatrix);
}

// ***** Merged transformation operations  
_vvoid _CVortexMatrix3D::_MergeMatrix_(const _vfloat aafNewMatrix[4][4])
{
  MultMatrixCCN(aafNewMatrix);
}
_vvoid _CVortexMatrix3D::_MergeMatrices_(_vfloat aafDstMatrix[4][4], const _vfloat aafSrcMatrix[4][4])
{
  MultMatricesDSD(aafDstMatrix, aafSrcMatrix);
}

// ***** Transpose operations
_vvoid _CVortexMatrix3D::TransposeMatrices(_vfloat aafDstMatrix[4][4], const _vfloat aafSrcMatrix[4][4])
{
#ifdef _ASM_SSE_  
  _vuint32 edi;
  __asm__ __volatile__
  (
    //                                   MCOL - C0 C1 C2 C3
    // Copy from buffer                  XPOS - F0 F1 F2 F3
    "movaps 16(%%esi), %%xmm1     \n" // xmm1 = 10 11 12 13 (ROW 1)
    "movaps %%xmm1, %%xmm4        \n" // xmm4 = xmm1
    "movaps (%%esi), %%xmm0       \n" // xmm0 = 00 01 02 03 (ROW 0)
    "shufps $0x11, %%xmm0, %%xmm1 \n" // xmm1 = 11 10 01 00 (mask = 0101)
    "shufps $0xBB, %%xmm0, %%xmm4 \n" // xmm4 = 13 12 03 02 (mask = 2323)
    "movaps 48(%%esi), %%xmm3     \n" // xmm3 = 30 31 32 33 (ROW 3)
    "movaps %%xmm3, %%xmm5        \n" // xmm5 = xmm3
    "movaps 32(%%esi), %%xmm2     \n" // xmm2 = 20 21 22 23 (ROW 2)
    "shufps $0x11, %%xmm2, %%xmm3 \n" // xmm3 = 31 30 21 20 (mask = 0101)
    "shufps $0xBB, %%xmm2, %%xmm5 \n" // xmm5 = 33 32 23 22 (mask = 2323)
    "movaps %%xmm1, %%xmm2        \n" // xmm2 = xmm1
    "shufps $0x77, %%xmm3, %%xmm1 \n" // xmm1 = 00 10 20 30 (mask = 1313)
    "movaps %%xmm1, (%%edi)       \n" // ROW0 = 00 10 20 30
    "shufps $0x22, %%xmm3, %%xmm2 \n" // xmm2 = 01 11 21 31 (mask = 0202)
    "movaps %%xmm2, 16(%%edi)     \n" // ROW1 = 01 11 21 31
    "movaps %%xmm4, %%xmm0        \n" // xmm0 = xmm4
    "shufps $0x77, %%xmm5, %%xmm4 \n" // xmm4 = 02 12 22 32 (mask = 1313)
    "movaps %%xmm4, 32(%%edi)     \n" // ROW2 = 02 12 22 32
    "shufps $0x22, %%xmm5, %%xmm0 \n" // xmm0 = 03 13 23 33 (mask = 0202)
    "movaps %%xmm0, 48(%%edi)     \n" // ROW3 = 03 13 23 33
    : "=D" (edi)                              \
    :  "0" (aafDstMatrix), "S" (aafSrcMatrix) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "memory"
  );
#else
  aafDstMatrix[0][0] = aafSrcMatrix[0][0]; aafDstMatrix[0][1] = aafSrcMatrix[1][0]; aafDstMatrix[0][2] = aafSrcMatrix[2][0]; aafDstMatrix[0][3] = aafSrcMatrix[3][0];
  aafDstMatrix[1][0] = aafSrcMatrix[0][1]; aafDstMatrix[1][1] = aafSrcMatrix[1][1]; aafDstMatrix[1][2] = aafSrcMatrix[2][1]; aafDstMatrix[1][3] = aafSrcMatrix[3][1];
  aafDstMatrix[2][0] = aafSrcMatrix[0][2]; aafDstMatrix[2][1] = aafSrcMatrix[1][2]; aafDstMatrix[2][2] = aafSrcMatrix[2][2]; aafDstMatrix[2][3] = aafSrcMatrix[3][2];
  aafDstMatrix[3][0] = aafSrcMatrix[0][3]; aafDstMatrix[3][1] = aafSrcMatrix[1][3]; aafDstMatrix[3][2] = aafSrcMatrix[2][3]; aafDstMatrix[3][3] = aafSrcMatrix[3][3];
#endif
}
_vvoid _CVortexMatrix3D::TransposeMatrices(_CVortexMatrix3D &csVDstMatrix, _CVortexMatrix3D &csVSrcMatrix)
{
  TransposeMatrices(csVDstMatrix.mt_aafMatrix, csVSrcMatrix.mt_aafMatrix);
}

// ***** Copy operations
_vfloat *_CVortexMatrix3D::GetMatrixArray(_vvoid)
{
  return(&mt_aafMatrix[0][0]);
}
_vvoid _CVortexMatrix3D::CopyMatrices(_vfloat aafDstMatrix[4][4], const _vfloat aafSrcMatrix[4][4])
{
#ifdef _ASM_SSE_
  _vuint32 edi;
  __asm__ __volatile__
  (
    "movaps   (%%esi), %%xmm0 \n"
    "movaps 16(%%esi), %%xmm1 \n"
    "movaps 32(%%esi), %%xmm2 \n"
    "movaps 48(%%esi), %%xmm3 \n"
    "movaps %%xmm0,   (%%edi) \n"
    "movaps %%xmm1, 16(%%edi) \n"
    "movaps %%xmm2, 32(%%edi) \n"
    "movaps %%xmm3, 48(%%edi) \n"
    : "=D" (edi)                              \
    :  "0" (aafDstMatrix), "S" (aafSrcMatrix) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "memory"
  );
#elif defined(_ASM_MMX_)
  _vuint32 edi;
  __asm__ __volatile__
  (
    "movq   (%%esi), %%mm0 \n"
    "movq  8(%%esi), %%mm1 \n"
    "movq 16(%%esi), %%mm2 \n"
    "movq 24(%%esi), %%mm3 \n"
    "movq 32(%%esi), %%mm4 \n"
    "movq 40(%%esi), %%mm5 \n"
    "movq 48(%%esi), %%mm6 \n"
    "movq 56(%%esi), %%mm7 \n"
    "movq %%mm0,   (%%edi) \n"
    "movq %%mm1,  8(%%edi) \n"
    "movq %%mm2, 16(%%edi) \n"
    "movq %%mm3, 24(%%edi) \n"
    "movq %%mm4, 32(%%edi) \n"
    "movq %%mm5, 40(%%edi) \n"
    "movq %%mm6, 48(%%edi) \n"
    "movq %%mm7, 56(%%edi) \n"
    "emms                  \n"
    : "=D" (edi)                              \
    :  "0" (aafDstMatrix), "S" (aafSrcMatrix) \
    : "mm0", "mm1", "mm2", "mm3", "mm4", "mm5", "mm6", "mm7", "memory"
  );
#else
  aafDstMatrix[0][0] = aafSrcMatrix[0][0]; aafDstMatrix[0][1] = aafSrcMatrix[0][1]; aafDstMatrix[0][2] = aafSrcMatrix[0][2]; aafDstMatrix[0][3] = aafSrcMatrix[0][3];
  aafDstMatrix[1][0] = aafSrcMatrix[1][0]; aafDstMatrix[1][1] = aafSrcMatrix[1][1]; aafDstMatrix[1][2] = aafSrcMatrix[1][2]; aafDstMatrix[1][3] = aafSrcMatrix[1][3];
  aafDstMatrix[2][0] = aafSrcMatrix[2][0]; aafDstMatrix[2][1] = aafSrcMatrix[2][1]; aafDstMatrix[2][2] = aafSrcMatrix[2][2]; aafDstMatrix[2][3] = aafSrcMatrix[2][3];
  aafDstMatrix[3][0] = aafSrcMatrix[3][0]; aafDstMatrix[3][1] = aafSrcMatrix[3][1]; aafDstMatrix[3][2] = aafSrcMatrix[3][2]; aafDstMatrix[3][3] = aafSrcMatrix[3][3];
#endif
}
_vvoid _CVortexMatrix3D::CopyMatrices(_CVortexMatrix3D &csVDstMatrix, _CVortexMatrix3D &csVSrcMatrix)
{
  CopyMatrices(csVDstMatrix.mt_aafMatrix, csVSrcMatrix.mt_aafMatrix);
}

// ***** Multiplication operations
_vvoid _CVortexMatrix3D::MultMatrixCCN(const _vfloat aafNewMatrix[4][4])
{
  MultMatricesDDS(mt_aafMatrix, aafNewMatrix);
}
_vvoid _CVortexMatrix3D::MultMatrixCCN(_CVortexMatrix3D &csVNewMatrix)
{
  MultMatrixCCN(csVNewMatrix.mt_aafMatrix);
}
_vvoid _CVortexMatrix3D::MultMatrixCNC(const _vfloat aafNewMatrix[4][4])
{
  MultMatricesDSD(mt_aafMatrix, aafNewMatrix);
}
_vvoid _CVortexMatrix3D::MultMatrixCNC(_CVortexMatrix3D &csVNewMatrix)
{
  MultMatrixCNC(csVNewMatrix.mt_aafMatrix);
}
_vvoid _CVortexMatrix3D::MultMatricesDDS(_vfloat aafDstMatrix[4][4], const _vfloat aafSrcMatrix[4][4])
{
#ifdef _ASM_SSE_
  static _vfloat _sse aafTmp[3][4];
  _vuint32 ebx, edi;
  __asm__ __volatile__
  (
    // Load the source matrix to register
    //                                           XPOS - F0 F1 F2 F3
    "movaps   (%%esi), %%xmm4         \n" // xmm4 = src[00 01 02 03]
    "movaps 16(%%esi), %%xmm5         \n" // xmm5 = src[10 11 12 13]
    "movaps 32(%%esi), %%xmm6         \n" // xmm6 = src[20 21 22 23]
    "movaps 48(%%esi), %%xmm7         \n" // xmm7 = src[30 31 32 33]
    // Multiply the row0
    "movss    (%%edi), %%xmm0         \n" // xmm0 = dst[00 XX XX XX]
    "shufps     $0x00, %%xmm0, %%xmm0 \n" // xmm0 = dst[00 00 00 00]
    "movss   4(%%edi), %%xmm1         \n" // xmm1 = dst[01 XX XX XX]
    "shufps     $0x00, %%xmm1, %%xmm1 \n" // xmm1 = dst[01 01 01 01]
    "movss   8(%%edi), %%xmm2         \n" // xmm2 = dst[02 XX XX XX]
    "shufps     $0x00, %%xmm2, %%xmm2 \n" // xmm2 = dst[02 02 02 02]
    "movss  12(%%edi), %%xmm3         \n" // xmm3 = dst[03 XX XX XX]
    "shufps     $0x00, %%xmm3, %%xmm3 \n" // xmm3 = dst[03 03 03 03]
    "mulps     %%xmm4, %%xmm0         \n" // xmm0 = dst[00 00 00 00] X src[00 01 02 03]
    "mulps     %%xmm5, %%xmm1         \n" // xmm1 = dst[01 01 01 01] X src[10 11 12 13]
    "addps     %%xmm1, %%xmm0         \n" // xmm0 = xmm0 + xmm1
    "mulps     %%xmm6, %%xmm2         \n" // xmm2 = dst[02 02 02 02] X src[20 21 22 23]
    "mulps     %%xmm7, %%xmm3         \n" // xmm3 = dst[03 03 03 03] X src[30 31 32 33]
    "addps     %%xmm3, %%xmm2         \n" // xmm2 = xmm2 + xmm3
    "addps     %%xmm2, %%xmm0         \n" // xmm0 = dst[row0] X src
    "movaps    %%xmm0, (%%ebx)        \n" // save row0 result to temporary matrix
    // Multiply the row1
    "movss  16(%%edi), %%xmm0         \n" // xmm0 = dst[10 XX XX XX]
    "shufps     $0x00, %%xmm0, %%xmm0 \n" // xmm0 = dst[10 10 10 10]
    "movss  20(%%edi), %%xmm1         \n" // xmm1 = dst[11 XX XX XX]
    "shufps     $0x00, %%xmm1, %%xmm1 \n" // xmm1 = dst[11 11 11 11]
    "movss  24(%%edi), %%xmm2         \n" // xmm2 = dst[12 XX XX XX]
    "shufps     $0x00, %%xmm2, %%xmm2 \n" // xmm2 = dst[12 12 12 12]
    "movss  28(%%edi), %%xmm3         \n" // xmm3 = dst[13 XX XX XX]
    "shufps     $0x00, %%xmm3, %%xmm3 \n" // xmm3 = dst[13 13 13 13]
    "mulps     %%xmm4, %%xmm0         \n" // xmm0 = dst[10 10 10 10] X src[00 01 02 03]
    "mulps     %%xmm5, %%xmm1         \n" // xmm1 = dst[11 11 11 11] X src[10 11 12 13]
    "addps     %%xmm1, %%xmm0         \n" // xmm0 = xmm0 + xmm1
    "mulps     %%xmm6, %%xmm2         \n" // xmm2 = dst[12 12 12 12] X src[20 21 22 23]
    "mulps     %%xmm7, %%xmm3         \n" // xmm3 = dst[13 13 13 13] X src[30 31 32 33]
    "addps     %%xmm3, %%xmm2         \n" // xmm2 = xmm2 + xmm3
    "addps     %%xmm2, %%xmm0         \n" // xmm0 = dst[row1] X src
    "movaps    %%xmm0, 16(%%ebx)      \n" // save row1 result to temporary matrix
    // Multiply the row2
    "movss  32(%%edi), %%xmm0         \n" // xmm0 = dst[20 XX XX XX]
    "shufps     $0x00, %%xmm0, %%xmm0 \n" // xmm0 = dst[20 20 20 20]
    "movss  36(%%edi), %%xmm1         \n" // xmm1 = dst[21 XX XX XX]
    "shufps     $0x00, %%xmm1, %%xmm1 \n" // xmm1 = dst[21 21 21 21]
    "movss  40(%%edi), %%xmm2         \n" // xmm2 = dst[22 XX XX XX]
    "shufps     $0x00, %%xmm2, %%xmm2 \n" // xmm2 = dst[22 22 22 22]
    "movss  44(%%edi), %%xmm3         \n" // xmm3 = dst[23 XX XX XX]
    "shufps     $0x00, %%xmm3, %%xmm3 \n" // xmm3 = dst[23 23 23 23]
    "mulps     %%xmm4, %%xmm0         \n" // xmm0 = dst[20 20 20 20] X src[00 01 02 03]
    "mulps     %%xmm5, %%xmm1         \n" // xmm1 = dst[21 21 21 21] X src[10 11 12 13]
    "addps     %%xmm1, %%xmm0         \n" // xmm0 = xmm0 + xmm1
    "mulps     %%xmm6, %%xmm2         \n" // xmm2 = dst[22 22 22 22] X src[20 21 22 23]
    "mulps     %%xmm7, %%xmm3         \n" // xmm3 = dst[23 23 23 23] X src[30 31 32 33]
    "addps     %%xmm3, %%xmm2         \n" // xmm2 = xmm2 + xmm3
    "addps     %%xmm2, %%xmm0         \n" // xmm0 = dst[row2] X src
    "movaps    %%xmm0, 32(%%ebx)      \n" // save row2 result to temporary matrix
    // Multiply the row3
    "movss  48(%%edi), %%xmm0         \n" // xmm0 = dst[30 XX XX XX]
    "shufps     $0x00, %%xmm0, %%xmm0 \n" // xmm0 = dst[30 30 30 30]
    "movss  52(%%edi), %%xmm1         \n" // xmm1 = dst[31 XX XX XX]
    "shufps     $0x00, %%xmm1, %%xmm1 \n" // xmm1 = dst[31 31 31 31]
    "movss  56(%%edi), %%xmm2         \n" // xmm2 = dst[32 XX XX XX]
    "shufps     $0x00, %%xmm2, %%xmm2 \n" // xmm2 = dst[32 32 32 32]
    "movss  60(%%edi), %%xmm3         \n" // xmm3 = dst[33 XX XX XX]
    "shufps     $0x00, %%xmm3, %%xmm3 \n" // xmm3 = dst[33 33 33 33]
    "mulps     %%xmm4, %%xmm0         \n" // xmm0 = dst[30 30 30 30] X src[00 01 02 03]
    "mulps     %%xmm5, %%xmm1         \n" // xmm1 = dst[31 31 31 31] X src[10 11 12 13]
    "addps     %%xmm1, %%xmm0         \n" // xmm0 = xmm0 + xmm1
    "mulps     %%xmm6, %%xmm2         \n" // xmm2 = dst[32 32 32 32] X src[20 21 22 23]
    "mulps     %%xmm7, %%xmm3         \n" // xmm3 = dst[33 33 33 33] X src[30 31 32 33]
    "addps     %%xmm3, %%xmm2         \n" // xmm2 = xmm2 + xmm3
    "addps     %%xmm2, %%xmm0         \n" // xmm0 = dst[row3] X src
    // Copy temporary matrix to destination matrix
    "movaps   (%%ebx), %%xmm1         \n" // xmm0 = tmp0
    "movaps 16(%%ebx), %%xmm2         \n" // xmm1 = tmp1
    "movaps 32(%%ebx), %%xmm3         \n" // xmm2 = tmp2
    "movaps %%xmm1,   (%%edi)         \n" // dst0 = tmp0
    "movaps %%xmm2, 16(%%edi)         \n" // dst1 = tmp1
    "movaps %%xmm3, 32(%%edi)         \n" // dst2 = tmp2
    "movaps %%xmm0, 48(%%edi)         \n" // dst3 = tmp3
    : "=b" (ebx),   "=D" (edi)                              \
    :  "0" (aafTmp), "1" (aafDstMatrix), "S" (aafSrcMatrix) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6", "xmm7", "memory"
  );
#else  
  // ..... Temporary variables
  _vfloat aafTmp[4][4];
  
  // ..... Do multiplication
  for(_vuint8 un8I = 0; un8I < 4; un8I++)
    for(_vuint8 un8J = 0; un8J < 4; un8J++)
       aafTmp[un8I][un8J] = (aafDstMatrix[un8I][0] * aafSrcMatrix[0][un8J]) +
                            (aafDstMatrix[un8I][1] * aafSrcMatrix[1][un8J]) +
                            (aafDstMatrix[un8I][2] * aafSrcMatrix[2][un8J]) +
                            (aafDstMatrix[un8I][3] * aafSrcMatrix[3][un8J]);
  CopyMatrices(aafDstMatrix, aafTmp);
#endif    
}
_vvoid _CVortexMatrix3D::MultMatricesDDS(_CVortexMatrix3D &csVDstMatrix, _CVortexMatrix3D &csVSrcMatrix)
{
  MultMatricesDDS(csVDstMatrix.mt_aafMatrix, csVSrcMatrix.mt_aafMatrix);
}
_vvoid _CVortexMatrix3D::MultMatricesDSD(_vfloat aafDstMatrix[4][4], const _vfloat aafSrcMatrix[4][4])
{
#ifdef _ASM_SSE_
  static _vfloat _sse aafTmp[3][4];
  _vuint32 ebx, edi;
  __asm__ __volatile__
  (
    // Load the destination matrix to register
    //                                           XPOS - F0 F1 F2 F3
    "movaps   (%%edi), %%xmm4         \n" // xmm4 = dst[00 01 02 03]
    "movaps 16(%%edi), %%xmm5         \n" // xmm5 = dst[10 11 12 13]
    "movaps 32(%%edi), %%xmm6         \n" // xmm6 = dst[20 21 22 23]
    "movaps 48(%%edi), %%xmm7         \n" // xmm7 = dst[30 31 32 33]
    // Multiply the row0
    "movss    (%%esi), %%xmm0         \n" // xmm0 = src[00 XX XX XX]
    "shufps     $0x00, %%xmm0, %%xmm0 \n" // xmm0 = src[00 00 00 00]
    "movss   4(%%esi), %%xmm1         \n" // xmm1 = src[01 XX XX XX]
    "shufps     $0x00, %%xmm1, %%xmm1 \n" // xmm1 = src[01 01 01 01]
    "movss   8(%%esi), %%xmm2         \n" // xmm2 = src[02 XX XX XX]
    "shufps     $0x00, %%xmm2, %%xmm2 \n" // xmm2 = src[02 02 02 02]
    "movss  12(%%esi), %%xmm3         \n" // xmm3 = src[03 XX XX XX]
    "shufps     $0x00, %%xmm3, %%xmm3 \n" // xmm3 = src[03 03 03 03]
    "mulps     %%xmm4, %%xmm0         \n" // xmm0 = src[00 00 00 00] X dst[00 01 02 03]
    "mulps     %%xmm5, %%xmm1         \n" // xmm1 = src[01 01 01 01] X dst[10 11 12 13]
    "addps     %%xmm1, %%xmm0         \n" // xmm0 = xmm0 + xmm1
    "mulps     %%xmm6, %%xmm2         \n" // xmm2 = src[02 02 02 02] X dst[20 21 22 23]
    "mulps     %%xmm7, %%xmm3         \n" // xmm3 = src[03 03 03 03] X dst[30 31 32 33]
    "addps     %%xmm3, %%xmm2         \n" // xmm2 = xmm2 + xmm3
    "addps     %%xmm2, %%xmm0         \n" // xmm0 = src[row0] X dst
    "movaps    %%xmm0, (%%ebx)        \n" // save row0 result to temporary matrix
    // Multiply the row1
    "movss  16(%%esi), %%xmm0         \n" // xmm0 = src[10 XX XX XX]
    "shufps     $0x00, %%xmm0, %%xmm0 \n" // xmm0 = src[10 10 10 10]
    "movss  20(%%esi), %%xmm1         \n" // xmm1 = src[11 XX XX XX]
    "shufps     $0x00, %%xmm1, %%xmm1 \n" // xmm1 = src[11 11 11 11]
    "movss  24(%%esi), %%xmm2         \n" // xmm2 = src[12 XX XX XX]
    "shufps     $0x00, %%xmm2, %%xmm2 \n" // xmm2 = src[12 12 12 12]
    "movss  28(%%esi), %%xmm3         \n" // xmm3 = src[13 XX XX XX]
    "shufps     $0x00, %%xmm3, %%xmm3 \n" // xmm3 = src[13 13 13 13]
    "mulps     %%xmm4, %%xmm0         \n" // xmm0 = src[10 10 10 10] X dst[00 01 02 03]
    "mulps     %%xmm5, %%xmm1         \n" // xmm1 = src[11 11 11 11] X dst[10 11 12 13]
    "addps     %%xmm1, %%xmm0         \n" // xmm0 = xmm0 + xmm1
    "mulps     %%xmm6, %%xmm2         \n" // xmm2 = src[12 12 12 12] X dst[20 21 22 23]
    "mulps     %%xmm7, %%xmm3         \n" // xmm3 = src[13 13 13 13] X dst[30 31 32 33]
    "addps     %%xmm3, %%xmm2         \n" // xmm2 = xmm2 + xmm3
    "addps     %%xmm2, %%xmm0         \n" // xmm0 = src[row1] X dst
    "movaps    %%xmm0, 16(%%ebx)      \n" // save row1 result to temporary matrix
    // Multiply the row2
    "movss  32(%%esi), %%xmm0         \n" // xmm0 = src[20 XX XX XX]
    "shufps     $0x00, %%xmm0, %%xmm0 \n" // xmm0 = src[20 20 20 20]
    "movss  36(%%esi), %%xmm1         \n" // xmm1 = src[21 XX XX XX]
    "shufps     $0x00, %%xmm1, %%xmm1 \n" // xmm1 = src[21 21 21 21]
    "movss  40(%%esi), %%xmm2         \n" // xmm2 = src[22 XX XX XX]
    "shufps     $0x00, %%xmm2, %%xmm2 \n" // xmm2 = src[22 22 22 22]
    "movss  44(%%esi), %%xmm3         \n" // xmm3 = src[23 XX XX XX]
    "shufps     $0x00, %%xmm3, %%xmm3 \n" // xmm3 = src[23 23 23 23]
    "mulps     %%xmm4, %%xmm0         \n" // xmm0 = src[20 20 20 20] X dst[00 01 02 03]
    "mulps     %%xmm5, %%xmm1         \n" // xmm1 = src[21 21 21 21] X dst[10 11 12 13]
    "addps     %%xmm1, %%xmm0         \n" // xmm0 = xmm0 + xmm1
    "mulps     %%xmm6, %%xmm2         \n" // xmm2 = src[22 22 22 22] X dst[20 21 22 23]
    "mulps     %%xmm7, %%xmm3         \n" // xmm3 = src[23 23 23 23] X dst[30 31 32 33]
    "addps     %%xmm3, %%xmm2         \n" // xmm2 = xmm2 + xmm3
    "addps     %%xmm2, %%xmm0         \n" // xmm0 = src[row2] X dst
    "movaps    %%xmm0, 32(%%ebx)      \n" // save row2 result to temporary matrix
    // Multiply the row3
    "movss  48(%%esi), %%xmm0         \n" // xmm0 = src[30 XX XX XX]
    "shufps     $0x00, %%xmm0, %%xmm0 \n" // xmm0 = src[30 30 30 30]
    "movss  52(%%esi), %%xmm1         \n" // xmm1 = src[31 XX XX XX]
    "shufps     $0x00, %%xmm1, %%xmm1 \n" // xmm1 = src[31 31 31 31]
    "movss  56(%%esi), %%xmm2         \n" // xmm2 = src[32 XX XX XX]
    "shufps     $0x00, %%xmm2, %%xmm2 \n" // xmm2 = src[32 32 32 32]
    "movss  60(%%esi), %%xmm3         \n" // xmm3 = src[33 XX XX XX]
    "shufps     $0x00, %%xmm3, %%xmm3 \n" // xmm3 = src[33 33 33 33]
    "mulps     %%xmm4, %%xmm0         \n" // xmm0 = src[30 30 30 30] X dst[00 01 02 03]
    "mulps     %%xmm5, %%xmm1         \n" // xmm1 = src[31 31 31 31] X dst[10 11 12 13]
    "addps     %%xmm1, %%xmm0         \n" // xmm0 = xmm0 + xmm1
    "mulps     %%xmm6, %%xmm2         \n" // xmm2 = src[32 32 32 32] X dst[20 21 22 23]
    "mulps     %%xmm7, %%xmm3         \n" // xmm3 = src[33 33 33 33] X dst[30 31 32 33]
    "addps     %%xmm3, %%xmm2         \n" // xmm2 = xmm2 + xmm3
    "addps     %%xmm2, %%xmm0         \n" // xmm0 = src[row3] X dst
    // Copy temporary matrix to destination matrix
    "movaps   (%%ebx), %%xmm1         \n" // xmm0 = tmp0
    "movaps 16(%%ebx), %%xmm2         \n" // xmm1 = tmp1
    "movaps 32(%%ebx), %%xmm3         \n" // xmm2 = tmp2
    "movaps %%xmm1,   (%%edi)         \n" // dst0 = tmp0
    "movaps %%xmm2, 16(%%edi)         \n" // dst1 = tmp1
    "movaps %%xmm3, 32(%%edi)         \n" // dst2 = tmp2
    "movaps %%xmm0, 48(%%edi)         \n" // dst3 = tmp3
    : "=b" (ebx),   "=D" (edi)                              \
    :  "0" (aafTmp), "1" (aafDstMatrix), "S" (aafSrcMatrix) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6", "xmm7", "memory"
  );
#else
  // ..... Temporary variables
  _vfloat aafTmp[4][4];
  
  // ..... Do multiplication
  for(_vuint8 un8I = 0; un8I < 4; un8I++)
    for(_vuint8 un8J = 0; un8J < 4; un8J++)
       aafTmp[un8I][un8J] = (aafSrcMatrix[un8I][0] * aafDstMatrix[0][un8J]) +
                            (aafSrcMatrix[un8I][1] * aafDstMatrix[1][un8J]) +
                            (aafSrcMatrix[un8I][2] * aafDstMatrix[2][un8J]) +
                            (aafSrcMatrix[un8I][3] * aafDstMatrix[3][un8J]);
  CopyMatrices(aafDstMatrix, aafTmp);
#endif
}
_vvoid _CVortexMatrix3D::MultMatricesDSD(_CVortexMatrix3D &csVDstMatrix, _CVortexMatrix3D &csVSrcMatrix)
{
  MultMatricesDSD(csVDstMatrix.mt_aafMatrix, csVSrcMatrix.mt_aafMatrix);
}

#endif
// ============================================================================================
// End of vmat3d.h
// ============================================================================================
