// ############################################################################################
//  vmdisp.cpp
//  Implementation of console-message-displayer functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <stdio.h>
#include <stdarg.h>
// ***** VortexGE headers
#include <vmdisp.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** Buffer size
#define _EBUF_SIZE_ 1024
#define _WBUF_SIZE_ 1024
#define _IBUF_SIZE_ 1024
#define _PBUF_SIZE_ 1024

// ============================================================================================
// Local variables declaration
// ============================================================================================
// ***** Message displayer target
static _TVortexFILE *_ptInfo_    = stdout;
static _TVortexFILE *_ptWarning_ = stdout;
static _TVortexFILE *_ptError_   = stdout;

// ***** Message buffer
static _EVortexERRCode _eERRCode_;
static _vchar          _acEBuff_[_EBUF_SIZE_];
static _vchar          _acWBuff_[_WBUF_SIZE_];
static _vchar          _acIBuff_[_IBUF_SIZE_];
static _vchar          _acPBuff_[_PBUF_SIZE_];

// ============================================================================================
// Defines
// ============================================================================================
// ***** ANSI's escape command
#define _ESC_CLEAR_SCREEN   "\033[2J\033[0;0H"
#define _ESC_COLOR_NORMAL   "\033[0m"
#define _ESC_COLOR_FATAL    "\033[1;37;41m"
#define _ESC_COLOR_ERROR    "\033[1;31;40m"
#define _ESC_COLOR_WARNING  "\033[1;33;40m"
#define _ESC_COLOR_INFO     "\033[1;34;40m"
#define _ESC_COLOR_LOCATION "\033[1;32;40m"
#define _ESC_COLOR_LINE     "\033[1;35;40m"
#define _ESC_COLOR_DETAIL   "\033[1;37;40m"

// ============================================================================================
// Exported functions implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// Message displayer
// --------------------------------------------------------------------------------------------
// ***** Clear screen
_vvoid VClrSrc(_vvoid)
{
#ifdef __USE_COLORED_MESSAGES__
  fprintf(stdout, "%s", _ESC_CLEAR_SCREEN); fflush(stdout);
  fprintf(stderr, "%s", _ESC_CLEAR_SCREEN); fflush(stderr);
#else
  system("clear");
#endif  
}

// ***** Display fatal error message
_vvoid VFatalError(const _vchar *pcMsg, ...)
{
  // ..... Temporary variables
  va_list vlTmp;

  // ..... Extract the pcMsg
  va_start(vlTmp, pcMsg);
  (void) vsnprintf(_acEBuff_, _EBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);
  
  // ..... Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptError_, "%s\n%sFATAL ERROR:%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_NORMAL, _ESC_COLOR_FATAL, _ESC_COLOR_LINE, _ESC_COLOR_DETAIL, _acEBuff_, _ESC_COLOR_NORMAL);
#else
  fprintf(_ptError_, "\nFATAL ERROR:\n------------------------------------------------------------\n%s\n", _acEBuff_);
#endif  
}
_vvoid VFatalError(const _vchar *pcFile, _vint32 n32Line, const _vchar *pcMsg, ...)
{
  // ..... Temporary variables
  va_list vlTmp;
  
  // ..... Extract the pcMsg
  va_start(vlTmp, pcMsg);
  (void) vsnprintf(_acEBuff_, _EBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);
  
  // ..... Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptError_, "%s\n%sFATAL ERROR:%s %s, line %d%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_NORMAL, _ESC_COLOR_FATAL, _ESC_COLOR_LOCATION, pcFile, n32Line,
          _ESC_COLOR_LINE, _ESC_COLOR_DETAIL, _acEBuff_, _ESC_COLOR_NORMAL);
#else
  fprintf(_ptError_, "\nFATAL ERROR: %s, line %d\n------------------------------------------------------------\n%s\n", pcFile, n32Line, _acEBuff_);
#endif  
}

// ***** Display error message
_vvoid VError(const _vchar *pcMsg, ...)
{
  // ..... Temporary variables
  va_list vlTmp;
  
  // ..... Extract the pcMsg
  va_start(vlTmp, pcMsg);
  (void) vsnprintf(_acEBuff_, _EBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);
  
  // ..... Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptError_, "%s\n%sERROR:%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_NORMAL, _ESC_COLOR_ERROR, _ESC_COLOR_LINE, _ESC_COLOR_DETAIL, _acEBuff_, _ESC_COLOR_NORMAL);
#else
  fprintf(_ptError_, "\nERROR:\n------------------------------------------------------------\n%s\n", _acEBuff_);
#endif  
}
_vvoid VError(const _vchar *pcFile, _vint32 n32Line, const _vchar *pcMsg, ...)
{
  // ..... Temporary variables
  va_list vlTmp;
  
  // ..... Extract the pcMsg
  va_start(vlTmp, pcMsg);
  (void) vsnprintf(_acEBuff_, _EBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);
  
  // ..... Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptError_, "%s\n%sERROR:%s %s, line %d%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_NORMAL, _ESC_COLOR_ERROR, _ESC_COLOR_LOCATION, pcFile, n32Line,
          _ESC_COLOR_LINE, _ESC_COLOR_DETAIL, _acEBuff_, _ESC_COLOR_NORMAL);
#else
  fprintf(_ptError_, "\nERROR: %s, line %d\n------------------------------------------------------------\n%s\n", pcFile, n32Line, _acEBuff_);
#endif  
}

// ***** Display warning message
_vvoid VWarning(const _vchar *pcMsg, ...)
{
  // ..... Temporary variables
  va_list vlTmp;
  
  // ..... Extract the pcMsg
  va_start(vlTmp, pcMsg);
  (void) vsnprintf(_acWBuff_, _WBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);
  
  // ..... Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptWarning_, "%s\n%sWARNING:%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_NORMAL, _ESC_COLOR_WARNING, _ESC_COLOR_LINE, _ESC_COLOR_DETAIL, _acWBuff_, _ESC_COLOR_NORMAL);
#else
  fprintf(_ptWarning_, "\nWARNING:\n------------------------------------------------------------\n%s\n", _acWBuff_);
#endif  
}
_vvoid VWarning(const _vchar *pcFile, _vint32 n32Line, const _vchar *pcMsg, ...)
{
  // ..... Temporary variables
  va_list vlTmp;
  
  // ..... Extract the pcMsg
  va_start(vlTmp, pcMsg);
  (void) vsnprintf(_acWBuff_, _WBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);
  
  // ..... Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptWarning_, "%s\n%sWARNING:%s %s, line %d%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_NORMAL, _ESC_COLOR_WARNING, _ESC_COLOR_LOCATION, pcFile, n32Line,
          _ESC_COLOR_LINE, _ESC_COLOR_DETAIL, _acWBuff_, _ESC_COLOR_NORMAL);
#else
  fprintf(_ptWarning_, "\nWARNING: %s, line %d\n------------------------------------------------------------\n%s\n", pcFile, n32Line, _acWBuff_);
#endif  
}
  
// ***** Display warning message
_vvoid VInfo(const _vchar *pcMsg, ...)
{
  // ..... Temporary variables
  va_list vlTmp;
  
  // ..... Extract the pcMsg
  va_start(vlTmp, pcMsg);
  (void) vsnprintf(_acIBuff_, _IBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);
  
  // ..... Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptInfo_, "%s\n%sINFO:%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_NORMAL, _ESC_COLOR_INFO, _ESC_COLOR_LINE, _ESC_COLOR_DETAIL, _acIBuff_, _ESC_COLOR_NORMAL);
#else
  fprintf(_ptInfo_, "\nINFO:\n------------------------------------------------------------\n%s\n", _acIBuff_);
#endif  
}
_vvoid VInfo(const _vchar *pcFile, _vint32 n32Line, const _vchar *pcMsg, ...)
{
  // ..... Temporary variables
  va_list vlTmp;
  
  // ..... Extract the pcMsg
  va_start(vlTmp, pcMsg);
  (void) vsnprintf(_acIBuff_, _IBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);
  
  // ..... Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptInfo_, "%s\n%sINFO:%s %s, line %d%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_NORMAL, _ESC_COLOR_INFO, _ESC_COLOR_LOCATION, pcFile, n32Line,
          _ESC_COLOR_LINE, _ESC_COLOR_DETAIL, _acIBuff_, _ESC_COLOR_NORMAL);
#else
  fprintf(_ptInfo_, "\nINFO: %s, line %d\n------------------------------------------------------------\n%s\n", pcFile, n32Line, _acIBuff_);
#endif  
}

// ***** Display message
_vvoid VPrint(const _vchar *pcMsg, ...)
{
  // ..... Temporary variables
  va_list vlTmp;
  
  // ..... Extract the pcMsg
  va_start(vlTmp, pcMsg);
  (void) vsnprintf(_acPBuff_, _IBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);
  
  // ..... Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptInfo_, "%s%s%s", _ESC_COLOR_DETAIL, _acPBuff_, _ESC_COLOR_NORMAL);
#else
  fprintf(_ptInfo_, "%s", _acPBuff_);
#endif  
}

// ***** Redirect message
_vvoid VRedirectMsgs(_TVortexFILE *ptInfo, _TVortexFILE *ptWarning, _TVortexFILE *ptError)
{
  _ptInfo_    = ptInfo;
  _ptWarning_ = ptWarning;
  _ptError_   = ptError;
}

// --------------------------------------------------------------------------------------------
// Error message controller
// --------------------------------------------------------------------------------------------
_vvoid VClearError(_vvoid)
{
  _eERRCode_   = VERR_None;
  _acEBuff_[0] = 0;
}
_vvoid VSetErrorCode(_EVortexERRCode eERRCode)
{
  _eERRCode_ = eERRCode;
}
_EVortexERRCode VGetLastErrorCode(_vvoid)
{
  return(_eERRCode_);
}
const _vchar *GetLastErrorString(_vvoid)
{
  if(_acEBuff_[0]) return(_acEBuff_);
  return(NULL);
}

// ============================================================================================
// End of vmdisp.cpp
// ============================================================================================
