// ############################################################################################
//  vp3dbuff.h
//  Definition & implementation of _CVortexPanel3DBuffer class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VP3DBUFF_H_
#define _VP3DBUFF_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vmat3d.h>
#include <vpanel3d.h>
#include <vpoint3d.h>

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The 3D panel buffer class
class _CVortexPanel3DBuffer
{
  protected:
    // Variables
    _CVortexTexture *mt_pcsTexture;  // The panel's texture
    _CVortexPoint3D *mt_pcsVList;    // The list of vertices
    _CVortexPanel3D *mt_pcsPList;    // The list of panels
    _vuint32        mt_un32TWidth1;  // The texture's width - 1
    _vuint32        mt_un32THeight1; // The texture's height - 1
    _vuint32        mt_un32VCount;   // The vertex count
    _vuint32        mt_un32PCount;   // The panel count
    _vuint32        mt_un32PCR;      // The R component of the panel's color
    _vuint32        mt_un32PCG;      // The G component of the panel's color
    _vuint32        mt_un32PCB;      // The B component of the panel's color
    
  public:
    // Constructor & destructor
    inline _CVortexPanel3DBuffer();
    inline ~_CVortexPanel3DBuffer();
    // Lighting functions
    inline _vvoid FSResetAmbientLight(_TVortexAmbientLight *ptVAmbientLight = NULL);
    inline _vvoid GSResetAmbientLight(_TVortexAmbientLight *ptVAmbientLight = NULL);
    inline _vvoid FSResetOmniLight(_TVortexOmniLight *ptVOmniLight = NULL);
    inline _vvoid GSResetOmniLight(_TVortexOmniLight *ptVOmniLight = NULL);
    inline _vvoid FSResetLights(_TVortexAmbientLight *ptVAmbientLight = NULL, _TVortexOmniLight *ptVOmniLight = NULL);
    inline _vvoid GSResetLights(_TVortexAmbientLight *ptVAmbientLight = NULL, _TVortexOmniLight *ptVOmniLight = NULL);
    // Texturing functions
    inline _vvoid SetColor(_vuint8 un8R, _vuint8 un8G, _vuint8 un8B);
    inline _vvoid SetTexture(_CVortexTexture *pcsTexture);
    inline _vvoid SetTexture(_CVortexAnimatedTexture *pcsAnimatedTexture);
    _vvoid CalcDefUnitTexCoords(_vvoid);
    _vvoid CalcDefTexCoords(_vvoid);
    _vvoid RefitTexCoords(_vvoid);
    // Common rendering functions
    inline _vvoid Transform(_CVortexMatrix3D &csTMatrix);
    inline _vvoid DynTransform(_CVortexMatrix3D &csCameraMatrix);
    inline _vvoid PreDynNSRender(_CVortexMatrix3D &csDynMatrix, _TVortexOmniLight *ptVOmniLightUsed = NULL);
    inline _vvoid PreDynFSRender(_CVortexMatrix3D &csDynMatrix, _TVortexOmniLight *ptVOmniLightUsed = NULL);
    inline _vvoid PreDynGSRender(_CVortexMatrix3D &csDynMatrix, _TVortexOmniLight *ptVOmniLightUsed = NULL);
    // Normal rendering functions
    inline _vvoid UFRender(_CVortexMatrix3D &csTMatrix); // Untextured, flat shaded
    inline _vvoid UGRender(_CVortexMatrix3D &csTMatrix); // Untextured, gouroud shaded
    inline _vvoid CFRender(_CVortexMatrix3D &csTMatrix); // Colored,    flat shaded
    inline _vvoid CGRender(_CVortexMatrix3D &csTMatrix); // Colored,    gouroud shaded
    inline _vvoid TNRender(_CVortexMatrix3D &csTMatrix); // Textured,   not shaded
    inline _vvoid TFRender(_CVortexMatrix3D &csTMatrix); // Textured,   flat shaded
    inline _vvoid TGRender(_CVortexMatrix3D &csTMatrix); // Textured,   gouroud shaded
    inline _vvoid DynUFRender(_CVortexMatrix3D &csTMatrix); // Untextured, flat shaded
    inline _vvoid DynUGRender(_CVortexMatrix3D &csTMatrix); // Untextured, gouroud shaded
    inline _vvoid DynCFRender(_CVortexMatrix3D &csTMatrix); // Colored,    flat shaded
    inline _vvoid DynCGRender(_CVortexMatrix3D &csTMatrix); // Colored,    gouroud shaded
    inline _vvoid DynTNRender(_CVortexMatrix3D &csTMatrix); // Textured,   not shaded
    inline _vvoid DynTFRender(_CVortexMatrix3D &csTMatrix); // Textured,   flat shaded
    inline _vvoid DynTGRender(_CVortexMatrix3D &csTMatrix); // Textured,   gouroud shaded
    // Fogged rendering functions
    inline _vvoid FogUFRender(_CVortexMatrix3D &csTMatrix); // Fogged, untextured, flat shaded
    inline _vvoid FogUGRender(_CVortexMatrix3D &csTMatrix); // Fogged, untextured, gouroud shaded
    inline _vvoid FogCFRender(_CVortexMatrix3D &csTMatrix); // Fogged, colored,    flat shaded
    inline _vvoid FogCGRender(_CVortexMatrix3D &csTMatrix); // Fogged, colored,    gouroud shaded
    inline _vvoid FogTNRender(_CVortexMatrix3D &csTMatrix); // Fogged, textured,   not shaded
    inline _vvoid FogTFRender(_CVortexMatrix3D &csTMatrix); // Fogged, textured,   flat shaded
    inline _vvoid FogTGRender(_CVortexMatrix3D &csTMatrix); // Fogged, textured,   gouroud shaded
    inline _vvoid DynFogUFRender(_CVortexMatrix3D &csTMatrix); // Fogged, untextured, flat shaded
    inline _vvoid DynFogUGRender(_CVortexMatrix3D &csTMatrix); // Fogged, untextured, gouroud shaded
    inline _vvoid DynFogCFRender(_CVortexMatrix3D &csTMatrix); // Fogged, colored,    flat shaded
    inline _vvoid DynFogCGRender(_CVortexMatrix3D &csTMatrix); // Fogged, colored,    gouroud shaded
    inline _vvoid DynFogTNRender(_CVortexMatrix3D &csTMatrix); // Fogged, textured,   not shaded
    inline _vvoid DynFogTFRender(_CVortexMatrix3D &csTMatrix); // Fogged, textured,   flat shaded
    inline _vvoid DynFogTGRender(_CVortexMatrix3D &csTMatrix); // Fogged, textured,   gouroud shaded
    // Builder functions
    _vbool InitFromMemory(_vfloat *pfVertices, _vuint32 un32VertexCount, 
                          _vuint32 *pfPanelPoints, _vuint32 un32PanelCount,
                          _TVortexAmbientLight *ptVAmbientLight = NULL, 
                          _TVortexOmniLight *ptVOmniLight = NULL,
                          _vbool bEnableSmoothShading = true);
    // File I/O functions
    _vbool SaveVPD(const _vchar *pcFileName);
    _vbool LoadVPD(const _vchar *pcFileName);
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 3D panel buffer class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexPanel3DBuffer::_CVortexPanel3DBuffer()
{
  mt_pcsTexture   = NULL;
  mt_pcsVList     = NULL;
  mt_pcsPList     = NULL;
  mt_un32TWidth1  = 0;
  mt_un32THeight1 = 0;
  mt_un32VCount   = 0;
  mt_un32PCount   = 0;
  mt_un32PCR      = MAX_COLOR_VALUE;
  mt_un32PCG      = MAX_COLOR_VALUE;
  mt_un32PCB      = MAX_COLOR_VALUE;
}
_CVortexPanel3DBuffer::~_CVortexPanel3DBuffer()
{
  if(mt_pcsVList) delete [] mt_pcsVList;
  if(mt_pcsPList) delete [] mt_pcsPList;
}

// ***** Lighting functions
_vvoid _CVortexPanel3DBuffer::FSResetAmbientLight(_TVortexAmbientLight *ptVAmbientLight)
{  
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    mt_pcsPList[un32Cnt]._CalcIntenAmbient_(ptVAmbientLight);
    mt_pcsPList[un32Cnt]._FSCalcInten_();
  }
}
_vvoid _CVortexPanel3DBuffer::GSResetAmbientLight(_TVortexAmbientLight *ptVAmbientLight)
{  
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    mt_pcsPList[un32Cnt]._CalcIntenAmbient_(ptVAmbientLight);
    mt_pcsPList[un32Cnt]._GSCalcInten_();
  }
}
_vvoid _CVortexPanel3DBuffer::FSResetOmniLight(_TVortexOmniLight *ptVOmniLight)
{
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    mt_pcsPList[un32Cnt]._FSCalcIntenOmni_(ptVOmniLight);
    mt_pcsPList[un32Cnt]._FSCalcInten_();
  }
}
_vvoid _CVortexPanel3DBuffer::GSResetOmniLight(_TVortexOmniLight *ptVOmniLight)
{
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    mt_pcsPList[un32Cnt]._GSCalcIntenOmni_(ptVOmniLight);
    mt_pcsPList[un32Cnt]._GSCalcInten_();
  }
}
_vvoid _CVortexPanel3DBuffer::FSResetLights(_TVortexAmbientLight *ptVAmbientLight, _TVortexOmniLight *ptVOmniLight)
{
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    mt_pcsPList[un32Cnt]._CalcIntenAmbient_(ptVAmbientLight);
    mt_pcsPList[un32Cnt]._FSCalcIntenOmni_(ptVOmniLight);
    mt_pcsPList[un32Cnt]._FSCalcInten_();
  }
}
_vvoid _CVortexPanel3DBuffer::GSResetLights(_TVortexAmbientLight *ptVAmbientLight, _TVortexOmniLight *ptVOmniLight)
{
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    mt_pcsPList[un32Cnt]._CalcIntenAmbient_(ptVAmbientLight);
    mt_pcsPList[un32Cnt]._GSCalcIntenOmni_(ptVOmniLight);
    mt_pcsPList[un32Cnt]._GSCalcInten_();
  }
}

// ***** Texturing functions
_vvoid _CVortexPanel3DBuffer::SetColor(_vuint8 un8R, _vuint8 un8G, _vuint8 un8B)
{
  mt_un32PCR = _vuint32(un8R);
  mt_un32PCG = _vuint32(un8G);
  mt_un32PCB = _vuint32(un8B);
}
_vvoid _CVortexPanel3DBuffer::SetTexture(_CVortexTexture *pcsTexture)
{
  mt_pcsTexture = pcsTexture;
}
_vvoid _CVortexPanel3DBuffer::SetTexture(_CVortexAnimatedTexture *pcsAnimatedTexture)
{
  mt_pcsTexture = (_CVortexTexture *) pcsAnimatedTexture;
}

// ***** Common rendering functions
_vvoid _CVortexPanel3DBuffer::Transform(_CVortexMatrix3D &csTMatrix)
{
  // ..... Temporary variables
  _vuint32 un32Cnt;
  
  // ..... Transform the vertex list
  for(un32Cnt = 0; un32Cnt < mt_un32VCount; un32Cnt++)
    csTMatrix.Transform(mt_pcsVList[un32Cnt]);

  // ..... Update the panel
  for(un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
    mt_pcsPList[un32Cnt].UpdateFrame();
}
_vvoid _CVortexPanel3DBuffer::DynTransform(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Temporary variables
  _vuint32 un32Cnt;
  
  // ..... Transform the vertex list
  for(un32Cnt = 0; un32Cnt < mt_un32VCount; un32Cnt++)
    csCameraMatrix.PstDynTransform(mt_pcsVList[un32Cnt]);

  // ..... Update the panel
  for(un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
    mt_pcsPList[un32Cnt].UpdateFrame();
}
_vvoid _CVortexPanel3DBuffer::PreDynNSRender(_CVortexMatrix3D &csDynMatrix, _TVortexOmniLight *ptVOmniLightUsed)
{
  // ..... Temporary variables
  _vuint32 un32Cnt;
  
  // ..... Transform the vertex list
  for(un32Cnt = 0; un32Cnt < mt_un32VCount; un32Cnt++)
    csDynMatrix.PreDynTransform(mt_pcsVList[un32Cnt]);
}
_vvoid _CVortexPanel3DBuffer::PreDynFSRender(_CVortexMatrix3D &csDynMatrix, _TVortexOmniLight *ptVOmniLightUsed)
{
  // ..... Static variables
  static _vint8 n8DFSCnt = 0;
  
  // ..... Temporary variables
  _vuint32 un32Cnt;
  
  // ..... Transform the vertex list
  for(un32Cnt = 0; un32Cnt < mt_un32VCount; un32Cnt++)
    csDynMatrix.PreDynTransform(mt_pcsVList[un32Cnt]);

  // ..... Update the lighting once each 25 frames
  if(n8DFSCnt <= 25)
  {
    n8DFSCnt++;
    return;
  }
  n8DFSCnt = 0;
  for(un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    mt_pcsPList[un32Cnt]._DynFSCalcIntenOmni_(ptVOmniLightUsed);
    mt_pcsPList[un32Cnt]._FSCalcInten_();
  }  
}
_vvoid _CVortexPanel3DBuffer::PreDynGSRender(_CVortexMatrix3D &csDynMatrix, _TVortexOmniLight *ptVOmniLightUsed)
{
  // ..... Static variables
  static _vint8 n8DGSCnt = 0;
  
  // ..... Temporary variables
  _vuint32 un32Cnt;
  
  // ..... Transform the vertex list
  for(un32Cnt = 0; un32Cnt < mt_un32VCount; un32Cnt++)
    csDynMatrix.PreDynTransform(mt_pcsVList[un32Cnt]);

  // ..... Update the lighting once each 25 frames
  if(n8DGSCnt <= 25)
  {
    n8DGSCnt++;
    return;
  }
  n8DGSCnt = 0;
  for(un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    mt_pcsPList[un32Cnt]._DynGSCalcIntenOmni_(ptVOmniLightUsed);
    mt_pcsPList[un32Cnt]._GSCalcInten_();
  }  
}

// ***** Normal rendering functions
_vvoid _CVortexPanel3DBuffer::UFRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].UFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].UFRender();
      }
  }
}
_vvoid _CVortexPanel3DBuffer::UGRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].UGProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].UGRender();
      }
  }
}
_vvoid _CVortexPanel3DBuffer::CFRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].UFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].CFRender(mt_un32PCR, mt_un32PCG, mt_un32PCB);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::CGRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].UGProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].CGRender(mt_un32PCR, mt_un32PCG, mt_un32PCB);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::TNRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].TFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].TNRender(mt_pcsTexture);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::TFRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].TFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].TFRender(mt_pcsTexture);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::TGRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
 
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].TGProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].TGRender(mt_pcsTexture);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynUFRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].UFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].UFRender();
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynUGRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].UGProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].UGRender();
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynCFRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].UFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].CFRender(mt_un32PCR, mt_un32PCG, mt_un32PCB);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynCGRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].UGProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].CGRender(mt_un32PCR, mt_un32PCG, mt_un32PCB);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynTNRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].TFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].TNRender(mt_pcsTexture);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynTFRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].TFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].TFRender(mt_pcsTexture);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynTGRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
 
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].TGProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].TGRender(mt_pcsTexture);
      }
  }
}

// ***** Fogged rendering functions
_vvoid _CVortexPanel3DBuffer::FogUFRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogUFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogUFRender();
      }
  }
}
_vvoid _CVortexPanel3DBuffer::FogUGRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogUGProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogUGRender();
      }
  }
}
_vvoid _CVortexPanel3DBuffer::FogCFRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogUFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogCFRender(mt_un32PCR, mt_un32PCG, mt_un32PCB);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::FogCGRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogUGProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogCGRender(mt_un32PCR, mt_un32PCG, mt_un32PCB);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::FogTNRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogTFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogTNRender(mt_pcsTexture);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::FogTFRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogTFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogTFRender(mt_pcsTexture);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::FogTGRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Transform the vertex points
  Transform(csTMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogTGProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogTGRender(mt_pcsTexture);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynFogUFRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogUFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogUFRender();
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynFogUGRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogUGProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogUGRender();
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynFogCFRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogUFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogCFRender(mt_un32PCR, mt_un32PCG, mt_un32PCB);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynFogCGRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogUGProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogCGRender(mt_un32PCR, mt_un32PCG, mt_un32PCB);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynFogTNRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogTFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogTNRender(mt_pcsTexture);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynFogTFRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogTFProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogTFRender(mt_pcsTexture);
      }
  }
}
_vvoid _CVortexPanel3DBuffer::DynFogTGRender(_CVortexMatrix3D &csCameraMatrix)
{
  // ..... Transform the vertex points
  DynTransform(csCameraMatrix);
  
  // ..... Render the panels
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    if(mt_pcsPList[un32Cnt].Invisible() == 0)
      if(mt_pcsPList[un32Cnt].IsVisible3D())
      {
        mt_pcsPList[un32Cnt].FogTGProject();
        if(mt_pcsPList[un32Cnt].IsVisible2D())
          mt_pcsPList[un32Cnt].FogTGRender(mt_pcsTexture);
      }
  }
}

#endif
// ============================================================================================
// End of vp3dbuff.h
// ============================================================================================
