// ############################################################################################
//  vthread.h
//  Definition & implementation of _CVortexThread class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VTHREAD_H_
#define _VTHREAD_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <stdlib.h>
#include <pthread.h>
// ***** VortexGE headers
#include <vbtypes.h>
 
// ============================================================================================
// Foward references of classes
// ============================================================================================
// ***** From "vthread.h"
class _CVortexThread;

// ============================================================================================
// Exported structures definition
// ============================================================================================
// ***** Thread data
typedef struct _TVortexThreadData
{
  _CVortexThread *pcsThread;
  _vbool         *pbExit;
  _vbool         *pbSuspend;
  _vvoid         *pvUserData;
} *_PTVortexThreadData;

// ============================================================================================
// Typedefs
// ============================================================================================
// ***** User thread-callback funtion
// !!!!! CAUTION !!!!!
// The user thread-callback funtion must be constructed like this :
//
// _vuint64 MyCallback(_TVortexThreadData *ptThreadData)
// {
//   while(!(*ptThreadData->pbExit))
//   {
//     if(*ptThreadData->pbSuspend) continue;
//
//     <perform user specific actions>
//
//   }
//   return(<return value>);
// }
//
typedef _vuint64 (*VThreadCallback)(_TVortexThreadData *ptThreadData);

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The thread class
class _CVortexThread
{
  protected:
    // Variables
    _TVortexThreadData mt_tTD;
    pthread_t          mt_ptThreadHandle;
    VThreadCallback    mt_pfnUserCallback;
    _vbool             mt_bExit;
    _vbool             mt_bSuspend;
    _vbool             mt_bTEN;  
    _vint64            mt_un64ExitStatus;
  
  public:
    // Constructor & destructor
    _CVortexThread();
    ~_CVortexThread();
    // Thread functions
    _vbool Initialize(VThreadCallback pfnUserCallback, _vvoid *pvUserData);
    _vbool Start(_vvoid); 
    _vbool Stop(_vuint32 un32TimeOutmS = 2500); 
    _vbool Kill(_vvoid);
    inline _vbool IsRunning(_vvoid);
    inline _vvoid Suspend(_vvoid);
    inline _vvoid Resume(_vvoid);
    inline _vuint64 GetExitStatus(_vvoid);

  // Friend functions
  friend _vvoid *_ThreadFunc_(_vvoid *pvData);
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The thread class
// --------------------------------------------------------------------------------------------
// ***** Thread functions
_vbool _CVortexThread::IsRunning(_vvoid)
{
  return(mt_ptThreadHandle != 0);
}
_vvoid _CVortexThread::Suspend(_vvoid)
{
  mt_bSuspend = true;
}
_vvoid _CVortexThread::Resume(_vvoid)
{
  mt_bSuspend = false;
}
_vuint64 _CVortexThread::GetExitStatus(_vvoid)
{
  return(mt_un64ExitStatus);
}

#endif
// ============================================================================================
// End of vthread.h
// ============================================================================================
