// ############################################################################################
//  vvidbase.h
//  Definition various base types used by _CVortexVideoEnv
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VVIDBASE_H_
#define _VVIDBASE_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** X11 headers 
#include <X11/Xlib.h>
// ***** VortexGE header
#include <vbtypes.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** Color creation macros
#define X11_RGB24(un8R, un8G, un8B) ((_vuint32(un8R) << 16) | (_vuint32(un8G) << 8) | _vuint32(un8B))
#ifdef __USE_RGB16_555__
#define X11_RGB16(un8R, un8G, un8B) ((_vuint16(un8R & 0xF8) << 7) | (_vuint16(un8G & 0xFC) << 2) | _vuint16(un8B >> 3))
#else
#define X11_RGB16(un8R, un8G, un8B) ((_vuint16(un8R & 0xF8) << 8) | (_vuint16(un8G & 0xFC) << 3) | _vuint16(un8B >> 3))
#endif

// ============================================================================================
// Typedefs
// ============================================================================================
// ***** Colors
typedef _vuint32 _TVortexColor;

// ***** Font
#ifdef __USE_X11_VIDEO__
typedef XFontStruct _TVortexFont;
#else
#error No video system selected
#endif

// ============================================================================================
// Exported structures definition
// ============================================================================================
// ***** Triplet uint8 RGB
struct _TVortexUN8RGB
{
  _vuint8 un8B, un8G, un8R;
} __attribute__((packed));

// ============================================================================================
// Exported structures definition
// ============================================================================================
// ***** Style
struct _TVortexStyle
{
  _TVortexFont *ptFont;             // The font
  _vuint32     un32WholeBorder;     // The whole border color
  _vuint32     un32LightBorder;     // The light border color
  _vuint32     un32DarkBorder;      // The dark border color
  _vuint32     un32Face;            // The face color 
  _vuint32     un32Text;            // The text color
  _vuint32     un32IconBorder;      // The icon border color
  _vuint32     un32IconBackground;  // The icon border color
  _vuint32     un32IconInformation; // The icon-information color
  _vuint32     un32IconQuestion;    // The icon-question color
  _vuint32     un32IconWarning;     // The icon-warning color
  _vuint32     un32IconError;       // The icon-error color
  _vint32      n32FillStyle;        // The fill style
};

// ***** The render buffer structure
struct _TVortexRenderBuffer
{
  _vuint32 un32UT;       // Updating types
  _vuint32 un32Width;    // Weight of the buffer
  _vuint32 un32Height;   // Height of the buffer
  _vuint32 un32Size;     // Size of the buffer and the Z buffer in bytes
  _vuint32 un32Size48;   // Size of the buffer and the Z buffer in dword divided by 8
  _vuint32 un32SizeSL;   // Size of the buffer and the Z buffer scanline in bytes
  _vuint32 *pun32LUT;    // Scalling LUT
  _vint32  *pn32ZBuffer; // The Z buffer
  _vuint8  *pun8Buffer;  // The render buffer
  _vfloat  fWidth2;      // Weight of the buffer / 2
  _vfloat  fHeight2;     // Height of the buffer / 2      
  _vint32  n32ZTrans;    // The Z buffer translation factor
  _vint32  n32ZWait;     // The Z buffer clearing delay
};

#endif
// ============================================================================================
// End of vvidbase.h
// ============================================================================================
