// ############################################################################################
//  bench1.cpp
//  Benchmark memory transfer operations
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ############################################################################################
// Header files
// ############################################################################################
// ***** C++ headers
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

// ***** VortexGE headers
#include <vbtypes.h>

// ############################################################################################
// Zero memory
// ############################################################################################
inline _vvoid ZeroMem4X8(_vuint32 *pun32Buffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 *pun32Tmp = pun32Buffer;
  for(_vuint32 un32Cnt = 0; un32Cnt < un32BufferSizeDividedBy32; un32Cnt++)
  {
    *pun32Tmp++ = 0; *pun32Tmp++ = 0; *pun32Tmp++ = 0; *pun32Tmp++ = 0;
    *pun32Tmp++ = 0; *pun32Tmp++ = 0; *pun32Tmp++ = 0; *pun32Tmp++ = 0;
  }
}
inline _vvoid ZeroMem8X4(_vuint32 *pun32Buffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint64 *pun64Tmp = (_vuint64*) pun32Buffer;
  for(_vuint32 un32Cnt = 0; un32Cnt < un32BufferSizeDividedBy32; un32Cnt++)
  {
    *pun64Tmp++ = 0; *pun64Tmp++ = 0; *pun64Tmp++ = 0; *pun64Tmp++ = 0;
  }
}
inline _vvoid ZeroMem4X8ASM(_vuint32 *pun32Buffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 ecx, edi;
  __asm__ __volatile__
  (
    "xor %%eax, %%eax \n"
    "shl $3, %%ecx    \n"
    "cld              \n"
    "rep stosl        \n"
    : "=&D" (edi),       "=&c" (ecx)                       \
    :   "0" (pun32Buffer), "1" (un32BufferSizeDividedBy32) \
    : "eax", "memory"
  );
}
inline _vvoid ZeroMem8X4MMX(_vuint32 *pun32Buffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 ecx, edi;
  __asm__ __volatile__
  (
    "pxor %%mm0, %%mm0       \n"
    "1:                      \n"
    "  movq %%mm0,   (%%edi) \n"
    "  movq %%mm0,  8(%%edi) \n"
    "  movq %%mm0, 16(%%edi) \n"
    "  movq %%mm0, 24(%%edi) \n"
    "  add  $32, %%edi       \n"
    "loop 1b                 \n"
    "emms                    \n"
    : "=&D" (edi),       "=&c" (ecx)                       \
    :   "0" (pun32Buffer), "1" (un32BufferSizeDividedBy32) \
    : "mm0", "memory"
  );
}
inline _vvoid ZeroMem8X4SSE(_vuint32 *pun32Buffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 ecx, edi;
  __asm__ __volatile__
  (
    "pxor %%mm0, %%mm0         \n"
    "1:                        \n"
    "  movntq %%mm0,   (%%edi) \n"
    "  movntq %%mm0,  8(%%edi) \n"
    "  movntq %%mm0, 16(%%edi) \n"
    "  movntq %%mm0, 24(%%edi) \n"
    "  add    $32, %%edi       \n"
    "loop 1b                   \n"
    "emms                      \n"
    : "=&D" (edi),       "=&c" (ecx)                       \
    :   "0" (pun32Buffer), "1" (un32BufferSizeDividedBy32) \
    : "mm0", "memory"
  );
}
inline _vvoid ZeroMem16X2SSE(_vuint32 *pun32Buffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 ecx, edi;
  __asm__ __volatile__
  (
    "xorps %%xmm0, %%xmm0        \n"
    "1:                          \n"
    "  movntps %%xmm0,   (%%edi) \n"
    "  movntps %%xmm0, 16(%%edi) \n"
    "  add     $32, %%edi        \n"
    "loop 1b                     \n"
    : "=&D" (edi),       "=&c" (ecx)                       \
    :   "0" (pun32Buffer), "1" (un32BufferSizeDividedBy32) \
    : "xmm0", "memory"
  );
}

// ############################################################################################
// Fill memory
// ############################################################################################
inline _vvoid FillMem4X8(_vuint32 *pun32Buffer, _vuint32 un32Value, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 *pun32Tmp = pun32Buffer;
  for(_vuint32 un32Cnt = 0; un32Cnt < un32BufferSizeDividedBy32; un32Cnt++)
  {
    *pun32Tmp++ = un32Value; *pun32Tmp++ = un32Value; *pun32Tmp++ = un32Value; *pun32Tmp++ = un32Value;
    *pun32Tmp++ = un32Value; *pun32Tmp++ = un32Value; *pun32Tmp++ = un32Value; *pun32Tmp++ = un32Value;
  }
}
inline _vvoid FillMem8X4(_vuint32 *pun32Buffer, _vuint32 un32Value, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint64 *pun64Tmp = (_vuint64 *) pun32Buffer;
  _vuint64 un64Value = (_vuint64(un32Value) << 32) | _vuint64(un32Value);
  for(_vuint32 un32Cnt = 0; un32Cnt < un32BufferSizeDividedBy32; un32Cnt++)
  {
    *pun64Tmp++ = un64Value; *pun64Tmp++ = un64Value; *pun64Tmp++ = un64Value; *pun64Tmp++ = un64Value;
  }
}
inline _vvoid FillMem4X8ASM(_vuint32 *pun32Buffer, _vuint32 un32Value, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 eax, ecx, edi;
  __asm__ __volatile__
  (
    "shl $3, %%ecx    \n"
    "cld              \n"
    "rep stosl        \n"
    : "=&D" (edi),       "=&a" (eax),     "=&c" (ecx)                       \
    :   "0" (pun32Buffer), "1" (un32Value), "2" (un32BufferSizeDividedBy32) \
    : "memory"
  );
}
inline _vvoid FillMem8X4MMX(_vuint32 *pun32Buffer, _vuint32 un32Value, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 ebx, ecx, edi;
  _vuint64 un64Value = (_vuint64(un32Value) << 32) | _vuint64(un32Value);
  __asm__ __volatile__
  (
    "movq (%%ebx), %%mm0     \n"
    "1:                      \n"
    "  movq %%mm0,   (%%edi) \n"
    "  movq %%mm0,  8(%%edi) \n"
    "  movq %%mm0, 16(%%edi) \n"
    "  movq %%mm0, 24(%%edi) \n"
    "  add  $32, %%edi       \n"
    "loop 1b                 \n"
    "emms                    \n"
    : "=&D" (edi),       "=&b" (ebx),      "=&c" (ecx)                       \
    :   "0" (pun32Buffer), "1" (&un64Value), "2" (un32BufferSizeDividedBy32) \
    : "mm0", "memory"
  );
}
inline _vvoid FillMem8X4SSE(_vuint32 *pun32Buffer, _vuint32 un32Value, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 ebx, ecx, edi;
  _vuint64 un64Value = (_vuint64(un32Value) << 32) | _vuint64(un32Value);
  __asm__ __volatile__
  (
    "movq (%%ebx), %%mm0       \n"
    "1:                        \n"
    "  movntq %%mm0,   (%%edi) \n"
    "  movntq %%mm0,  8(%%edi) \n"
    "  movntq %%mm0, 16(%%edi) \n"
    "  movntq %%mm0, 24(%%edi) \n"
    "  add    $32, %%edi       \n"
    "loop 1b                   \n"
    "emms                      \n"
    : "=&D" (edi),       "=&b" (ebx),      "=&c" (ecx)                       \
    :   "0" (pun32Buffer), "1" (&un64Value), "2" (un32BufferSizeDividedBy32) \
    : "mm0", "memory"
  );
}
inline _vvoid FillMem16X2SSE(_vuint32 *pun32Buffer, _vuint32 un32Value, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 ebx, ecx, edi;
  _vuint64 un64Value = (_vuint64(un32Value) << 32) | _vuint64(un32Value);
  __asm__ __volatile__
  (
    "movhps (%%ebx), %%xmm0      \n"
    "movlps (%%ebx), %%xmm0      \n"
    "1:                          \n"
    "  movntps %%xmm0,   (%%edi) \n"
    "  movntps %%xmm0, 16(%%edi) \n"
    "  add     $32, %%edi        \n"
    "loop 1b                     \n"
    : "=&D" (edi),       "=&b" (ebx),      "=&c" (ecx)                       \
    :   "0" (pun32Buffer), "1" (&un64Value), "2" (un32BufferSizeDividedBy32) \
    : "xmm0", "memory"
  );
}

// ############################################################################################
// Copy memory
// ############################################################################################
inline _vvoid CopyMem4X8(_vuint32 *pun32DstBuffer, _vuint32 *pun32SrcBuffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 *pun32Dst = pun32DstBuffer;
  _vuint32 *pun32Src = pun32SrcBuffer;
  for(_vuint32 un32Cnt = 0; un32Cnt < un32BufferSizeDividedBy32; un32Cnt++)
  {
    *pun32Dst++ = *pun32Src++; *pun32Dst++ = *pun32Src++; *pun32Dst++ = *pun32Src++; *pun32Dst++ = *pun32Src++;
    *pun32Dst++ = *pun32Src++; *pun32Dst++ = *pun32Src++; *pun32Dst++ = *pun32Src++; *pun32Dst++ = *pun32Src++;
  }
}
inline _vvoid CopyMem8X4(_vuint32 *pun32DstBuffer, _vuint32 *pun32SrcBuffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint64 *pun64Dst = (_vuint64 *) pun32DstBuffer;
  _vuint64 *pun64Src = (_vuint64 *) pun32SrcBuffer;
  for(_vuint32 un32Cnt = 0; un32Cnt < un32BufferSizeDividedBy32; un32Cnt++)
  {
    *pun64Dst++ = *pun64Src++; *pun64Dst++ = *pun64Src++; *pun64Dst++ = *pun64Src++; *pun64Dst++ = *pun64Src++;
  }
}
inline _vvoid CopyMem4X8ASM(_vuint32 *pun32DstBuffer, _vuint32 *pun32SrcBuffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 ecx, esi, edi;
  __asm__ __volatile__
  (
    "shl $3, %%ecx \n"
    "cld           \n"
    "rep movsl     \n"
    : "=&D" (edi),          "=&S" (esi),          "=&c" (ecx)                       \
    :   "0" (pun32DstBuffer), "1" (pun32SrcBuffer), "2" (un32BufferSizeDividedBy32) \
    : "memory"
  );
}
inline _vvoid CopyMem8X4MMX(_vuint32 *pun32DstBuffer, _vuint32 *pun32SrcBuffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 ecx, esi, edi;
  __asm__ __volatile__
  (
    "movq   (%%esi), %%mm0   \n"
    "movq  8(%%esi), %%mm1   \n"
    "movq 16(%%esi), %%mm2   \n"
    "movq 24(%%esi), %%mm3   \n"
    "add  $32, %%esi         \n"
    "1:                      \n"
    "  movq %%mm0,   (%%edi) \n"
    "  movq %%mm1,  8(%%edi) \n"
    "  movq %%mm2, 16(%%edi) \n"
    "  movq %%mm3, 24(%%edi) \n"
    "  add  $32, %%edi       \n"
    "  movq   (%%esi), %%mm0 \n"
    "  movq  8(%%esi), %%mm1 \n"
    "  movq 16(%%esi), %%mm2 \n"
    "  movq 24(%%esi), %%mm3 \n"
    "  add  $32, %%esi       \n"
    "loop 1b                 \n"
    "emms                    \n"
    : "=&D" (edi),          "=&S" (esi),          "=&c" (ecx)                       \
    :   "0" (pun32DstBuffer), "1" (pun32SrcBuffer), "2" (un32BufferSizeDividedBy32) \
    : "mm0", "mm1", "mm2", "mm3", "memory"
  );
}
inline _vvoid CopyMem8X4SSE(_vuint32 *pun32DstBuffer, _vuint32 *pun32SrcBuffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 ecx, esi, edi;
  __asm__ __volatile__
  (
    "prefetchnta   (%%esi)     \n"
    "prefetchnta 32(%%esi)     \n"
    "prefetchnta 64(%%esi)     \n"
    "1:                        \n"
    "  prefetchnta 96(%%esi)   \n"
    "  movq     (%%esi), %%mm0 \n"
    "  movq    8(%%esi), %%mm1 \n"
    "  movq   16(%%esi), %%mm2 \n"
    "  movq   24(%%esi), %%mm3 \n"
    "  add    $32, %%esi       \n"
    "  movntq %%mm0,   (%%edi) \n"
    "  movntq %%mm1,  8(%%edi) \n"
    "  movntq %%mm2, 16(%%edi) \n"
    "  movntq %%mm3, 24(%%edi) \n"
    "  add    $32, %%edi       \n"
    "loop 1b                   \n"
    "emms                      \n"
    : "=&D" (edi),          "=&S" (esi),          "=&c" (ecx)                       \
    :   "0" (pun32DstBuffer), "1" (pun32SrcBuffer), "2" (un32BufferSizeDividedBy32) \
    : "mm0", "mm1", "mm2", "mm3", "memory"
  );
}
inline _vvoid CopyMem16X2SSE(_vuint32 *pun32DstBuffer, _vuint32 *pun32SrcBuffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 ecx, esi, edi;
  __asm__ __volatile__
  (
    "prefetchnta   (%%esi)       \n"
    "prefetchnta 32(%%esi)       \n"
    "prefetchnta 64(%%esi)       \n"
    "1:                          \n"
    "  prefetchnta 96(%%esi)     \n"
    "  movaps    (%%esi), %%xmm0 \n"
    "  movaps  16(%%esi), %%xmm1 \n"
    "  add     $32, %%esi        \n"
    "  movntps %%xmm0,   (%%edi) \n"
    "  movntps %%xmm1, 16(%%edi) \n"
    "  add     $32, %%edi        \n"
    "loop 1b                     \n"
    : "=&D" (edi),          "=&S" (esi),          "=&c" (ecx)                       \
    :   "0" (pun32DstBuffer), "1" (pun32SrcBuffer), "2" (un32BufferSizeDividedBy32) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "memory"
  );
}

// ############################################################################################
// Logger
// ############################################################################################
inline _vvoid Log(const _vchar *pcMsg, ...)
{
  // ..... Temporary variables
  _vchar  acBuff[1024];
  va_list vlTmp;
  
  // ..... Extract the pcMsg
  va_start(vlTmp, pcMsg);
  (void) vsnprintf(acBuff, 1024, pcMsg, vlTmp);
  va_end(vlTmp);
  
  // ..... Print the message
  printf(acBuff);
}

// ############################################################################################
// Aligned memory allocation
// ############################################################################################
_vvoid *amalloc(_vuint32 un32Size, _vuint32 un32Alignment = 16)
{
  // ..... Temporary variables
  _vvoid   *pvTmp;
  _vuint32 un32R;

  // ..... Make sure that the alignment >= sizeof(_vvoid *)
  if(un32Alignment < sizeof(_vvoid *)) un32Alignment = sizeof(_vvoid *);
 
  // ..... Allocate aligned memory 
  un32R = posix_memalign(&pvTmp, un32Alignment, un32Size);
  if(un32R)
  {
    if(un32R == EINVAL) printf("The alignment must be a power of two and a multiple of %d (size of _vvoid *)\n", sizeof(_vvoid *));
    if(un32R == ENOMEM) printf("Not enough memory\n");
    return(NULL);
  }
  return(pvTmp);
}

// ############################################################################################
// Validations
// ############################################################################################
_vvoid vfill(_vuchar *pucBuffer, _vuint32 un32Size, _vuchar ucVal)
{
  memset(pucBuffer, ucVal, un32Size);
}
_vbool vcheck(_vuchar *pucBuffer, _vuint32 un32Size, _vuchar ucVal)
{
  _vuchar  *pucBuff = pucBuffer;
  _vuint32 un32Cnt;
  for(un32Cnt = 0; un32Cnt < un32Size; un32Cnt++)
    if(*pucBuff++ != ucVal) return(false);
  return(true);    
}

// ############################################################################################
// Main program (benchmaker)
// ############################################################################################
#define TOT_ITER 256
int main(void)
{
  // ***** Local variables
  _vuchar  *pucSrc, *pucDst;
  _vuint32 un32Cnt, un32Size, un32Size32;
  _vuint64 un64TS, un64TE, un64TC;
  _vbool   bVal;
  
  // ***** Allocate dummy memories
  un32Size = 1024 * 1024; un32Size32 = un32Size >> 5;
  pucSrc = (_vuchar *) amalloc(un32Size);
  pucDst = (_vuchar *) amalloc(un32Size);
  
  // ***** Benchmark ZeroMemory functions
  Log("\nBenchmarking ZeroMemory functions ...\n");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    memset(pucDst, 0, un32Size);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); memset(pucDst, 0, un32Size); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  C++ memset     : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    ZeroMem4X8((_vuint32 *) pucDst, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); ZeroMem4X8((_vuint32 *) pucDst, un32Size32); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  ZeroMem4X8     : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    ZeroMem8X4((_vuint32 *) pucDst, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); ZeroMem8X4((_vuint32 *) pucDst, un32Size32); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  ZeroMem8X4     : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    ZeroMem4X8ASM((_vuint32 *) pucDst, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); ZeroMem4X8ASM((_vuint32 *) pucDst, un32Size32); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  ZeroMem4X8ASM  : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    ZeroMem8X4MMX((_vuint32 *) pucDst, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); ZeroMem8X4MMX((_vuint32 *) pucDst, un32Size32); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  ZeroMem8X4MMX  : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    ZeroMem8X4SSE((_vuint32 *) pucDst, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); ZeroMem8X4SSE((_vuint32 *) pucDst, un32Size32); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  ZeroMem8X4SSE  : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    ZeroMem16X2SSE((_vuint32 *) pucDst, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); ZeroMem16X2SSE((_vuint32 *) pucDst, un32Size32); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  ZeroMem16X2SSE : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  
  // ***** Benchmark FillMemory functions
  Log("\nBenchmarking FillMemory functions ...\n");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    memset(pucDst, 0xF0F0F0F0, un32Size);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); memset(pucDst, 0xF0, un32Size); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  C++ memset     : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    FillMem4X8((_vuint32 *) pucDst, 0xF0F0F0F0, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); FillMem4X8((_vuint32 *) pucDst, 0xF0F0F0F0, un32Size32); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  FillMem4X8     : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    FillMem8X4((_vuint32 *) pucDst, 0xF0F0F0F0, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); FillMem8X4((_vuint32 *) pucDst, 0xF0F0F0F0, un32Size32); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  FillMem8X4     : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    FillMem4X8ASM((_vuint32 *) pucDst, 0xF0F0F0F0, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); FillMem4X8ASM((_vuint32 *) pucDst, 0xF0F0F0F0, un32Size32); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  FillMem4X8ASM  : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    FillMem8X4MMX((_vuint32 *) pucDst, 0xF0F0F0F0, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); FillMem8X4MMX((_vuint32 *) pucDst, 0xF0F0F0F0, un32Size32); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  FillMem8X4MMX  : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    FillMem8X4SSE((_vuint32 *) pucDst, 0xF0F0F0F0, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); FillMem8X4SSE((_vuint32 *) pucDst, 0xF0F0F0F0, un32Size32); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  FillMem8X4SSE  : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    FillMem16X2SSE((_vuint32 *) pucDst, 0xF0F0F0F0, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); FillMem16X2SSE((_vuint32 *) pucDst, 0xF0F0F0F0, un32Size32); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  FillMem16X2SSE : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  
  // ***** Benchmark CopyMemory functions
  memset(pucSrc, 0x88, un32Size); 
  Log("\nBenchmarking CopyMemory functions ...\n");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    memcpy(pucDst, pucSrc, un32Size);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); memcpy(pucDst, pucSrc, un32Size); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  C++ memcpy     : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    CopyMem4X8((_vuint32 *) pucDst, (_vuint32 *) pucSrc, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); CopyMem4X8((_vuint32 *) pucDst, (_vuint32 *) pucSrc, un32Size32); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  CopyMem4X8     : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    CopyMem8X4((_vuint32 *) pucDst, (_vuint32 *) pucSrc, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); CopyMem8X4((_vuint32 *) pucDst, (_vuint32 *) pucSrc, un32Size32); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  CopyMem8X4     : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    CopyMem4X8ASM((_vuint32 *) pucDst, (_vuint32 *) pucSrc, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); CopyMem4X8ASM((_vuint32 *) pucDst, (_vuint32 *) pucSrc, un32Size32); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  CopyMem4X8ASM  : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    CopyMem8X4MMX((_vuint32 *) pucDst, (_vuint32 *) pucSrc, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); CopyMem8X4MMX((_vuint32 *) pucDst, (_vuint32 *) pucSrc, un32Size32); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  CopyMem8X4MMX  : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    CopyMem8X4SSE((_vuint32 *) pucDst, (_vuint32 *) pucSrc, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); CopyMem8X4SSE((_vuint32 *) pucDst, (_vuint32 *) pucSrc, un32Size32); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  CopyMem8X4SSE  : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++)
  {
    un64TS = clock();
    CopyMem16X2SSE((_vuint32 *) pucDst, (_vuint32 *) pucSrc, un32Size32);
    un64TE = clock();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); CopyMem16X2SSE((_vuint32 *) pucDst, (_vuint32 *) pucSrc, un32Size32); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  CopyMem16X2SSE : %7.2f ticks : %s\n", _vdouble(un64TC) / _vdouble(TOT_ITER), bVal ? "OK" : "BAD");
  
  // ***** Free memories
  free(pucSrc);
  free(pucDst);
  
  // ***** End
  return(0);
}

// ############################################################################################
// End of bench1.cpp
// ############################################################################################
