// ############################################################################################
//  rotbox2.cpp
//  A demo application for VortexGE
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vmain.h>

// ============================================================================================
// Local functions prototype
// ============================================================================================
// ***** Initialization functions
_vbool InitializeSystem(_vvoid);
_vbool LoadImages(_vvoid);
_vbool CreateScene(_vvoid);

// ============================================================================================
// Local variables declaration
// ============================================================================================
// ***** Objects
static _CVortexVideoEnv      csVidEnv;
static _CVortex3DEnv         cs3DEnv;
static _CVortexPanel3DBuffer csGround, csBaseBox, csRotBox;
static _CVortexMatrix3D      csMatCamera;
static _CVortexImageBuilder  csImgBuild;
static _CVortexImage         csImgBack;
static _CVortexTexture       csTGround, csTBox;
static _CVortexHNode         csBaseNode, *pcsRotNode1, *pcsRotNode2;
#include <sys/ioctl.h>
#include <linux/vt.h>
#include <errno.h>

// ============================================================================================
// Main program
// ============================================================================================
int main(_vvoid)
{
  // ..... Local variables
  _vuint64          un64Frames, un64Ticks1, un64Ticks2;
  _vfloat           fFPS = 0, fOldFPS = 1, fXDir = _vfloat(0.2);
  _vchar            acBuff[64];
  _vint16           n16XCnt = 200;
  _EVortexKeySymbol eKS;
  
  // ..... Initialization
  if(!InitializeSystem()) return(1);
  if(!LoadImages()) return(1);
  
  // ..... Create scene
  cs3DEnv.SetDefaultOmniLightPosition(40, 200, 150);
  cs3DEnv.SetDefaultOmniLightColor(31, 31, 16);
  CreateScene();
  csGround.SetTexture(&csTGround);
  csGround.CalcDefTexCoords();
  csBaseBox.SetTexture(&csTBox);
  csBaseBox.CalcDefTexCoords();
  csRotBox.SetTexture(&csTBox);
  csRotBox.CalcDefTexCoords();
  cs3DEnv.SetFogState(12, 12, 12, 14);
  
  // ..... Initialize camera
  //       This is the same as in the "rotbox 1"  
  csMatCamera.Identity();
  csMatCamera.Translate(0, -100, -50);
  csMatCamera.Rotate(-495, 0, 0);
  csMatCamera.DumpMatrix(); 
  //       This one is using the new _CVortexMatrix3D::SetToLookAtMatrix() function
  //       Note that this one is easier but slower since SetToLookAtMatrix() does complex calculations
  _vfloat afCamPos[3]    = {0, 100, 50};
  _vfloat afCamTarget[3] = {0, 5, 150};
  csMatCamera.SetToLookAtMatrix(afCamPos, afCamTarget);
  csMatCamera.DumpMatrix(); 
  
  // ..... Initialize hierarchy
  csBaseNode.StcTranslate(0, -5, 150);
  pcsRotNode1 = csBaseNode.AddChild();
  pcsRotNode1->DynSetPosition(0, 20, 0);
  pcsRotNode2 = pcsRotNode1->AddChild();
  pcsRotNode2->StcTranslate(-30, 0, 0);
  pcsRotNode2->StcScale(0.5, 0.5, 0.5);
  pcsRotNode2->DynSetPosition(50, 10, 0);

  // ..... Main loop
  VInfo(_F_, _L_, "Main loop started\n");
  csVidEnv.AutoRepeatOff();
  usleep(100000); csVidEnv.ClearEvent();
  un64Frames = 0;
  un64Ticks1 = cs3DEnv.GetTicks();
  while(1)
  {
    // Calculate FPS
    fFPS = cs3DEnv.CalcFPS();
    if(fOldFPS != fFPS)
    {
      fOldFPS = fFPS;
      sprintf(acBuff, "%05.1f fps", fFPS);
    }
    un64Frames++;
    
    // Update everything if the application is active
    if(csVidEnv.IsActive())
    {
      // Update frame
      csBaseNode.DynTranslate(fXDir , 0, 0);
      pcsRotNode1->DynRotate(0, 10, 0);
      pcsRotNode2->DynRotate(0, -25, 0);
      csBaseNode.UpdateMatrix();
      n16XCnt++;
      if(n16XCnt > 400)
      {
        fXDir = -fXDir;
        n16XCnt = 0 ;
      }

      // Clear scene
      cs3DEnv.ClearRenderBuffer(csImgBack);
      cs3DEnv.ClearZBuffer();
      
      // Render the ground
      csGround.FogTGRender(csMatCamera);
      // Render the base box
      csBaseBox.PreDynGSRender(csBaseNode.Matrix());
      csBaseBox.DynFogTGRender(csMatCamera);
      // Render the 1st rotating box
      csRotBox.PreDynGSRender(pcsRotNode1->Matrix());
      csRotBox.DynFogTGRender(csMatCamera);
      // Render the 2nd rotating box
      csRotBox.PreDynGSRender(pcsRotNode2->Matrix());
      csRotBox.DynFogTGRender(csMatCamera);

      // Update display
      csVidEnv.DisplayRenderImage();
      csVidEnv.SetForegroundColor(csVidEnv.CreateColor(255, 255, 255));
      csVidEnv.DrawText(10, 20, acBuff, 9);
      csVidEnv.DrawText(10, 50, "Press 'A' to move camera one step left", 38);
      csVidEnv.DrawText(10, 65, "Press 'S' to move camera one step right", 39);
      csVidEnv.DrawText(10, 80, "Press 'W' to move camera one step up", 36);
      csVidEnv.DrawText(10, 95, "Press 'Z' to move camera one step down", 38);
      csVidEnv.DrawText(10, 110, "Press 'C' to move camera back to center", 39);
      csVidEnv.DrawText(10, 140, "Press escape to exit ...", 24);
      csVidEnv.UpdateMainWindow();
 
      // Pump event
      csVidEnv.DispatchEvent();
      if(csVidEnv.CheckEvent())
      {
        if(csVidEnv.EventKeyRelease())
        {
          eKS = csVidEnv.GetKeySym();
          if(eKS == VKS_ESC)
            break;
          else if(eKS == 'c')
          {
            afCamPos[0] = 0;
            afCamPos[1] = 100;
            csMatCamera.SetToLookAtMatrix(afCamPos, afCamTarget);
          }
          else if(eKS == 'a')
          {
            if(afCamPos[0] > _vfloat(-30.0))
            { 
              afCamPos[0] -= 10;
              csMatCamera.SetToLookAtMatrix(afCamPos, afCamTarget);
            }
          }
          else if(eKS == 's')
          {
            if(afCamPos[0] < _vfloat(30.0))
            { 
              afCamPos[0] += 10;
              csMatCamera.SetToLookAtMatrix(afCamPos, afCamTarget);
            }
          }
          else if(eKS == 'z')
          {
            if(afCamPos[1] > _vfloat(70.0))
            { 
              afCamPos[1] -= 10;
              csMatCamera.SetToLookAtMatrix(afCamPos, afCamTarget);
            }
          }
          else if(eKS == 'w')
          {
            if(afCamPos[1] < _vfloat(130.0))
            { 
              afCamPos[1] += 10;
              csMatCamera.SetToLookAtMatrix(afCamPos, afCamTarget);
            }
          }
        }
      }
    }
    else
    {
      while(csVidEnv.IsInactive())
      {
        // Calculate FPS
        fFPS = cs3DEnv.CalcFPS();
        if(fOldFPS != fFPS)
        {
          fOldFPS = fFPS;
          sprintf(acBuff, "%05.1f fps", fFPS);
        }
        un64Frames++;
        
        // Update display
        if(csVidEnv.IsVisible())
        {
          csVidEnv.DisplayRenderImage();
          csVidEnv.SetForegroundColor(csVidEnv.CreateColor(255, 255, 255));
          csVidEnv.DrawText(10, 20, acBuff, 9);
          csVidEnv.UseDefaultBigFont();
          csVidEnv.DrawTextXCenter(150, "Inactive State", 14);
          csVidEnv.UseDefaultSmallFont();
          csVidEnv.UpdateMainWindow();
        }
        
        // Pump event
        csVidEnv.DispatchEvent();
        usleep(100000);
      }
      csVidEnv.ClearEvent();
    }
  }
  un64Ticks2 = cs3DEnv.GetTicks();
  
  // ..... Calculate & display the avegare frame rate
  sprintf(acBuff, "Average frame rate = %.2f fps\n \nClick \"OK\" to exit ...", cs3DEnv.CalcAvgFPS(un64Frames, un64Ticks1, un64Ticks2));
  csVidEnv.MessageBox("End Program", acBuff);
  VInfo(_F_, _L_, "Program ended, average frame rate = %.2f FPS", cs3DEnv.CalcAvgFPS(un64Frames, un64Ticks1, un64Ticks2));
  
  // ..... OK !
  return(0);
}

// ============================================================================================
// Local functions implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// Initialization functions
// --------------------------------------------------------------------------------------------
// ***** Initialize system
_vbool InitializeSystem(_vvoid)
{
  // ..... Initialize main stuff
  if(!csVidEnv.CreateMainWindow(640, 480, "Demo : Rotating Box 2")) return(false);
  if(!csVidEnv.CreateRenderImage(640, 480)) return(false);
  if(!csVidEnv.CreateRenderBuffer(320, 240)) return(false);
  if(!cs3DEnv.Initialize(&csVidEnv)) return(false);

  // ..... OK !
  return(true);
}

// ***** Load images
_vbool LoadImages(_vvoid)
{
  // ..... Load background image
  if(!csImgBuild.LoadBMP("../vgmediaf/bmp/back1")) return(false);
  csImgBuild.ResizeToFitBackground();
  csImgBuild.AdjustBrightness(0.25, 0);
  csImgBuild.SmearEffect(5);
  if(!csImgBuild.TransferToImage(csImgBack)) return(false);
  
  // ..... Load texture image 1
  if(!csImgBuild.LoadBMP("../vgmediaf/bmp/tex1")) return(false);
  csImgBuild.Resize(256, 256);
  if(!csImgBuild.TransferToTexture(csTGround)) return(false);

  // ..... Load texture image 2
  if(!csImgBuild.LoadBMP("../vgmediaf/bmp/tex2")) return(false);
  csImgBuild.Resize(64, 64);
  if(!csImgBuild.TransferToTexture(csTBox)) return(false);
  csImgBuild.FreeImage();

  // ..... OK !
  return(true);
}


// ***** Create scene
_vbool CreateScene(_vvoid)
{
  // ..... Temporary variables
  _vfloat  afVG[]    = {-100, -10, 100,   -100, -10, 200,   100, -10, 200,   100, -10, 100,
                        -100,   0, 100,   -100,   0, 200,   100,   0, 200,   100,   0, 100};
  _vuint32 aun32PG[] = {3, 2, 1, 0,   4, 5, 6, 7,
                        0, 4, 7, 3,   3, 7, 6, 2,    2, 6, 5, 1,   1, 5, 4, 0};
  _vfloat  afVB[]    = {0,  0, -20,   -20,  0, 0,   0,  0, 20,   20,  0, 0,
                        0, 35, -20,   -20, 35, 0,   0, 35, 20,   20, 35, 0};
  _vuint32 aun32PB[] = {3, 2, 1, 0,   4, 5, 6, 7,
                        0, 4, 7, 3,   3, 7, 6, 2,    2, 6, 5, 1,   1, 5, 4, 0};
  _vfloat  afVR[]    = {0,  0, -10,   0,  0, 10,   60,  0, 10,   60,  0, -10,
                        0, 10, -10,   0, 10, 10,   60, 10, 10,   60, 10, -10};
  _vuint32 aun32PR[] = {3, 2, 1, 0,   4, 5, 6, 7,
                        0, 4, 7, 3,   3, 7, 6, 2,    2, 6, 5, 1,   1, 5, 4, 0};
  // ..... Create scene
  if(!csGround.InitFromMemory(afVG, 8, aun32PG, 6)) return(false);
  if(!csBaseBox.InitFromMemory(afVB, 8, aun32PB, 6)) return(false);
  if(!csRotBox.InitFromMemory(afVR, 8, aun32PR, 6)) return(false);
  return(true);
}  
 
// ============================================================================================
// End of rotbox.cpp
// ============================================================================================
