// ############################################################################################
//  font.cpp
//  Fonts displayer
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ############################################################################################
// Header files
// ############################################################################################
// ***** C++ headers
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// ***** X11 headers
#include <X11/Xlib.h>
#include <X11/Xutil.h>

// ***** VortexGE headers
#include <vbtypes.h>

// ############################################################################################
// Macros
// ############################################################################################
#define _XLOOPS_ 48
#define _YLOOPS_ 6
  
// ############################################################################################
// Functions
// ############################################################################################
// These characters are actually ASCII codes, if you open this file with Anjuta or KEdit you 
// will see them correctly
//   ⅓⅔⅛⅜⅝⅞←↑→↓↔↕↨▪▫▲►▼◄○●☺☻☼♀♂♠♣♥♦♪♫
//   ⅓ÇüéâäàåçêëèïîìÄÅÉæÆôöòûùÿÖÜ¢£¥₧ƒáíóúñÑªº¿⌐¬½¼¡«»░▒▓│┤╡╢╖╕╣║╗╝╜╛┐└┴┬├─┼╞╟╚╔╩╦╠═╬╧╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀αßΓπΣσµτΦΘΩδ∞φε∩≡±≥≤⌠⌡÷≈³¹†√ⁿ²■₪₫∆⌂
// These are Japanese hiragana and katakana
//   ぁあぃいぅうぇえぉおかがきぎくぐけげこごさざしじすずせぜそぞただちぢっつづてでとどなにぬねのはばぱひびぴふぶぷへべぺほぼぽまみむめもゃやゅゆょよらりるれろゎわゐゑをんゔ゛゜ゝゞ
//   ァアィイゥウェエォオカガキギクグケゲコゴサザシジスズセゼソゾタダチヂッツヅテデトドナニヌネノハバパヒビピフブプヘベペホボポマミムメモャヤュユョヨラリルレロヮワヰヱヲンヴヵヶヷヸヹヺ・ーヽヾ
//   ｡｢｣､･ｦｧｨｩｪｫｬｭｮｯｰｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐﾑﾒﾓﾔﾕﾖﾗﾘﾙﾚﾛﾜﾝﾞﾟ
// Any body know how to translate those unicode to XChar2b ?

// ***** Get XCharStruct
XCharStruct *CVT(XFontStruct *ptFont, _vuint32 un32Char)
{
  // ..... Temporary variables
  _vuint32 un32Rows, un32Pages;
  _vuint32 un32Byte1, un32Byte2;
  _vbool   bValid;

  // ..... Check for validity
  bValid    = true;
  un32Rows  = ptFont->max_byte1 - ptFont->min_byte1 + 1;
  un32Pages = ptFont->max_char_or_byte2 - ptFont->min_char_or_byte2 + 1;
  if(un32Rows == 1)
  {
    un32Byte1 = 0;
    un32Byte2 = 0;
    if((ptFont->min_char_or_byte2 > un32Char) || (ptFont->max_char_or_byte2 < un32Char)) bValid = false;
  }
  else
  {
    un32Byte1 = un32Char >> 8;
    un32Byte2 = un32Char & 0xFF;
    if((ptFont->min_char_or_byte2 > un32Byte2) || (ptFont->max_char_or_byte2 < un32Byte2) ||
       (ptFont->min_byte1 > un32Byte1) || (ptFont->max_byte1 < un32Byte1)) bValid = false;
  }
  
  // ..... Return the char
  if(bValid)
  {
    if(ptFont->per_char)
    {
      if(un32Rows == 1)
        return(ptFont->per_char + (un32Char - ptFont->min_char_or_byte2));
      else
        return(ptFont->per_char + ((un32Byte1 - ptFont->min_byte1) * un32Pages) + (un32Byte2 - ptFont->min_char_or_byte2));
    }
    else
      return(&ptFont->min_bounds);
  }
  return(NULL);
}

// ***** Draw fonts
_vvoid DrawFont(const _vchar *pcFont, const _vchar *pcReg, _vuint8 un8Size, _vuint8 un8Page)
{
  // ..... Temporary variables
  Display       *ptDisp;
  _vint32       n32Screen;
  _vint32       n32Depth;
  XFontStruct   *ptFont[4];
  Window        tWin;
  _vchar        acTitle[256], acFont[256];
  XSizeHints    tSize;
  GC            tGC;
  Pixmap        tPixmap;
  XEvent        tEvent;
  _vint32       n32FW, n32FH, n32FA, n32FT;
  _vint32       n32W, n32H, n32F, n32Y, n32X, n32Pos;
  _vbool        bExit;
  XChar2b       wcP;
  
  // ..... Open display
  ptDisp    = XOpenDisplay(NULL);
  n32Screen = DefaultScreen(ptDisp);
  n32Depth  = DefaultDepth(ptDisp, n32Screen);
  if(n32Depth == 8)
  {
    printf("Must use 16 or 24 bpp !");
    return;
  }

  // ..... Create 1st font
  sprintf(acFont, "*-%s-medium-r-*-*-%u-*-*-*-*-*-%s-*", pcFont, un8Size, pcReg);
  ptFont[0] = XLoadQueryFont(ptDisp, acFont);
  
  // ..... Create 2nd font
  sprintf(acFont, "*-%s-medium-i-*-*-%u-*-*-*-*-*-%s-*", pcFont, un8Size, pcReg);
  ptFont[1] = XLoadQueryFont(ptDisp, acFont);
  if(ptFont[1] == NULL)
  {
    sprintf(acFont, "*-%s-medium-o-*-*-%u-*-*-*-*-*-%s-*", pcFont, un8Size, pcReg);
    ptFont[1] = XLoadQueryFont(ptDisp, acFont);
  }
  
  // ..... Create 3rd font
  sprintf(acFont, "*-%s-bold-r-*-*-%u-*-*-*-*-*-%s-*", pcFont, un8Size, pcReg);
  ptFont[2] = XLoadQueryFont(ptDisp, acFont);
  
  // ..... Create 4th font
  sprintf(acFont, "*-%s-bold-i-*-*-%u-*-*-*-*-*-%s-*", pcFont, un8Size, pcReg);
  ptFont[3] = XLoadQueryFont(ptDisp, acFont);
  if(ptFont[3] == NULL)
  {
    sprintf(acFont, "*-%s-bold-o-*-*-%u-*-*-*-*-*-%s-*", pcFont, un8Size, pcReg);
    ptFont[3] = XLoadQueryFont(ptDisp, acFont);
  }
  
  // ..... Check
  if(ptFont[0] == NULL) return;
  if(ptFont[1] == NULL) return;
  if(ptFont[2] == NULL) return;
  if(ptFont[3] == NULL) return;
  
  // ..... Calculate font size
  n32FW = n32FH = n32FA = 0;
  for(n32F = 0; n32F < 4; n32F++)
  {
    if(ptFont[n32F])
    {
      n32FT = ptFont[n32F]->max_bounds.width;
      if(n32FT > n32FW) n32FW = n32FT;
      n32FT = ptFont[n32F]->ascent + ptFont[n32F]->descent;
      if(n32FT > n32FH) n32FH = n32FT;
      n32FT = ptFont[n32F]->ascent;
      if(n32FT > n32FA) n32FA = n32FT;
    }
  }
  n32FW += 1; n32FH += 1;
  n32W = n32FW * _XLOOPS_;
  n32H = n32FH * _YLOOPS_ * 4;
  if(n32W <= 0) n32W = 10;
  if(n32H <= 0) n32H = 10;
  
  // ..... Create window
  sprintf(acTitle, "%s, %s, page %u, %upt ", pcFont, pcReg, un8Page, un8Size);
  tWin = XCreateSimpleWindow(ptDisp, RootWindow(ptDisp, n32Screen),
                              0, 0, n32W, n32H,
                              4, WhitePixel(ptDisp, n32Screen),
                              BlackPixel(ptDisp, n32Screen));
  tSize.flags      = PMinSize | PMaxSize;
  tSize.min_width  = tSize.max_width  = n32W;
  tSize.min_height = tSize.max_height = n32H;
  XSetStandardProperties(ptDisp, tWin,
                         acTitle, acTitle, None, NULL, 0, &tSize);
  XSelectInput(ptDisp, tWin, ButtonPressMask | ExposureMask);
  XMapWindow(ptDisp, tWin); 
  XMoveWindow(ptDisp, tWin, 10, 10);

  // ..... Create GC
  tGC = XCreateGC(ptDisp, RootWindow(ptDisp, n32Screen), 0, NULL);
  
  // ..... Create pixmap
  tPixmap = XCreatePixmap(ptDisp, tWin, n32W, n32H, n32Depth);
  
  // ..... Print characters
  XSetForeground(ptDisp, tGC, 0x00000000);
  XFillRectangle(ptDisp, tPixmap, tGC, 0, 0, n32W, n32H);
  XSetForeground(ptDisp, tGC, 0xFFFFFFFF);
  wcP.byte1 = un8Page;
  for(n32F = 0; n32F < 4; n32F++)
  {
    if(ptFont[n32F])
    {
      XSetFont(ptDisp, tGC, ptFont[n32F]->fid);
      n32Pos = 0;
      for(n32Y = _YLOOPS_ * n32F; n32Y < (_YLOOPS_ * n32F + _YLOOPS_); n32Y++)
        for(n32X = 0; n32X < _XLOOPS_; n32X++)
        {
          if((_vuint8(n32Pos) >= ptFont[n32F]->min_char_or_byte2) && 
             (_vuint8(n32Pos) <= ptFont[n32F]->max_char_or_byte2))
          {
            wcP.byte2 = _vuchar(n32Pos);
            XDrawString16(ptDisp, tPixmap, tGC, n32X * n32FW, n32Y * n32FH + n32FA, &wcP, 1);
          }
          n32Pos++;
          if(n32Pos > 255) n32Pos = 255;
        }
    }
  }
  
  // ..... Main loop
  bExit = false;
  while(!bExit)
  {
    // Handle events
    if(XPending(ptDisp))
    {
      XNextEvent(ptDisp, &tEvent);
      switch(tEvent.type)
      {  
        case Expose:
          XCopyArea(ptDisp, tPixmap, tWin, tGC, 0, 0, n32W, n32H, 0, 0);
          break;
        case ButtonPress:
          bExit = true;   
          break;
        default:
          break;
      }
    } 
  }
   
  // ..... Clean up
  XFreePixmap(ptDisp, tPixmap);
  XFreeGC(ptDisp, tGC);
  XUnmapWindow(ptDisp, tWin); 
  XDestroyWindow(ptDisp, tWin);
  if(ptFont[0]) XFreeFont(ptDisp, ptFont[0]);
  if(ptFont[1]) XFreeFont(ptDisp, ptFont[1]);
  if(ptFont[2]) XFreeFont(ptDisp, ptFont[2]);
  if(ptFont[3]) XFreeFont(ptDisp, ptFont[3]);
  XCloseDisplay(ptDisp);
}

// ############################################################################################
// Main programs
// ############################################################################################
int main(_vvoid)
{
  // ..... Temporary variables
  _vuint8 aun8S[] = {8, 10, 12, 14, 16};
  _vuint8 un8C, un8M = sizeof(aun8S);
  
  // ..... Draw fonts
  for(un8C = 0; un8C < un8M; un8C++) DrawFont("courier", "iso10646", aun8S[un8C], 0);
  for(un8C = 0; un8C < un8M; un8C++) DrawFont("mincho", "jisx0208.1983", aun8S[un8C], 36);
  for(un8C = 0; un8C < un8M; un8C++) DrawFont("mincho", "jisx0208.1983", aun8S[un8C], 37);
  
  // ..... OK !
  return(0);
}

// ############################################################################################
// End of font.cpp
// ############################################################################################
